/*
 *    Brutus source file implementing a Windows(R) INI file helper class.
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "inifile.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#if defined(_WIN32)
#include <sys/types.h>
#define strcasecmp(__str1__, __str2__) _stricmp(__str1__, __str2__) 
#endif


// converts the null terminated string 'str' into dosmode (![\r]\n ==> \r\n)
static bool dosmode(char **str)
{
	if ((!(*str)) || (!(*str[0])))
		return true;

	char *retv = NULL;

	// take care of short strings first
	size_t input_len = strlen(*str);
	if (1 == input_len) {
		if ('\n' == (*str)[0]) {
			retv = (char*)realloc((void*)(*str), sizeof(char)*3);
			if (!retv)
				return false;

			(*str) = retv;
			(*str)[0] = '\r';
			(*str)[1] = '\n';
			(*str)[2] = '\0';
		}
		return true;
	}

	// get number of \n's which are not preceeded by \r
	off_t pos = sizeof(char);
	size_t size = input_len + sizeof(char); // reserve room for \0
	while (true) {
		if (!(*str)[pos])
			break;
		
		if ('\n' == (*str)[pos]) {
			if ('\r' != (*str)[pos-1]) 
				size++;			
		}
		pos++;
	}

	retv = (char *)malloc(sizeof(char)*size);
	if (!retv)
		return false;
	memset((void*)retv, '\0', size);

	// compose dosmode string
	pos = 0;
	off_t retv_pos = 0;
	while (true) {
		if (!(*str)[pos])
			break;
		
		if ('\n' == (*str)[pos]) {
			if (pos && ('\r' != (*str)[pos-1]))
				retv[retv_pos++] = '\r';
			else if (!pos)
				retv[retv_pos++] = '\r';
		}
		retv[retv_pos] = (*str)[pos];
		retv_pos++;
		pos++;
	}
	free(*str);
	*str = retv;
	
	return true;
}

/* Will return a null terminated string containg the all text
 * from the present position in the file until the first INI 
 * section or EOF. 
 *
 * Returns NULL if empty or on error. */
static char *get_data(FILE *file)
{
	size_t size = 0;
	char *retv = NULL;
	bool in_comment = false;
	off_t pos = ftell(file);
	char c = EOF;
	bool empty = true;

	// get size of head
	while (1) {
		// For clarity, not speed
		if (!fread((void*)&c, 1, 1, file))
			break;

		if (';' == c) {
			in_comment = true;
			empty = false;
			goto next;
		}

		if (in_comment && (c == '\n')) {
			in_comment = false;
			goto next;
		}

		if (!in_comment) {
			if ((c == '[') || (c == EOF))
				break; 
			if (empty) 
				empty = !!isspace(c);
		}
	next:		
		size++;
		continue;
	}
	if (ferror(file))
		return NULL; 
	
	if (empty) {
		if (c == '[')
			fseek(file, -1, SEEK_CUR); // the next character must be '['
		return NULL;
	}
	fseek(file, pos, SEEK_SET);

	// record head
	retv = (char*)malloc(sizeof(char)*(size + 1)); // make room for '\0'
	for (size_t n=0; n<size; n++) {
		fread((void*)&c, 1, 1, file);
		retv[n] = c;
		pos++;
	} // the file position indicator is now pointing at '[' or EOF
	retv[size] = '\0';

	if (!dosmode(&retv)) {
		free(retv);
		return NULL;
	}

	return retv;
}

/* Will return a null terminated string containg the all text
 * from the present position in the file until and including 
 * the first '\n'.
 *
 * Returns NULL on error. */
static char *get_line(FILE *file)
{
	size_t size = 0;
	char *retv = NULL;
	off_t pos = ftell(file);
	char c = EOF;
	bool empty = true;

	// get size of section
	while (1) {
		// For clarity, not speed
		if (!fread((void*)&c, 1, 1, file))
			break;
		size++;

		if (empty) 
			empty = !isalnum(c);

		if ((c == '\n') || (c == EOF))
			break;		

		continue;
	}
	if (ferror(file) || empty)
		return NULL; // error

	fseek(file, pos, SEEK_SET);

	// record line
	retv = (char*)malloc(sizeof(char)*(size + 1)); // make room for '\0'
	for (size_t n=0; n<size; n++) {
		fread((void*)&c, 1, 1, file);
		retv[n] = c;
		pos++;
	}
	retv[size] = '\0';

	if (!dosmode(&retv)) {
		free(retv);
		return NULL;
	}

	return retv;
}

/* Reads all of file_name into recv.
 *
 * There must be no '\0' or single '\r' characters in the 
 * specified input file. \r\n are allowed. ![\r]\n\r are not
 * allowed. No empty sections are allowed.
 *
 * I am reading one byte of the file at a time. This is to 
 * avoid a second copy/allocation. Not that this method is 
 * particular fast, but it makes the purpose so much clearer.
 * This function is not a bottleneck, therefore clarity before 
 * speed.
 *
 * Returns false on error. */
static bool read_file(const char *file_name, IniFile *recv)
{
	FILE *file = NULL;
	char *section = NULL;
	char *data = NULL;

	file = fopen(file_name, "r");
	if (!file) 
		return false;
	
	// get heading
	data = get_data(file); // next character is now '[' or EOF
	if ((data) && (strlen(data))) {
		recv->add_head(data);
		free(data);
	}

	// EOF?
	if (feof(file))
		goto out;

	// the next character is now '['
	while (1) {
		section = get_line(file);
		if (!section)
			goto out;
		
		data = get_data(file);
		if (!data) 
			data = strdup(""); // the next iteration will exit the loop

		recv->add_section((const char*)section, (const char*)data);
		free(section);
		free(data);

		if (feof(file))
			break;
	}

out:
	fclose(file);

	return true;
}

IniFile::IniFile(const char *name) 
        : head_(NULL),
	  content_(NULL),
	  section_count_(0)
{
	if (name)
		read_file(name, this);
}

IniFile::~IniFile()
{
	for (unsigned long n=0; n<section_count_; n++) {
		free(content_[n].name);
		free(content_[n].data);
	}
	free(content_);
}

bool IniFile::add_head(const char *head)
{
	char *tmp_head = strdup(head);

	if (!dosmode(&tmp_head)) {
		free(tmp_head);
		return false;
	}


	void *tmp = NULL;
	if (head_) {
		tmp = realloc((void*)head_, (sizeof(char)*strlen(head_)) + (sizeof(char)*strlen(tmp_head)) + sizeof(char));
		if (!tmp) {
			free(tmp_head);
			return false;
		}

		head_ = (char*)tmp;
		strcat(head_, tmp_head);
		free(tmp_head);

		return true;
	} else
		head_ = tmp_head;

	return true;
}

bool IniFile::duplicate(const char *section_name, const char *data)
{
	bool retv = false;
	char *in_data = NULL;
	char *query_data = NULL;

	if (!section_name)
		return false;
	if (!data)
		return false;
	if (!strlen(section_name))
		return false;
	if (!strlen(data))
		return false;

	// convert the substring to uppercase and remove \r\n from the end
	query_data = strdup(data);
	if (!query_data)
		return false;
	if ('\n' == query_data[strlen(query_data)-1])
		query_data[strlen(query_data)-1] = '\0';
	if ('\r' == query_data[strlen(query_data)-2])
		query_data[strlen(query_data)-2] = '\0';
	if (!_strupr(query_data))
		goto out;

	// look for duplicate
	for (unsigned long n=0; n<section_count_; n++) {
		if (!strcasecmp(content_[n].name, section_name)) {
			in_data = strdup(content_[n].data);
			if (!in_data)
				goto out;
			if (!_strupr(in_data))
				goto out;

			if (strstr(in_data, query_data))
				retv = true;

			goto out;
		}
	}

out:
	free(in_data);
	free(query_data);

	return retv;
}

bool IniFile::add_section(const char *section_name, const char *data)
{
	char *tmp_section_name = strdup(section_name);
	char *tmp_data = strdup(data);

	if (!dosmode(&tmp_section_name)) {
		free(tmp_section_name);
		free(tmp_data);
		return false;
	}

	if (!dosmode(&tmp_data)) {
		free(tmp_section_name);
		free(tmp_data);
		return false;
	}

	if (duplicate(tmp_section_name, tmp_data)) {
		free(tmp_section_name);
		free(tmp_data);
		return true;
	}

	// look for pre-existing section
	void *tmp = NULL;
	for (unsigned long n=0; n<section_count_; n++) {
		if (!strcasecmp(content_[n].name, tmp_section_name)) {
			free(tmp_section_name);
			tmp = realloc((void*)content_[n].data, (sizeof(char)*strlen(content_[n].data)) + (sizeof(char)*strlen(tmp_data)) + sizeof(char));
			if (!tmp) {
				free(tmp_data);
				return false;
			}
			content_[n].data = (char*)tmp;
			
			strcat(content_[n].data, (const char*)tmp_data);
			free(tmp_data);

			return true;
		}
	}

	// new section with data
	tmp = (void*)content_;
	content_ = (struct section*)realloc(tmp, sizeof(struct section)*(++section_count_));
	if (!content_) {
		content_ = (struct section*)tmp;
		return false;
	}
	content_[section_count_-1].name = tmp_section_name;
	content_[section_count_-1].data = tmp_data;

	return true;
}

bool IniFile::add(const IniFile & ini_file)
{
	const size_t add_count = ini_file.section_count_;

	bool retv = true;
	if (ini_file.head_) 
		retv = retv && add_head((const char*)ini_file.head_);

	for (unsigned long n=0; n<add_count; n++) {
		retv = retv && add_section((const char*)ini_file.content_[n].name, (const char*)ini_file.content_[n].data);
	}

	return retv;
}

int IniFile::commit(const char *file_name)
{
	FILE *file = NULL;

	file = fopen(file_name, "wb");
	if (!file) 
		return false;

	bool retv = false;
	if (head_) {
		fwrite((const void*)head_, sizeof(char), strlen(head_), file);
		if (ferror(file))
			goto out;
	}

	for  (unsigned long n=0; n<section_count_; n++) {
		fwrite((const void*)content_[n].name, sizeof(char), strlen(content_[n].name), file);
		if (ferror(file))
			goto out;

		fwrite((const void*)content_[n].data, sizeof(char), strlen(content_[n].data), file);
		if (ferror(file))
			goto out;
	}
	retv = true;

out:
	fclose(file);

	return retv;
}


