/*
 *    Brutus source file implementing the ProxyManager servant.
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <NT_Service/ProxyMgr/proxy_managerS_impl.h>

struct thread_args {
        BRUTUS_ProxyManager_i *proxy_manager;
        HANDLE proxy_handle;
        pid_t proxy_pid;
};

// Will remove the proxy entry from the list if it terminates by it-self
DWORD WINAPI ExitThread(LPVOID lpParameter)
{
        struct thread_args *args = (struct thread_args*)lpParameter;

        WaitForSingleObject(args->proxy_handle, INFINITE);
        args->proxy_manager->removeProxy(args->proxy_pid);

        HeapFree(GetProcessHeap(), 0, lpParameter);
        
        return 0;
}


/* This method will be invoked by the factory client (BrutusLogon::)
 * if PROCESS_PER_SESSION is enabled */
CORBA::Boolean BRUTUS_ProxyManager_i::setUser(const char *LifeLineIOR,
					      const char *UserName,
					      const char *Domain,
					      const char *Password,
					      const char *MAPIProfileName,
					      const char *MAPIProfilePassword,
					      ::CORBA::ULong MaxNotificationDelay,
					      ::CORBA::ULong Flags
					      ACE_ENV_ARG_DECL)
	ACE_THROW_SPEC((CORBA::SystemException))
{
	User *user_data = new User(LifeLineIOR, UserName, Domain, Password, MAPIProfilePassword, (unsigned long)MaxNotificationDelay, (unsigned long)Flags);
	if (!user_data)
		throw CORBA::NO_MEMORY();

	if (!user_data->create_token())
		return (::CORBA::Boolean)0;

        std::pair<std::string, User*> spawn_entry(MAPIProfileName, user_data);
	std::pair<SLit, bool> p;
	while (true) {
		{
			ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_);
			p = spawn_list_.insert(spawn_entry);
		}
		if (p.second)
			break;
		else
			throw CORBA::NO_MEMORY();			
	}

	return (::CORBA::Boolean)1;
}

/* This method (or setSpawnStatus()) is invoked by a proxy process 
 * to get the data necessary to bootstrap an IMAPISession for a Brutus 
 * client. One of setSession() or setSpawnStatus() is subsequently 
 * invoked by the proxy */
void BRUTUS_ProxyManager_i::getUser(CORBA::String_out LifeLineIOR,
				    CORBA::String_out MAPIProfileName,
				    CORBA::String_out UserName,
				    CORBA::String_out Domain,
				    CORBA::String_out Password,
				    CORBA::String_out MAPIProfilePassword,
				    ::CORBA::ULong_out MaxNotificationDelay,
				    ::CORBA::ULong_out Flags
				    ACE_ENV_ARG_DECL)
	ACE_THROW_SPEC((CORBA::SystemException))
{
	CORBA::String_var life_line_ior = (const char*)"";
	CORBA::String_var mapi_profile = (const char*)"";
	CORBA::String_var user = (const char*)"";
	CORBA::String_var domain = (const char*)"";
	CORBA::String_var password = (const char*)"";
	CORBA::String_var profile_password = (const char*)"";

	{
		ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_);
		SLit i;

		for (i = spawn_list_.begin(); i != spawn_list_.end(); i++) 
			if (!i->second->taken) {
				i->second->taken = true;
				break;
			}
                
		if (i != spawn_list_.end()) {
			mapi_profile = CORBA::string_dup(i->first.c_str());
			life_line_ior = CORBA::string_dup(i->second->life_line_ior);
			user = CORBA::string_dup(i->second->name);
			domain = CORBA::string_dup(i->second->domain);
			password = CORBA::string_dup(i->second->password);
			profile_password = CORBA::string_dup(i->second->mapi_pwd);
			MaxNotificationDelay = (::CORBA::ULong)i->second->max_notification_delay;
			Flags = (::CORBA::ULong)i->second->flags;
		}
	}

	LifeLineIOR = life_line_ior._retn();
	MAPIProfileName = mapi_profile._retn();
	UserName = user._retn();
	Domain = domain._retn();
	Password = password._retn();
	MAPIProfilePassword = profile_password._retn();

        return;
}


/* This method (or setSpawnStatus()) is invoked by a proxy process 
 * when a IMAPISession is ready for a Brutus client. No other ProxyManager
 * methods are then invoked by the proxy */
void BRUTUS_ProxyManager_i::setSession(const char *MAPIProfileName,
                                       const char *SessionIOR,
				       ::CORBA::Boolean Sharable
                                       ACE_ENV_ARG_DECL)
        ACE_THROW_SPEC((CORBA::SystemException))
{
	ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_);
	SLit i = spawn_list_.find(MAPIProfileName);
	if (i != spawn_list_.end()) {
		i->second->ior = CORBA::string_dup(SessionIOR);
		i->second->status = ::BRUTUS::ProxyManager::spawn_ok;
		i->second->sharable = Sharable;
	}

        return;
}

/* This method (or setSession()) is invoked by a proxy process 
 * when a IMAPISession could not be created for a Brutus client. 
 * No other ProxyManager methods are then invoked by the proxy */
void BRUTUS_ProxyManager_i::setSpawnStatus(const char *MAPIProfileName,
					   ::BRUTUS::ProxyManager::SpawnStatus Status
					   ACE_ENV_ARG_DECL)
	ACE_THROW_SPEC((CORBA::SystemException))
{
	ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_);
	SLit i = spawn_list_.find(MAPIProfileName);
	if (i != spawn_list_.end())
		i->second->status = Status;

        return;
}

/* This method will be invoked by the factory client (BrutusLogon::)
 * immediately after setUser() */
::BRUTUS::ProxyManager::SpawnStatus BRUTUS_ProxyManager_i::getSession(const char *MAPIProfileName,
								      CORBA::String_out SessionIOR
								      ACE_ENV_ARG_DECL)
	ACE_THROW_SPEC((CORBA::SystemException))
{
	::BRUTUS::ProxyManager::SpawnStatus retv = ::BRUTUS::ProxyManager::spawn_error_spawn;
	CORBA::String_var ior_str = (const char*)"";
	static HANDLE wait_obj = CreateEvent(NULL, TRUE, FALSE, NULL);
	HANDLE exit_thread = NULL;
	pid_t proxy_pid = 0;
	
	SLit i;
	while (true) { 
		{
			ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_);
			i = spawn_list_.find(MAPIProfileName);
		}
		if (i != spawn_list_.end())
			break;
		WaitForSingleObject(wait_obj, 50); // sleep 50 miliseconds
	}

	if (!(i->second->flags | ::BRUTUS::BRUTUS_UNIQUE_SESSION)) { // see if we can/may re-use an existing session proxy...
		ACE_Write_Guard<ACE_RW_Mutex> guard(proxy_mutex_);
		unsigned long len = proxy_list_.size();
		unsigned long p;
		for (p = 0; p < len; p++) {
			if (!strcmp(MAPIProfileName, proxy_list_[p]->get_mapi_profile())) {
				if (!proxy_list_[p]->sharable)
					continue;

				ior_str = (const char*)proxy_list_[p]->get_session_ior_string(); // copy ior string
				retv = ::BRUTUS::ProxyManager::spawn_duplicate;
				goto out;
			}
		}
	}

	// spawn a proxy process...
	ProxyProcess *proxy = NULL;
	try {
		proxy = new ProxyProcess();
        }
        catch (std::bad_alloc &) {
                throw CORBA::NO_MEMORY();
        }

	if (-1 == proxy->spawn(i->second->primary_token)) { // spawn failed
		delete proxy;
		goto out;
	}

	{ // spawn succeeded
		ACE_Write_Guard<ACE_RW_Mutex> guard(proxy_mutex_); 
		if (!proxy->set_mapi_profile(MAPIProfileName)) {
			delete proxy;
			throw CORBA::NO_MEMORY();
		}
			
		proxy_pid = proxy->getpid();

                // launch thread to monitor the proxy process  
                struct thread_args *args = NULL;
                
                try {
                        args = (struct thread_args*)HeapAlloc(GetProcessHeap(), 
                                                              0, 
                                                              sizeof(struct thread_args));
                }
                catch (...) {
                        delete proxy;
			throw CORBA::NO_MEMORY();
                }
                if (!args) {
                        delete proxy;
			throw CORBA::NO_MEMORY();
                }

                args->proxy_manager = this;
                args->proxy_handle = proxy->gethandle();
                args->proxy_pid = proxy_pid;
                exit_thread = CreateThread(NULL,
                                           0,
                                           ExitThread,
                                           (void*)args,
                                           CREATE_SUSPENDED,
                                           NULL);
                proxy->set_exit_thread_handle(exit_thread);
                proxy_list_.push_back(proxy);
	}

        // The entry in proxy_list_ must be completed 
        // before the thread is started. The reason is 
        // that the thread might, if the proxy terminated 
        // early, call this->removeProxy() immediately. We 
        // therefore do not activate the thread before we 
        // are out of the mutex protected scope.
        ResumeThread(exit_thread);

        // wait for status to change
        retv = ::BRUTUS::ProxyManager::spawn_unknown;
	while (::BRUTUS::ProxyManager::spawn_unknown == retv) {
		{
			ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_); 
			retv = i->second->status;
		}
                        
		WaitForSingleObject(wait_obj, 50); // sleep 50 miliseconds

		{
			ACE_Write_Guard<ACE_RW_Mutex> guard(proxy_mutex_); 
			if (!proxy->running()) // the proxy entry will be removed by the exit thread
				goto proxy_dead;

			continue;
		}
		
	proxy_dead: // the proxy process terminated
		{
			ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_); 
			retv = i->second->status;
		}
		break;
	}

	// the ior member is now populated because of invocation order decipline 
	if (::BRUTUS::ProxyManager::spawn_ok == retv) {
		ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_);
		ior_str = (char*)i->second->ior; // saves a CORBA::string_dup()
		if (!proxy->set_session_ior_string(ior_str.in())) {
			delete proxy;
			throw CORBA::NO_MEMORY();
		}
		i->second->ior = NULL; // to avoid double free()'ing

		proxy->sharable = i->second->sharable ? true : false;
	}

out:
	SessionIOR = ior_str._retn();

	// delete the spawn entry
	delete i->second;
	{
		ACE_Write_Guard<ACE_RW_Mutex> guard(spawn_mutex_);
		spawn_list_.erase(i);
	}

	return retv;
}

CORBA::ULong BRUTUS_ProxyManager_i::getProxyCount(ACE_ENV_SINGLE_ARG_DECL)
	ACE_THROW_SPEC((CORBA::SystemException))
{
	ACE_Write_Guard<ACE_RW_Mutex> guard(proxy_mutex_); 

	return (CORBA::ULong)proxy_list_.size();
}

// no harm done if the proxy entry do not exist
void BRUTUS_ProxyManager_i::removeProxy(pid_t pid)
{
	ACE_Write_Guard<ACE_RW_Mutex> guard(proxy_mutex_); 
	
	for (unsigned long n = 0; n < proxy_list_.size(); n++) {
		if (pid == proxy_list_[n]->getpid()) {
			proxy_list_[n]->kill();
			Sleep(0);
			delete proxy_list_[n];
			proxy_list_.erase(proxy_list_.begin() + n);

			break;
		}
	}
}
