/*
 *    Brutus source file implementing the main() function.
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <vector>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <malloc.h>

#include <TAO/tao/corba.h>

#include <NT_Service/macros.h>
#include <NT_Service/config.h>
#include <NT_Service/defines.h>
#include <NT_Service/poa_utils.h>
#include <NT_Service/conv_utils.h>
#include <NT_Service/servant_activator.h>

#include <NT_Service/servant_impl/IMAPISessionS_impl.h>
#include <NT_Service/ProxyMgr/proxy_managerC.h>
#include <idl/BrutusCheckC.h>
#include <idl/return_codesC.h>
#include <idl/definesC.h>

#include <windows.h>
#include <userenv.h>
#include <lm.h>
#define MAPIGUID_H // do not include mapiguid.h
#include <mapix.h>
#include <NT_Service/templates.i>

// termination signal
volatile HANDLE stop_ = NULL;

static inline LPWSTR asci_to_unicode(LPCSTR str)
{
        LPWSTR wstr = NULL;

        DWORD len = lstrlen(str);
        if (len > 0) {
                // get buffer size and allocate utf8 string buffer
                int wlen = MultiByteToWideChar(CP_ACP, 0, str, -1, NULL, 0);
                wstr = (LPWSTR)calloc(wlen, sizeof(wchar_t));
                if (!wstr)
                        goto out;

                // convert to UTF8
                if (!MultiByteToWideChar(CP_UTF8, 0, str, -1, wstr, wlen)) {
                        free(wstr);
                        wstr = NULL;
                        goto out;
                }
        }

out:
        return wstr;
}

static inline char *widechar_to_utf8(LPCWSTR wstr)
{
        char *utf8_str = NULL;

        DWORD wlen = lstrlenW(wstr);
        if (wlen > 0) {
                // get buffer size and allocate utf8 string buffer
                int utf8_len = WideCharToMultiByte(CP_UTF8, 0, wstr, wlen, NULL, 0, NULL, NULL);
                utf8_str = (char*)calloc(utf8_len, sizeof(char));
                if (!utf8_str)
                        goto out;

                // convert to UTF8
                if (!WideCharToMultiByte(CP_UTF8, 0, wstr, wlen, utf8_str, utf8_len, NULL, NULL)) {
                        free(utf8_str);
                        utf8_str = NULL;
                        goto out;
                }
        }

out:
        return utf8_str;
}

static bool get_domain_controller(const char *domain, 
                                  char **domain_controller)
{
        bool retv = false;
        NET_API_STATUS st; // retval from net query
        LPWSTR dc = NULL; // domain controller
        LPWSTR w_domain = NULL;

        *domain_controller = NULL;

        w_domain = asci_to_unicode(domain);
        if (domain && !w_domain)
                goto out;

        // get domain controller name
        st = NetGetAnyDCName(NULL, w_domain, (LPBYTE *)&dc);
        if (st != NERR_Success) 
                goto out;

        *domain_controller = widechar_to_utf8(dc);

out:
        free(w_domain);
        if (dc) 
                NetApiBufferFree(dc);

        return retv;
}

// Returns true of successful. Caller must free profile_path if 
// the function succeeds.
static bool get_profile_path(const char *user,
                             const char *domain, 
                             char **profile_path)
{
        bool retv = false;
        NET_API_STATUS st; // retval from net query
        LPUSER_INFO_3 pinfo = NULL; // level 3 user info
        LPWSTR dc = NULL; // domain controller
        LPWSTR w_domain = NULL;
        LPWSTR w_user = NULL;

        *profile_path = NULL;

        w_domain = asci_to_unicode(domain);
        if (domain && !w_domain)
                goto out;

        w_user = asci_to_unicode(user);
        if (!w_user)
                goto out;

        // get domain controller
        st = NetGetAnyDCName(NULL, w_domain, (LPBYTE *)&dc);
        if (st != NERR_Success) 
                goto out;

        st = NetUserGetInfo((LPCWSTR)dc, (LPCWSTR)w_user, 3, (LPBYTE *)&pinfo);
        if (st != NERR_Success) 
                goto out;

        if (pinfo) {
                if (pinfo->usri3_profile) {
                        DWORD w_len = lstrlenW(pinfo->usri3_profile);
                        if (w_len > 0) {
                                // get buffer size and allocate utf8 string buffer
                                int utf8_len = WideCharToMultiByte(CP_UTF8, 
                                                                   0, 
                                                                   pinfo->usri3_profile,
                                                                   w_len,
                                                                   NULL, 
                                                                   0,
                                                                   NULL,
                                                                   NULL);
                                *profile_path = (char*)calloc(utf8_len, sizeof(char));
                                if (!*profile_path)
                                        goto out;

                                // convert to ASCI
                                if (!WideCharToMultiByte(CP_UTF8, 
                                                         0,
                                                         pinfo->usri3_profile, 
                                                         w_len, 
                                                         *profile_path, 
                                                         utf8_len,
                                                         NULL, 
                                                         NULL)) {
                                        free(*profile_path);
                                        goto out;
                                }
                        }
                }
        }

out:
        free(w_domain);
        free(w_user);
        if (dc) 
                NetApiBufferFree(dc);
        if (pinfo)
                NetApiBufferFree(pinfo);

        return retv;
}


// MAPI_ALLOW_OTHERS is ignored due to security concerns
// MAPI_NEW_SESSION is forced due to security concerns
// All UI related flags are ignored
// MAPI_FORCE_DOWNLOAD is ignored. Unclear how it should be interpreted
static inline FLAGS native_logon_flags(const BRUTUS::BDEFINE Flags) 
{
        BRUTUS::BDEFINE flags = Flags;
        FLAGS retval = 0;

        retval |= MAPI_EXTENDED;
        retval |= MAPI_NEW_SESSION;

        if (flags & BRUTUS::BRUTUS_UNIQUE_SESSION) // ignore this flag here
                FLAGS_OFF(BRUTUS::BDEFINE, flags, BRUTUS::BRUTUS_UNIQUE_SESSION);

        if (flags & BRUTUS::BRUTUS_MAPI_NO_MAIL) {
                retval |= MAPI_NO_MAIL;
                FLAGS_OFF(BRUTUS::BDEFINE, flags, BRUTUS::BRUTUS_MAPI_NO_MAIL);
        }
        if (flags & BRUTUS::BRUTUS_MAPI_DEFAULT_SERVICES) {
                retval |= MAPI_DEFAULT_SERVICES;
                FLAGS_OFF(BRUTUS::BDEFINE, flags, BRUTUS::BRUTUS_MAPI_DEFAULT_SERVICES);
        }
        if (flags & BRUTUS::BRUTUS_MAPI_TIMEOUT_SHORT) {
                retval |= MAPI_TIMEOUT_SHORT;
                FLAGS_OFF(BRUTUS::BDEFINE, flags, BRUTUS::BRUTUS_MAPI_TIMEOUT_SHORT);
        }
        if (flags & BRUTUS::BRUTUS_MAPI_USE_DEFAULT) {
                retval |= MAPI_USE_DEFAULT;
                FLAGS_OFF(BRUTUS::BDEFINE, flags, BRUTUS::BRUTUS_MAPI_USE_DEFAULT);
        }
        if (flags & BRUTUS::BRUTUS_MAPI_EXPLICIT_PROFILE) {
                retval |= MAPI_EXPLICIT_PROFILE;
                FLAGS_OFF(BRUTUS::BDEFINE, flags, BRUTUS::BRUTUS_MAPI_EXPLICIT_PROFILE);
        }
        // Must(?) fix handling of Unicode characters
        if (flags & BRUTUS::BRUTUS_MAPI_UNICODE) {
                retval |= MAPI_UNICODE;
                FLAGS_OFF(BRUTUS::BDEFINE, flags, BRUTUS::BRUTUS_MAPI_UNICODE);
        }

        if (flags) {
                char msg[128] = {0};
                sprintf(msg, "Unknown or unsupported flag(s) from BRUTUS : %X", flags); 
                BRUTUS_LOG_BUG(msg);
        }

        return retval;
}

static BRUTUS::BRESULT create_session(const unsigned long LifeCheckDelay,
				      ::BRUTUS::BrutusCheck_ptr LifeLine,
				      const char *ProfileName,
                                      const char *ProfilePassword,
				      CORBA::ULong MaxNotifyDelay,
                                      ::BRUTUS::BDEFINE Flags,
                                      PortableServer::POA_ptr POA,
                                      ::BRUTUS::IMAPISession_out Session,
                                      HANDLE Stop)
{
        BRUTUS_LOG_REF;
        Session = BRUTUS::IMAPISession::_nil();

        LPMAPISESSION mapi_session = NULL;
        FLAGS flags = native_logon_flags(Flags);

        HRESULT hr = MAPILogonEx(0,
                                 (char*)ProfileName,
                                 (char*)ProfilePassword,
                                 flags,
                                 &mapi_session);

        BRUTUS::BRESULT br;
        if (!hresult_to_bresult(hr, br)) {
                if (mapi_session) {
                        BRUTUS_LOG_SUB_ERR("MAPI session is non-NIL but MAPILogonEx failed.");
                        try {
                                mapi_session->Release();
                        }
                        catch (...) {
                                BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                        }
                }
                return BRUTUS::BRUTUS_INTERNAL_ERROR;
        }
        if (S_OK != hr) {
                if (mapi_session) {
                        BRUTUS_LOG_SUB_ERR("MAPI session is non-NIL but MAPILogonEx failed.");
                        try {
                                mapi_session->Release();
                        }
                        catch (...) {
                                BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                        }
                }
                {
                        char msg[128] = {0};
                        const char *reason;

			switch (hr) {
			case MAPI_E_TIMEOUT :
                                reason = "MAPI_E_TIMEOUT";
				break;
                        case MAPI_E_NOT_FOUND :
				reason = "MAPI_E_NOT_FOUND";
				break;
			default :
                                reason = "MAPI_E_LOGON_FAILED";
			}
                        sprintf(msg, "MAPILogonEx(%s) failed - %s", ProfileName, reason); 
                        BRUTUS_LOG_SUB_PRIVATE(msg);
                }
                return br;
        }

        // create the child poa for this mapi session
        PortableServer::POA_var session_poa = PortableServer::POA::_duplicate(POA);
        if (CORBA::is_nil(session_poa)) {
                BRUTUS_LOG_SUB_ERR("Could not create session POA");
                try {
                        mapi_session->Release();
                }
                catch (...) {
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                }
                return BRUTUS::BRUTUS_INTERNAL_ERROR;
        }

        // create the IMAPISession servant
        BRUTUS_IMAPISession_i *session_servant = NULL;
        try {
                session_servant = new BRUTUS_IMAPISession_i(mapi_session, 
							    MaxNotifyDelay,
                                                            session_poa.in(), 
                                                            NULL, 
                                                            0,
                                                            Stop,
							    LifeLine,
							    LifeCheckDelay);
 	}
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_ERR("No memory");
                try {
                        mapi_session->Release();
                }
                catch (...) {
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                }
                return BRUTUS::BRUTUS_MAPI_E_NOT_ENOUGH_MEMORY;
        }
        if (!session_servant) {
                BRUTUS_LOG_SUB_ERR("No memory");
                try {
                        mapi_session->Release();
                }
                catch (...) {
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                }
                return BRUTUS::BRUTUS_MAPI_E_NOT_ENOUGH_MEMORY;
        }

        // create the IMAPISession object and register it in the session poa
        PortableServer::ObjectId_var oid;
        try {
                oid = session_poa->activate_object(session_servant);
        } 
        catch (const PortableServer::POA::WrongPolicy &) {
                BRUTUS_LOG_SUB_CRITICAL("PortableServer::POA::WrongPolicy exception caught");
                try {
                        delete session_servant;
                        mapi_session->Release();
                }
                catch (...) {
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                }
                return BRUTUS::BRUTUS_INTERNAL_ERROR;
        } 
        catch (const PortableServer::POA::ServantAlreadyActive &) {
                BRUTUS_LOG_SUB_CRITICAL("PortableServer::POA::ServantAlreadyActive exception caught");
                try {
                        delete session_servant;
                        mapi_session->Release();
                }
                catch (...) {
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                }
                return BRUTUS::BRUTUS_INTERNAL_ERROR;
        }

        CORBA::Object_var obj = CORBA::Object::_nil();
        try {
                obj = session_poa->id_to_reference(oid);
        }
        catch (const PortableServer::POA::WrongPolicy &) {
                BRUTUS_LOG_SUB_CRITICAL("PortableServer::POA::WrongPolicy exception caught");
        } 
        catch (const PortableServer::POA::ObjectNotActive &) {
                BRUTUS_LOG_SUB_CRITICAL("PortableServer::POA::ObjectNotActive exception caught");
        }
        if (CORBA::is_nil(obj)) {
                BRUTUS_LOG_SUB_CRITICAL("PortableServer::POA::id_to_reference() returned NIL");
                try {
                        mapi_session->Release();
                }
                catch (...) {
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                }
                return BRUTUS::BRUTUS_INTERNAL_ERROR;
        }
        BRUTUS::IMAPISession_var session = BRUTUS::IMAPISession::_narrow(obj);
        if (CORBA::is_nil(session)) {
                BRUTUS_LOG_SUB_CRITICAL("BRUTUS::IMAPISession::_narrow() returned NIL");
                try {
                        mapi_session->Release();
                }
                catch (...) {
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
                }
                return BRUTUS::BRUTUS_INTERNAL_ERROR;
        }

        Session = session._retn();
        {
                char msg[128] = {0};
                sprintf(msg, "%s has logged on", ProfileName); 
                BRUTUS_LOG_SUB_PRIVATE(msg);
        }

        BRUTUS_LOG_SUB_INF("Leaving create_session()");

        return BRUTUS::BRUTUS_S_OK;
}

// will realloc and strcat src to *dest, freeing *dest if 
// not successfull. Returns a pointer to the longer string.
static inline char *append_string(char **dest, const char *src)
{
        char *retv = NULL;
        size_t size;


        if (!*dest) {
                size = sizeof(char) * (strlen(src) + 1);
                retv = (char*)calloc(size, sizeof(char));
        } else {
                size = sizeof(char) * (strlen(*dest) + strlen(src) + 1);
                retv = (char*)realloc((void*)*dest, size);
        }
        if (!retv) {
                free(*dest);
                return NULL;
        }

        strcat(retv, src);
        *dest = retv;

        return *dest;
}

// Will realloc and append arg to argv and increment 
// argc, freeing *argv if not successfull.
// Returns a pointer to the new argv.
static char **build_argv(int *argc, char **argv[], char *arg)
{
        size_t size = sizeof(char*) * ++(*argc);

        char **retv = (char**)realloc((void**)*argv, size);
        if (!retv) {
                free(*argv);
                return NULL;
        }
        *argv = retv;
        (*argv)[*argc - 1] = arg;

        return *argv;
}

static void signal_handler(int sig)
{
	SetEvent(stop_);
}

int main(int argc, char **argv)
{
        BRUTUS_LOG_REF;

        bool do_stop = false; // there is a small potential deadlock window which is closed by this
        int retv = EXIT_FAILURE; 
        unsigned long n;
        
        // Windows Profile support
        HANDLE user_token = NULL;
        PROFILEINFO pinfo = { 0 };
        pinfo.dwSize = 0;

        // CORBA
        CORBA::ORB_var orb = CORBA::ORB::_nil();
        PortableServer::POA_var root_poa = PortableServer::POA::_nil();
        PortableServer::POAManager_var root_poa_manager = PortableServer::POAManager::_nil();
        PortableServer::POA_var session_poa = PortableServer::POA::_nil();
        PortableServer::POAManager_var session_poa_manager = PortableServer::POAManager::_nil();
        PortableServer::ServantManager_var session_activator = PortableServer::ServantManager::_nil();
        ::BRUTUS::IMAPISession_var session = ::BRUTUS::IMAPISession::_nil();
        ::BRUTUS::ProxyManager_var proxy_manager = ::BRUTUS::ProxyManager::_nil();
	::BRUTUS::BrutusCheck_var life_line = ::BRUTUS::BrutusCheck::_nil();
        ::BRUTUS::ProxyManager::SpawnStatus st = ::BRUTUS::ProxyManager::spawn_unknown;
        ::BRUTUS::BRESULT br = ::BRUTUS::BRUTUS_UNKNOWN_ERROR;

        // MAPI
        HRESULT hr = E_FAIL;
        LPMAPISESSION mapi_session = NULL;
        bool mapi_init_ok = false;
        bool profiles_on_disk = false;
        CORBA::ULong max_notify_delay = 0;
        CORBA::ULong ulFlags = 0;
        CORBA::String_var life_line_ior = "";
        CORBA::String_var proxy_profile = "";
        CORBA::String_var mapi_profile_password = "";
	unsigned long life_check_delay = 0;

        // signal handling
	signal(SIGABRT, signal_handler);
	signal(SIGFPE, signal_handler);
	signal(SIGILL, signal_handler);
	signal(SIGINT, signal_handler);
	signal(SIGSEGV, signal_handler);
	signal(SIGTERM , signal_handler);

        // disable error dialogs 
        SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOOPENFILEERRORBOX);

	{ // Change current working directory
		char *c = NULL;
		char wdir[_MAX_PATH] = { '\0' };

		memcpy(wdir, argv[0], strlen(argv[0]));
		c = &wdir[strlen(argv[0])-1];
		while ('\\' != *c)
			*(c--) = '\0';
		*c = '\0';
		SetCurrentDirectory(wdir);
	}

        { // constrain configuration resources to this scope
                Config *conf = NULL;
                try {
                        conf = new Config(INTERNAL_ORB_STACK); // stack name
                }
                catch (std::bad_alloc &) {
                        BRUTUS_LOG_SUB_CRITICAL("No memory");
                        st = ::BRUTUS::ProxyManager::spawn_error_config;
                        goto spawn_exit;
                }
                if (!conf) {
                        BRUTUS_LOG_SUB_CRITICAL("No memory");
                        st = ::BRUTUS::ProxyManager::spawn_error_config;
                        goto spawn_exit;
                }

                // load configuration
                if (conf->load(BRUTUS_CONF)) {
                        st = ::BRUTUS::ProxyManager::spawn_error_config;
			delete conf;
                        goto spawn_exit;
                }

                // MAPI options
                char *profiles_on_disk_str = NULL;
                profiles_on_disk_str = conf->get_value("MAPI_PROFILES_ON_DISK");
                if ((profiles_on_disk_str) && (!strcmp(profiles_on_disk_str, "yes"))) 
                        profiles_on_disk = true;
                free(profiles_on_disk_str);

		// Execution options
                char *life_check_delay_str = NULL;
                life_check_delay_str = conf->get_value("CLIENT_LIFE_CHECK_DELAY");
                if (life_check_delay_str)
                        life_check_delay = strtoul(life_check_delay_str, NULL, 10);
		else
                        life_check_delay = 0;
                free(life_check_delay_str);

		// ORB init command line
		int orb_argc = 0;
		char **orb_argv = NULL;
		std::vector<char*> cmd_line;
		cmd_line.resize(0);

                // ORB options
                char *orb_cmd_line_option = NULL;
                build_argv(&orb_argc, &orb_argv, PROXY_EXE); // add prog name
                orb_cmd_line_option = conf->get_value(INTERNAL_ORB_STACK);
                while (orb_cmd_line_option) {
                        cmd_line.push_back(orb_cmd_line_option);
                        build_argv(&orb_argc, &orb_argv, orb_cmd_line_option);
                        orb_cmd_line_option = conf->get_value(INTERNAL_ORB_STACK);
                }
                delete conf;

                try {
                        orb = CORBA::ORB_init(orb_argc, orb_argv);
                }
                catch (...) {
			cmd_line.clear();
			free(orb_argv);
			orb_argv = NULL;
			orb_argc = 0;
                        goto exit;
                }

		// free ORB command line options
		for (n = 0; n < cmd_line.size(); n++) {
			free(cmd_line[n]);
		}
		cmd_line.clear();
		free(orb_argv);
		orb_argv = NULL;

                if (CORBA::is_nil(orb.in())) 
                        goto exit;
        } 
       
        { // get ProxyManager user data
                try {
                        CORBA::Object_var object = orb->string_to_object(PROXY_MANAGER_IOR);
                        proxy_manager = BRUTUS::ProxyManager::_narrow(object.in());
                        if (CORBA::is_nil(proxy_manager.in())) 
                                goto exit;
                }
                catch (...) {
                        goto exit;
                }

                CORBA::String_var user_name = "";
                CORBA::String_var domain = "";
                CORBA::String_var password = "";
                try {
                        proxy_manager->getUser(life_line_ior.out(),
					       proxy_profile.out(), 
                                               user_name.out(),
                                               domain.out(), 
                                               password.out(), 
                                               mapi_profile_password.out(), 
					       max_notify_delay,
                                               ulFlags);
                }
                catch (...) {
                        st = ::BRUTUS::ProxyManager::spawn_error_corba;
                        goto spawn_exit;
                }
                proxy_manager = ::BRUTUS::ProxyManager::_nil();

                // MAPI gets profile information from HKEY_CURRENT_USER but the registry
                // profile for the impersonated user is not automatically loaded. We must
                // therefore take steps to ensure that any existing or roaming profile is
                // properly loaded before proceeeding unless we get the profile from disk.
                if (!profiles_on_disk) {
                        char *profile_path = NULL;
                        char *dc = NULL;

                        if (!get_domain_controller((const char*)domain.in(), &dc)) {
                                st = ::BRUTUS::ProxyManager::spawn_error_impersonation;
                                goto spawn_exit;        
                        }

                        if (!get_profile_path((const char*)user_name.in(), (const char*)domain.in(), &profile_path)) {
                                free(dc);
                                st = ::BRUTUS::ProxyManager::spawn_error_impersonation;
                                goto spawn_exit;        
                        }

                        pinfo.dwSize = sizeof(PROFILEINFO);
                        pinfo.dwFlags = PI_NOUI;
                        pinfo.lpUserName = strdup((const char*)user_name.in());
                        pinfo.lpProfilePath = profile_path;
                        pinfo.lpDefaultPath = NULL;
                        pinfo.lpServerName = strdup(dc);;
                        pinfo.lpPolicyPath = NULL;
                        free(dc);
                        free(profile_path);

			// get user token to load profile
                        if (!LogonUser((char*)((const char*)user_name.in()), 
                                       (char*)((const char*)domain.in()),
                                       (char*)((const char*)password.in()),
                                       LOGON32_LOGON_INTERACTIVE,
                                       LOGON32_PROVIDER_DEFAULT,
                                       &user_token)) {
                                st = ::BRUTUS::ProxyManager::spawn_error_authentication;
                                goto spawn_exit;
                        }
                        if (!LoadUserProfile(user_token, &pinfo)) {
                                st = ::BRUTUS::ProxyManager::spawn_error_impersonation;
                                goto spawn_exit;        
                        }                        
                }
        }

        ///////////////////////////////////////////////////
        //       start remaining CORBA sub-systems       //
        ///////////////////////////////////////////////////

        stop_ = CreateEvent(NULL, true, false, NULL);
        if (!stop_) {
                st = ::BRUTUS::ProxyManager::spawn_error_spawn;
                goto spawn_exit;
        }

        // root poa
        root_poa = resolve_init<PortableServer::POA>(orb.in(), "RootPOA");
        if (CORBA::is_nil(root_poa.in())) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create root session POA");
                st = ::BRUTUS::ProxyManager::spawn_error_corba;
                goto spawn_exit;
        }

        // root poa manager
        root_poa_manager = root_poa->the_POAManager();
        if (CORBA::is_nil(root_poa_manager.in())) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create session POAManager");
                st = ::BRUTUS::ProxyManager::spawn_error_corba;
                goto spawn_exit;
        }
        root_poa_manager->activate();

        // create session POA 
        session_poa = create_poa((const char*)"", PortableServer::POAManager::_nil(), root_poa.in());
        if (CORBA::is_nil(session_poa.in())) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create session POA");
                st = ::BRUTUS::ProxyManager::spawn_error_corba;
                goto spawn_exit;
        }

        // session poa manager
        session_poa_manager = session_poa->the_POAManager();
        if (CORBA::is_nil(session_poa_manager.in())) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create session POAManager");
                st = ::BRUTUS::ProxyManager::spawn_error_corba;
                goto spawn_exit;
        }
        session_poa_manager->activate();

        // create and register the session servant activator.
        try {
                session_activator = new BrutusActivator_i();
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_CRITICAL("No memory");
                st = ::BRUTUS::ProxyManager::spawn_error_exception;
                goto spawn_exit;
        }
        try {
                session_poa->set_servant_manager(session_activator.in());
        }
        catch (...) {
                BRUTUS_LOG_SUB_CRITICAL("Exception caught when setting Servant Activator");
                st = ::BRUTUS::ProxyManager::spawn_error_exception;
                goto spawn_exit;
        }

        // re-get the proxy manager reference
        try {
                CORBA::Object_var object = orb->string_to_object(PROXY_MANAGER_IOR);
                proxy_manager = BRUTUS::ProxyManager::_narrow(object.in());

                if (CORBA::is_nil(proxy_manager.in()))
                        goto exit;
        }
        catch (...) {
                goto exit;
        }

        //////////////////////////////////////
        //               MAPI               //
        //////////////////////////////////////

        // MAPI initialization
        MAPIINIT_0 mapi_init;
        mapi_init.ulFlags = MAPI_MULTITHREAD_NOTIFICATIONS | (profiles_on_disk ? MAPI_TEMPORARY_PROFILES : 0);
        mapi_init.ulVersion = MAPI_INIT_VERSION;
        hr = MAPIInitialize(&mapi_init);
        if (hr != S_OK) {
                BRUTUS_LOG_SUB_CRITICAL("Could not initialize MAPI");
                st = ::BRUTUS::ProxyManager::spawn_error_mapi_init;
                goto spawn_exit;
        } else 
                mapi_init_ok = true;

	// get the lifeline
	try {
		CORBA::Object_var object = orb->string_to_object(life_line_ior.in());
		life_line = BRUTUS::BrutusCheck::_narrow(object.in());
		if (CORBA::is_nil(life_line.in())) {
			st = BRUTUS::ProxyManager::spawn_error_spawn;
			goto spawn_exit;
		}
	}
	catch (...) {
		st = BRUTUS::ProxyManager::spawn_error_exception;
		goto spawn_exit;
	}

        // MAPI logon
        br = create_session(life_check_delay, 
			    life_line._retn(), 
			    proxy_profile.in(), 
			    mapi_profile_password.in(), 
			    max_notify_delay, 
			    ulFlags, 
			    session_poa.in(), 
			    session.out(), 
			    stop_);
        if (BRUTUS::BRUTUS_S_OK == br) {
                BRUTUS_LOG_SUB_INF("MAPI Session created");
		CORBA::String_var ior_str = (const char*)"";
		try {
			ior_str = orb->object_to_string(session.in());
                }
		catch (const CORBA::Exception &e) {
			st = BRUTUS::ProxyManager::spawn_error_exception;
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
			e._tao_print_exception("CORBA exception caught\n");
			goto spawn_exit;
		}
		catch (...) {
			st = BRUTUS::ProxyManager::spawn_error_exception;
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
			goto spawn_exit;
		}

		try {
			proxy_manager->setSession(proxy_profile.in(), 
						  ior_str.in(), 
						  (ulFlags | ::BRUTUS::BRUTUS_UNIQUE_SESSION) ? (::CORBA::Boolean)0 : (::CORBA::Boolean)1); // last call
                }
		catch (const CORBA::Exception &e) {
			st = BRUTUS::ProxyManager::spawn_error_exception;
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
			e._tao_print_exception("CORBA exception caught\n");
			goto spawn_exit;
		}
		catch (...) {
			st = BRUTUS::ProxyManager::spawn_error_exception;
                        BRUTUS_LOG_SUB_CRITICAL("Exception caught");
			goto spawn_exit;
		}
        } else {
                BRUTUS_LOG_SUB_CRITICAL("Could not create MAPI Session - No profile or wrong user credentials");
		if (::BRUTUS::BRUTUS_MAPI_E_NOT_FOUND)
			st = BRUTUS::ProxyManager::spawn_error_mapi_no_profile;
		else
			st = BRUTUS::ProxyManager::spawn_error_mapi_logon;

                goto spawn_exit;
        }                                
        proxy_manager = ::BRUTUS::ProxyManager::_nil();

        ////////////////////////////////////////////
        //             Activate proxy             //
        ////////////////////////////////////////////

        // stop_ is set to true when BRUTUS_IMAPISession_i::LogOff() is called or 
        // when any signal is recieved, in which case ~BRUTUS_IMAPISession_i() is
        // invoked when the object is destoyed when the POA is destroyed.
        BRUTUS_LOG_SUB_INF("Starting ORB event loop");
	while (true) {
                if (orb->work_pending()) 
                        orb->perform_work();

                if (do_stop)
                        break;

                if (WAIT_OBJECT_0 == WaitForSingleObject(stop_, 0)) {
                        do_stop = true;
                        break;
                }
        }
        retv = EXIT_SUCCESS;

        goto exit;

spawn_exit:
        try {
                if (CORBA::is_nil(proxy_manager.in())) {
                        if (CORBA::is_nil(orb.in()))
                                orb = CORBA::ORB_init(argc, argv, "BRUTUS_PROXY_ERROR_ORB");
        
                        CORBA::Object_var object = orb->string_to_object(PROXY_MANAGER_IOR);
                        proxy_manager = BRUTUS::ProxyManager::_narrow(object.in());
                        if (CORBA::is_nil(proxy_manager.in())) 
                                goto exit;
                }
                proxy_manager->setSpawnStatus(proxy_profile.in(), st); // last call
        }
        catch (...) {
                BRUTUS_LOG_SUB_CRITICAL("Caught exception while setting spawn status");
        }

exit:
        try {
                if (!CORBA::is_nil(session_poa_manager.in())) 
                        session_poa_manager->deactivate(1, 1);

                if (!CORBA::is_nil(session_poa.in())) 
                        session_poa->destroy(1, 1);

                if (!CORBA::is_nil(root_poa_manager.in())) 
                        root_poa_manager->deactivate(1, 1);

                if (!CORBA::is_nil(root_poa.in())) 
                        root_poa->destroy(1, 1);
                
                orb->shutdown(1);
                orb->destroy();
        }
        catch (...) {
        }

        // close MAPI
	try {
		if (mapi_init_ok)
			MAPIUninitialize();
	}
        catch (...) {
        }

        // only used if not using profiles on disk 
        if (pinfo.dwSize) {
                free(pinfo.lpUserName);
                free(pinfo.lpServerName);
                free(pinfo.lpProfilePath);
                UnloadUserProfile(user_token, pinfo.hProfile);
        }

        // close security token
        if (user_token)
                CloseHandle(user_token);

        if (stop_)
                CloseHandle(stop_);

        ExitProcess(retv); 
}
