/*
 *    Brutus source file implementing the main() function.
 *    Copyright (C) 2004 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


//**************************************************************************
//
// BASED ON ACE SAMPLE CODE BY:
//    Gonzalo Diethelm <gonzo@cs.wustl.edu> 
//    and Steve Huston <shuston@riverace.com>
//
//**************************************************************************

#include <NT_Service/main.h>

#include <vector>
#include <fstream>
#include <cstdlib>
#include <errno.h>
#include <atlbase.h>

#include <ace/SSL/SSL_Context.h>
#include <ace/Get_Opt.h>

#include <NT_Service/ntsvc.h>
#include <NT_Service/macros.h>
#include <NT_Service/config.h>
#include <NT_Service/defines.h>
#include <NT_Service/rsa_utils.h>

#include <TAO/tao/ORBInitializer_Registry.h>
#include <NT_Service/orbinitializer.h>


// The global orb instances
CORBA::ORB_var global_server_orb = CORBA::ORB::_nil();
CORBA::ORB_var global_internal_orb = CORBA::ORB::_nil();

const int password_length = 128;
char ssl_private_key_password[password_length] = {0};

// SSL callback
static int pem_passwd_cb(char *buf, int size, int rwflag, void *userdata)
{
        strncpy(buf, ssl_private_key_password, size);
        buf[size - 1] = '\0';

        return(strlen(buf));
}

static inline void sec_free_memory(void *buf, size_t size)
{
        // scramble buffer
        memset(buf, 0xFF, size);
        if (buf)
                free(buf);
}

static inline void scramble_password(void)
{
        // scramble password buffer
        memset((void*)ssl_private_key_password, 0xFF, password_length);
}

// Will check for and if possible enable required 
// privileges.
// Returns false on error, true otherwise.
static bool SetPrivilege(LPCTSTR lpszPrivilege,  // name of privilege to enable/disable
                         bool bEnablePrivilege)  // to enable or disable privilege
{
        TOKEN_PRIVILEGES tp;
        LUID luid;

        if (!LookupPrivilegeValue(NULL,            // lookup privilege on local system
                                  lpszPrivilege,   // privilege to lookup 
                                  &luid)) {        // receives LUID of privilege
                BRUTUS_SYSLOG_GETLASTERROR("LookupPrivilegeValue error");
                return false; 
        }

        tp.PrivilegeCount = 1;
        tp.Privileges[0].Luid = luid;
        if (bEnablePrivilege)
                tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
        else
                tp.Privileges[0].Attributes = 0;

        // Enable the privilege or disable all privileges.
        HANDLE this_process = NULL;
        if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &this_process)) {
                BRUTUS_SYSLOG_GETLASTERROR("OpenProcessToken error");
                return false; 
        }
        if (!AdjustTokenPrivileges(this_process, 
                                   FALSE, 
                                   &tp, 
                                   sizeof(TOKEN_PRIVILEGES), 
                                   (PTOKEN_PRIVILEGES)NULL, 
                                   (PDWORD)NULL)) { 
                CloseHandle(this_process);
                BRUTUS_SYSLOG_GETLASTERROR("AdjustTokenPrivileges failed");
                return false; 
        } 
        CloseHandle(this_process);

        if (GetLastError() == ERROR_NOT_ALL_ASSIGNED) {
		ACE_LOG_MSG->open("Brutus Server", ACE_Log_Msg::SYSLOG);
		ACE_DEBUG((LM_CRITICAL, ACE_TEXT("The token does not have the specified privilege")));
                return false;
        } 

        return true;
}

// will realloc and strcat src to *dest, freeing *dest if 
// not successfull. Returns a pointer to the longer string.
static inline char *append_string(char **dest, const char *src)
{
        char *retv;
        size_t size;


        if (!*dest) {
                size = sizeof(char) * (strlen(src) + 1);
                retv = (char*)calloc(size, sizeof(char));
        } else {
                size = sizeof(char) * (strlen(*dest) + strlen(src) + 1);
                retv = (char*)realloc((void*)*dest, size);
        }
        if (!retv) {
                free(*dest);
                return NULL;
        }

        strcat(retv, src);
        *dest = retv;

        return *dest;
}

// Will realloc and append arg to argv and increment 
// argc, freeing *argv if not successfull.
// Returns a pointer to the new argv.
static char **build_argv(int *argc, char **argv[], char *arg)
{
        size_t size = sizeof(char*) * ++(*argc);

        char **retv = (char**)realloc((void**)*argv, size);
        if (!retv) {
                free(*argv);
                return NULL;
        }
        *argv = retv;
        (*argv)[*argc - 1] = arg;

        return *argv;
}


// Default for the -i (install) option
#define DEFAULT_SERVICE_INIT_STARTUP SERVICE_AUTO_START

class Process
{
public:
        Process(void);
        ~Process(void);

        int run(int argc, 
                ACE_TCHAR *argv[], 
                const bool ProfilesOnDisk,
                const bool ProcessPerSession,
                const bool OnlyAdministrator,
                const bool Authenticate,
		const unsigned long LifeCheckDelay);

private:
        void parse_args(int argc, ACE_TCHAR *argv[]);
        void print_usage_and_die(void);

private:
        char progname[128];

        int opt_install;
        int opt_remove;
        int opt_start;
        int opt_kill;
        int opt_type;
        int opt_debug;

        int opt_startup;
};

typedef ACE_Singleton<Process, ACE_Mutex> PROCESS;

Process::Process(void) 
        : opt_install (0),
          opt_remove (0),
          opt_start (0),
          opt_kill (0),
          opt_type (0),
          opt_debug (0),
          opt_startup (0)
{
        ACE_OS::strcpy(progname, "brutus");
        ACE::init();
}

Process::~Process(void)
{
        ACE::fini();
}

void Process::print_usage_and_die(void)
{
        ACE_DEBUG ((LM_INFO,
                    "Usage: %s"
                    " -in -r -s -k -tn -d\n"
                    "  -i: Install Brutus Server as an NT service, with specified startup\n"
                    "  -r: Remove Brutus Server from the Service Manager\n"
                    "  -s: Start Brutus Server\n"
                    "  -k: Kill Brutus Server\n"
                    "  -t: Set startup for an installed Brutus Server\n"
                    "  -d: Debug; run Brutus Server as a regular application\n",
                    progname,
                    0));
        ACE_OS::exit(1);
}

void Process::parse_args(int argc, ACE_TCHAR *argv[])
{
        ACE_Get_Opt get_opt(argc, argv, "i:rskt:d");
        int c; 

        while (-1 != (c = get_opt ()))
                switch (c) {
                case 'i':
                        opt_install = 1;
                        opt_startup = ACE_OS::atoi(get_opt.optarg);
                        if (opt_startup <= 0) {
                                print_usage_and_die();
                        }
                        break;
                case 'r':
                        opt_remove = 1;
                        break;
                case 's':
                        opt_start = 1;
                        break;
                case 'k':
                        opt_kill = 1;
                        break;
                case 't':
                        opt_type = 1;
                        opt_startup = ACE_OS::atoi(get_opt.optarg);
                        if (opt_startup <= 0) {
                                print_usage_and_die();
                        }
                        break;
                case 'd':
                        opt_debug = 1;
                        break;
                default:
                        // -i can also be given without a value - if so, it defaults
                        // to defined value.
                        if (!ACE_OS::strcmp(get_opt.argv_[get_opt.optind-1], "-i")) {
                                opt_install = 1;
                                opt_startup = DEFAULT_SERVICE_INIT_STARTUP;
                        } else {
                                print_usage_and_die();
                        }
                        break;
                }
}

// Define a function to handle Ctrl+C to cleanly shut this down.
BOOL WINAPI ConsoleHandler(DWORD ctrlType)
{
        ACE_UNUSED_ARG (ctrlType);
        SERVICE::instance()->handle_control(SERVICE_CONTROL_STOP);
        return TRUE;
}

ACE_NT_SERVICE_DEFINE(Brutus,
                      Service,
                      "Brutus Server");

int Process::run(int argc, 
                 ACE_TCHAR *argv[], 
                 const bool ProfilesOnDisk,
                 const bool ProcessPerSession,
                 const bool OnlyAdministrator,
                 const bool Authenticate,
		 const unsigned long LifeCheckDelay)
{
        SERVICE::instance()->name("Brutus",
                                  "Brutus Server");

        SERVICE::instance()->SetOptions((bool)opt_debug,
                                        ProfilesOnDisk, 
                                        ProcessPerSession, 
                                        OnlyAdministrator,
                                        Authenticate,
					LifeCheckDelay);

        parse_args(argc, argv);

        if (opt_install && !opt_remove) return SERVICE::instance()->insert(opt_startup);

        if (opt_remove && !opt_install) return SERVICE::instance()->remove();

        if (opt_start && opt_kill) print_usage_and_die();

        if (opt_start) return SERVICE::instance()->start_svc();

        if (opt_kill) return SERVICE::instance()->stop_svc();

        if (opt_type) return SERVICE::instance()->startup(opt_startup);

        // If we get here, we either run the app in debug mode (-d) or are
        // being called from the service manager to start the service.

        if (opt_debug) {
                SetConsoleCtrlHandler(&ConsoleHandler, true);
                SERVICE::instance()->svc();
        } else {
                ACE_NT_SERVICE_RUN(Brutus,
                                   SERVICE::instance(),
                                   ret);
                if (!ret) {
                        ACE_ERROR ((LM_ERROR,
                                    "%p\n",
                                    "Couldn't start service"));
                }
        }
  
        return 0;
}

static inline bool get_xbit_number(const unsigned char bits, const char *num_str, unsigned long *ul)
{
        unsigned long max;

        switch (bits) {
        case 8 :
                max = 0xff;
                break;
        case 16 :
                max = 0xffff;
                break;
        case 32 :
                max = 0xffffffff;
                break;
        default :
                return false;
        }

        if (!num_str)
                return false;

        if (*num_str == '-')
                return false;

        char *tmp;
        unsigned long num = strtoul(num_str, &tmp, 10);
        if (errno == ERANGE)
                return false;

        if (*tmp != '\0')
                return false;

        if (num > max)
                return false;

        *ul = (unsigned long)num;

        return true;
}

int ACE_TMAIN (int argc, ACE_TCHAR* argv[])
{
	ACE_LOG_MSG->open("Brutus Server", ACE_Log_Msg::SYSLOG);
        int retv = -1;

        // ssl configuration
        char *use_ssl_str = NULL;
        bool use_ssl = false;
        char *ssl_certificate = NULL;
        char *ssl_certificate_format = NULL;
        char *ssl_private_key = NULL;
        char *ssl_private_key_password_str = NULL;
        char *ssl_cert_file = NULL;
        char *ssl_cert_file_env = NULL;
        char *ssl_cert_dir = NULL;
        char *ssl_cert_dir_env = NULL;
        SSL_CTX *ctx;

        // iiop configuration
        char *use_iiop_str = NULL;
        bool use_iiop = false;

        // ORB options
        int orb_argc = 0;
        char **orb_argv = NULL;
        char *orb_cmd_line_option = NULL;
        int internal_orb_argc = 0;
        char **internal_orb_argv = NULL;
        char *internal_orb_cmd_line_option = NULL;
        std::vector<char*> cmd_line;
        cmd_line.resize(0);

        // Interceptor
        ORBInitializer *initializer = NULL;
        PortableInterceptor::ORBInitializer_var orb_initializer = PortableInterceptor::ORBInitializer::_nil();

        // MAPI options
        bool profiles_on_disk = true;
        bool process_per_session = true;
        bool only_administrator = false;
        bool authenticate = true;

        // Cluster?
        bool forward = false;

	// check health status of client applications?
	unsigned long life_check_delay = 0;

        // configuration helper class
        Config *conf = NULL;

	{ // Change current working directory
		char *c = NULL;
		char wdir[_MAX_PATH] = { '\0' };

		memcpy(wdir, argv[0], strlen(argv[0]));
		c = &wdir[strlen(argv[0])-1];
		while ('\\' != *c)
			*(c--) = '\0';
		*c = '\0';
		SetCurrentDirectory(wdir);
	}

        // create svc.conf
        std::ofstream svc_conf("svc.conf", ifstream::trunc|ifstream::out);
        if (!svc_conf) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Configuration error: Could not create TAO Service Configuration file")));
                goto exit;
        }

        // create configuration class
        try {
                conf = new Config(SERVER_ORB_STACK); // stack name
        }
        catch (std::bad_alloc &) {
                goto no_mem;
        }
        if (!conf) 
                goto no_mem;

        // create public and private RSA key files in PEM format
        if (!create_rsa_files(PUBLIC_KEY, PRIVATE_KEY)) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Initialization error: Could not create RSA keys")));
                goto exit;
        }

        // TAO tuning options
        svc_conf << "static Server_Strategy_Factory \"-ORBConcurrency reactive -ORBPOALock thread\"\n";

        // load configuration
        retv = conf->load(BRUTUS_CONF);
        if (retv) {
                char msg[128] = {'\0'};
                if (retv > 0)
                        sprintf(msg, "Configuration error: Error in config file at line %d", retv);
                else
                        strcat(msg, "Configuration error: Could not open brutus.conf");

                ACE_DEBUG((LM_CRITICAL, ACE_TEXT(msg)));
                retv = -1;
                goto exit;
        }

	{ // Check that we have been configured
                char *configured_str = conf->get_value("CONFIGURED");
                if (!configured_str || strcmp(configured_str, "yes")) {
			int n = 0;
			bool debug_run = false;

			for (n = 0; n < argc; n++) {
				if (!strcmp("-d", argv[n])) {
					debug_run = true;
					break;
				}
			}

			// ignore CONFIGURED setting if "-d" is parsed on the cmd line
			if (!debug_run) {
				free(configured_str);
				ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Brutus Server is not configured. Please edit brutus.conf by hand or run Brutus Server Setup.")));
				retv = -1;
				goto exit;
			}
		}
		free(configured_str);
	}
                
        {
                char *forward_str = conf->get_value("FORWARD");
                if ((forward_str) && (!strcmp(forward_str, "yes"))) 
                        forward = true;
                free(forward_str);
        }

        {
                char *profiles_on_disk_str = conf->get_value("MAPI_PROFILES_ON_DISK");
                if ((profiles_on_disk_str) && (!strcmp(profiles_on_disk_str, "yes"))) 
                        profiles_on_disk = true;
                free(profiles_on_disk_str);
        }

	{
                char *life_check_delay_str = NULL;
                life_check_delay_str = conf->get_value("CLIENT_LIFE_CHECK_DELAY");
                if (life_check_delay_str)
                        life_check_delay = strtoul(life_check_delay_str, NULL, 10);
		else
                        life_check_delay = 0;
                free(life_check_delay_str);
	}

        // set .mmp path in the registry
        if (profiles_on_disk) {
                char *mmp_dir = conf->get_value("MMP_DIRECTORY");
                if (mmp_dir) {
                        if (!CreateDirectory(mmp_dir, NULL)) {
                                DWORD err = GetLastError();
                                if (ERROR_ALREADY_EXISTS != err) { 
                                        ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create MMP directory")));
                                        free(mmp_dir);
                                        goto exit;
                                }
                        }
                        
                        DWORD disposition = 0;
                        CRegKey *reg = new CRegKey();
                        if (!reg) {
                                free(mmp_dir);
                                goto no_mem;
                        }

                        long reg_retv = reg->Create(HKEY_LOCAL_MACHINE,
                                                    "Software\\Microsoft\\Windows Messaging Subsystem\\",
                                                    REG_NONE,
                                                    REG_OPTION_NON_VOLATILE,
                                                    KEY_WRITE | KEY_READ,
                                                    NULL,
                                                    &disposition);
                        switch (disposition) {
                        case REG_CREATED_NEW_KEY :
                                break;
                        case REG_OPENED_EXISTING_KEY :
                                break;
                        default :
                                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Unknown disposition for HKEY_LOCAL_MACHINE/Software/Microsoft/Windows Messaging Subsystem")));
                        }
                        if (reg_retv != ERROR_SUCCESS) {
                                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not open HKEY_LOCAL_MACHINE/Software/Microsoft/Windows Messaging Subsystem")));
                                free(mmp_dir);
                                delete reg;
                                goto exit;
                        }

                        reg_retv = reg->SetStringValue("ProfileDirectory",
						       mmp_dir,
						       REG_SZ);
                        free(mmp_dir);
                        if (reg_retv != ERROR_SUCCESS) {
                                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not set ProfileDirectory")));
                                delete reg;
                                goto exit;
                        }

                        delete reg;
                }
        }

        // protocol selections
        use_ssl_str = conf->get_value("USE_SSL");
        if ((use_ssl_str) && (!strcmp(use_ssl_str, "yes"))) 
                use_ssl = true;
        free(use_ssl_str);

        use_iiop_str = conf->get_value("USE_IIOP");
        if ((use_iiop_str) && (!strcmp(use_iiop_str, "yes"))) 
                use_iiop = true;
        free(use_iiop_str);

        if ((!use_iiop) && (!use_ssl)) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Configuration error: You must use at least one of SSL or IIOP")));
                goto exit;
        }

        // ssl configuration
        if (use_ssl) {
                ssl_certificate = conf->get_value("SSL_CERT");
                if (!ssl_certificate) {
                        ACE_DEBUG((LM_CRITICAL, ACE_TEXT("SSL_CERT: No SSL certificate")));
                        goto exit;
                }

                ssl_certificate_format = conf->get_value("SSL_CERT_FORMAT");
                if (!ssl_certificate_format) {
                        ACE_DEBUG((LM_CRITICAL, ACE_TEXT("SSL_CERT_FORMAT: No SSL certificate format")));
                        goto exit;
                }

                ssl_private_key = conf->get_value("SSL_PRIVATE_KEY");
                if (!ssl_private_key) {
                        ACE_DEBUG((LM_CRITICAL, ACE_TEXT("SSL_PRIVATE_KEY: No SSL private key")));
                        goto exit;
                }

                // load password and setup a callback function to answer password queries
                ssl_private_key_password_str = conf->get_value("SSL_PRIVATE_KEY_PASSWORD");
                if (!ssl_private_key_password_str) {
                        ACE_DEBUG((LM_CRITICAL, ACE_TEXT("SSL_PRIVATE_KEY_PASSWORD_STR: No SSL private key password")));
                        goto exit;
                } else {
                        strncpy(ssl_private_key_password, ssl_private_key_password_str, password_length);
                        sec_free_memory(ssl_private_key_password_str, password_length);

                        ctx = ACE_SSL_Context::instance()->context();
                        SSL_CTX_set_default_passwd_cb(ctx, pem_passwd_cb);
                }

                char *ssliop_options = "dynamic SSLIOP_Factory Service_Object * TAO_SSLIOP:_make_TAO_SSLIOP_Protocol_Factory() \"";;
                if (use_iiop) {
                        if(!append_string(&ssliop_options, "-SSLNoProtection ")) 
                                goto no_mem;
                }

                // SSL certificate
                if (!append_string(&ssliop_options, "-SSLCertificate ")) 
                        goto no_mem;
                if (!append_string(&ssliop_options, ssl_certificate_format))
                        goto no_mem;
                if (!append_string(&ssliop_options, ":")) 
                        goto no_mem;
                if (!append_string(&ssliop_options, ssl_certificate))
                        goto no_mem;
                if (!append_string(&ssliop_options, " ")) 
                        goto no_mem;
                
                // SSL private key
                if (!append_string(&ssliop_options, "-SSLPrivateKey "))
                        goto no_mem;
                if (!append_string(&ssliop_options, ssl_certificate_format))
                        goto no_mem;
                if (!append_string(&ssliop_options, ":"))
                        goto no_mem;
                if (!append_string(&ssliop_options, ssl_private_key))
                        goto no_mem;
                if (!append_string(&ssliop_options, " "))
                        goto no_mem;

                // SSL authenticate - LAST OPTION!!!
                if (!append_string(&ssliop_options, "--SSLAuthenticate SERVER"))
                        goto no_mem;
                // close option string
                if (!append_string(&ssliop_options, "\"\n"))
                        goto no_mem;

                // write to service configuration file
                svc_conf << ssliop_options;
                svc_conf << "static Resource_Factory \"-ORBProtocolFactory SSLIOP_Factory\"\n";
                free(ssliop_options);

                // create SSL_CERT_FILE environment variable
                ssl_cert_file = conf->get_value("SSL_CERT_FILE");
                if (ssl_cert_file) {
                        ssl_cert_file_env = (char*)calloc((strlen("SSL_CERT_FILE=") + strlen(ssl_cert_file) + 1), sizeof(char));
                        if (!ssl_cert_file_env)
                                goto no_mem;

                        strcpy(ssl_cert_file_env, "SSL_CERT_FILE=");
                        strcat(ssl_cert_file_env, ssl_cert_file);
                        _putenv(ssl_cert_file_env);
                }

                // create SSL_CERT_DIR environment variable
                ssl_cert_dir = conf->get_value("SSL_CERT_DIR");
                if (ssl_cert_dir) {
                        ssl_cert_dir_env = (char*)calloc((strlen("SSL_CERT_DIR=") + strlen(ssl_cert_dir) + 1), sizeof(char));
                        if (!ssl_cert_dir_env)
                                goto no_mem;

                        strcpy(ssl_cert_dir_env, "SSL_CERT_DIR=");
                        strcat(ssl_cert_dir_env, ssl_cert_dir);
                        _putenv(ssl_cert_dir_env);
                }
        }
        svc_conf.close();

        // Generic options
        {
                char *process_per_session_str = conf->get_value("PROCESS_PER_SESSION");
                if (!process_per_session_str)
                        process_per_session = true; // default if option is missing
                else if (!strcmp(process_per_session_str, "yes"))
                        process_per_session = true;
		else
                        process_per_session = false;
                free(process_per_session_str);

                char *only_administrator_str = conf->get_value("ONLY_ADMINISTRATOR");
		if (!only_administrator_str)
                        only_administrator = true; // default if option is missing
                else if (!strcmp(only_administrator_str, "yes"))
                        only_administrator = true;
                else 
                        only_administrator = false;
                free(only_administrator_str);

                char *authenticate_str = conf->get_value("NT_AUTHENTICATION");
		if (!authenticate_str)
                        authenticate = true; // default if option is missing
		else if (!strcmp(authenticate_str, "yes"))
                        authenticate = true;
		else
                        authenticate = false;
                free(authenticate_str);

		if (authenticate) {
                        bool privs_ok = true;
                        if (!SetPrivilege(SE_TCB_NAME, true)) {
                                privs_ok = false;
                                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Enable SE_TCB_NAME (Act as part of the operating system) failed")));
                        }
                        if (!SetPrivilege(SE_CHANGE_NOTIFY_NAME, true)) {
                                privs_ok = false;
                                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Enable SE_CHANGE_NOTIFY_NAME (Bypass traverse checking) failed")));
                        }
                        if (!privs_ok) {                                                                             
				ACE_DEBUG((LM_CRITICAL, ACE_TEXT("One or more required process privileges are missing. These must be assigned for your Brutus Server to work. Please see previous log entries for missing privileges.")));
				retv = -1;
				goto exit;
			}
		}
        }

        // Server ORB command-line options
        build_argv(&orb_argc, &orb_argv, argv[0]); // add prog name
        orb_cmd_line_option = conf->get_value(SERVER_ORB_STACK);
        while (orb_cmd_line_option) {
                cmd_line.push_back(orb_cmd_line_option);
                build_argv(&orb_argc, &orb_argv, orb_cmd_line_option);
                orb_cmd_line_option = conf->get_value(SERVER_ORB_STACK);
        }

        // re-create configuration class
        delete conf;
        conf = NULL;
        try {
                conf = new Config(INTERNAL_ORB_STACK); // stack name
        }
        catch (std::bad_alloc &) {
                goto no_mem;
        }
        if (!conf) 
                goto no_mem;

        // load configuration
        retv = conf->load(BRUTUS_CONF);
        if (retv) {
                char msg[128] = {'\0'};
                if (retv > 0)
                        sprintf(msg, "Configuration error: Error in config file at line %d", retv);
                else
                        strcat(msg, "Configuration error: Could not open brutus.conf");

                ACE_DEBUG((LM_CRITICAL, ACE_TEXT(msg)));
                retv = -1;
                goto exit;
        }
                

        // Internal ORB command-line options
        build_argv(&internal_orb_argc, &internal_orb_argv, argv[0]); // add prog name
        internal_orb_cmd_line_option = conf->get_value(INTERNAL_ORB_STACK);
        while (internal_orb_cmd_line_option) {
                cmd_line.push_back(internal_orb_cmd_line_option);
                build_argv(&internal_orb_argc, &internal_orb_argv, internal_orb_cmd_line_option);
                internal_orb_cmd_line_option = conf->get_value(INTERNAL_ORB_STACK);
        }
        delete conf;
        conf = NULL;

        // Initiate internal ORB without interceptor
        global_internal_orb = CORBA::ORB_init(internal_orb_argc, internal_orb_argv, "omc.brutus.orb.internal");

        // register the interceptor
        try {
                initializer = new ORBInitializer(forward, global_internal_orb.in());
        }
        catch (std::bad_alloc &) {
                goto no_mem;
        }
        if (!initializer)
                goto no_mem;

        orb_initializer = initializer;
        try {
                PortableInterceptor::register_orb_initializer(orb_initializer.in()
                                                              ACE_ENV_ARG_PARAMETER);
        }
        catch (...) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Exception caught from register_orb_initializer()")));
                goto exit;
        }

        // Initiate ORB with interceptor
        global_server_orb = CORBA::ORB_init(orb_argc, orb_argv, "omc.brutus.orb.server");
        
        // run the service
        retv = PROCESS::instance()->run(argc, 
                                        argv, 
                                        profiles_on_disk,
                                        process_per_session,
                                        only_administrator,
                                        authenticate,
					life_check_delay);
        goto exit;

no_mem:
        ACE_DEBUG((LM_CRITICAL, ACE_TEXT("No memory")));
                        
exit:
        scramble_password();

        if (conf)
                delete conf;
        free(ssl_certificate);
        free(ssl_certificate_format);
        free(ssl_private_key);
        free(ssl_private_key_password_str);
        free(ssl_cert_file);
        free(ssl_cert_file_env);
        free(ssl_cert_dir);
        free(ssl_cert_dir_env);

        for (unsigned long n = 0; n < cmd_line.size(); n++)
                free(cmd_line[n]);
        cmd_line.clear();
        free(orb_argv);
        free(internal_orb_argv);

        ACE_DEBUG((LM_SHUTDOWN, ACE_TEXT("Brutus is down")));

        ExitProcess(retv); // avoid MAPIUninitialize memory access violation at exit
        return retv; // kill compiler warning
}

#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)
template class ACE_Singleton<Process, ACE_Mutex>;
#elif defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)
#pragma instantiate ACE_Singleton<Process, ACE_Mutex>
#endif /* ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION */
