/*
 *    Brutus source file implementing a managed group class.
 *    Copyright (C) 2004 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


//***************************************************************************
//
// Brutus makes use of the corbaloc mechanism as descibed in the TAO FAQ.
//
// The server key for the BrutusLogOn object is:
//   
//     BrutusLogOn::IOR_TABLE_KEY
// 
//***************************************************************************


#include <NT_Service/managed_group.h>
#include <NT_Service/defines.h>
#include <NT_Service/poa_utils.h>
#include <NT_Service/macros.h>
#include <tao/IORTable/IORTable.h>

#define MAPIGUID_H // do not include mapiguid.h
#include <mapix.h>
#include <NT_Service/templates.i>

ManagedBrutusGroup::ManagedBrutusGroup(const char *PoaName,
                                       PortableServer::POA_ptr ServerPOA,
                                       PortableServer::POA_ptr SessionParentPOA,
                                       PortableServer::ServantManager_ptr Activator,
                                       CORBA::ORB_ptr ServerORB,
                                       CORBA::ORB_ptr InternalORB)
{
        BRUTUS_LOG_REF;
        BRUTUS_LOG_SUB_INF("Entering ManagedBrutusGroup::ManagedBrutusGroup()");

        CORBA::ORB_var server_orb = CORBA::ORB::_duplicate(ServerORB);
        CORBA::ORB_var internal_orb = CORBA::ORB::_duplicate(InternalORB);
        PortableServer::ServantManager_var session_activator = PortableServer::ServantManager::_duplicate(Activator);                                           
        
        group_name = (char*)calloc(sizeof(char) * (strlen("BRUTUS_MANAGED_GROUP") + 1), sizeof(char));
        if (!group_name) {
                BRUTUS_LOG_SUB_CRITICAL("No memory");
                throw 0;
        }
        sprintf(group_name, "BRUTUS_MANAGED_GROUP"); 

        // create a child poa for this managed group with the specified name
        poa_ = create_poa((const char*)PoaName, PortableServer::POAManager::_nil(), ServerPOA);
        if (CORBA::is_nil(poa_)) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create POA");
                free(group_name);
                throw 0;
        }
        poa_manager_ = poa_->the_POAManager();
        poa_manager_->activate();

        // create and register the server servant activator.
        try {
                server_activator_ = new BrutusActivator_i();
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_ERR("No memory");
                throw 0;
        }
	try {
		poa_->set_servant_manager(server_activator_.in());
	}
	catch (const CORBA::Exception &e) {
		fflush(NULL);
		e._tao_print_exception("Could not set server servant activator");
                BRUTUS_LOG_SUB_CRITICAL("Exception caught when setting Servant Activator");
                throw 0;
	}

        // Create the external BrutusLogOn pseudo object.
        // Reference counting will eventually delete it.
        try {
                logon_obj_impl_ = new BRUTUS_BrutusLogOn_i_external();
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_ERR("No memory");
                throw 0;
        }
        logon_obj_oid_ = poa_->activate_object(logon_obj_impl_);
        CORBA::Object_var obj = poa_->id_to_reference(logon_obj_oid_);
        logon_obj_ = BRUTUS::BrutusLogOn::_narrow(obj);

        // 
        // Register the BrutusLogOn object in the server ORB IOR table
        //

        // Turn the object reference into an IOR string 
        CORBA::String_var ior_string = server_orb->object_to_string(logon_obj_.in());

        // Get a reference to the IOR Table 
        CORBA::Object_var table_obj = server_orb->resolve_initial_references("IORTable");
        IORTable::Table_var table = IORTable::Table::_narrow(table_obj.in());

        // Bind the stringified IOR in the IOR Table 
        char *logon_name = (char*)calloc(sizeof(char) * (strlen(::BRUTUS::BrutusLogOn::IOR_TABLE_KEY) + 1), sizeof(char));
        if (!logon_name) {
                BRUTUS_LOG_SUB_CRITICAL("No memory");
                throw 0;
        }
        sprintf(logon_name, ::BRUTUS::BrutusLogOn::IOR_TABLE_KEY);
        table->bind((const char*)logon_name, ior_string.in());

        std::string msg = (const char*)logon_name;
        free(logon_name);
        msg += " IOR server key created";
        BRUTUS_LOG_SUB_INF(msg.c_str());

        BRUTUS_LOG_SUB_INF("Leaving ManagedBrutusGroup::ManagedBrutusGroup()");
}


ManagedBrutusGroup::~ManagedBrutusGroup(void)
{
        BRUTUS_LOG_REF;
        BRUTUS_LOG_SUB_INF("Entering ManagedBrutusGroup::~ManagedBrutusGroup()");

        // deactivate objects
        poa_->deactivate_object(logon_obj_oid_);

        // clean up POAs
        poa_manager_->deactivate(1, 1);
        poa_->destroy(1, 1);

        std::string msg = (const char*)group_name;
        msg += " destroyed";
        free(group_name);
        BRUTUS_LOG_SUB_INF(msg.c_str());

        BRUTUS_LOG_SUB_INF("Leaving ManagedBrutusGroup::~ManagedBrutusGroup()");
}


PortableServer::POAManager::State ManagedBrutusGroup::get_factory_state()
{
        return poa_manager_->get_state();
}


void ManagedBrutusGroup::activate_factory()
{
        poa_manager_->activate();
}


void ManagedBrutusGroup::hold_factory_requests(CORBA::Boolean wait_for_completion)
{
        poa_manager_->hold_requests(wait_for_completion);
}


void ManagedBrutusGroup::discard_factory_requests(CORBA::Boolean wait_for_completion)
{
        poa_manager_->discard_requests(wait_for_completion);
}
