/*
 *    Brutus source file implementing the Brutus NT Service.
 *    Copyright (C) 2004 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// This is for the purpose of having the copyright notice 
// in the binary as well as in the source code.
const char copyright[] = "Copyright (C) 2004, 2005 OMC Denmark ApS.";

//**************************************************************************
//
// BASED ON ACE SAMPLE CODE BY:
//    Gonzalo Diethelm <gonzo@cs.wustl.edu> 
//    and Steve Huston <shuston@riverace.com>
//
//**************************************************************************


#include <NT_Service/ntsvc.h>

#include <string>

#include <tao/corba.h>
#include <tao/IORTable/IORTable.h>
#include <tao/PortableServer/PortableServer.h>
#include <ace/OS.h>
#include <ace/Task.h>
#include <ace/Reactor.h>
#include <ace/Log_Msg.h>

#include <NT_Service/servant_impl/BrutusSessionS_impl.h>
#include <NT_Service/servant_impl/BrutusLogOnS_impl.h>
#include <NT_Service/servant_activator.h>
#include <NT_Service/managed_group.h>
#include <NT_Service/poa_utils.h>
#include <NT_Service/main.h>
#include <NT_Service/macros.h>
#include <NT_Service/defines.h>

#include <NT_Service/ProxyMgr/proxy_managerS_impl.h>
#include <NT_Service/servant_impl/log_serverS_impl.h>

#define MAPIGUID_H // do not include mapiguid.h
#include <mapix.h>
#include <NT_Service/templates.i>


class ORB_Thread_Base : public ACE_Task_Base {
public:
        ORB_Thread_Base(void)
                {
                };

        void set_orb(CORBA::ORB_ptr Broker)
                {
                        orb_ = CORBA::ORB::_duplicate(Broker);
                };

        virtual int svc(void)
                {
                        if (CORBA::is_nil(orb_))
                                throw 1;
                        
                        orb_->run();

                        return 0;
                };

private:
        CORBA::ORB_var orb_;
};

Service::Service (void)
{
        // Remember the Reactor instance.
        reactor(ACE_Reactor::instance());
        
        // sensible defaults
        profiles_on_disk_ = true;
        process_per_session_ = true;
        only_administrator_ = false;
        authenticate_ = true;
	life_check_delay_ = 0;
}


// The number of threads wich serves Brutus requests. 
// Each of these threads has one POA, one BrutusLogOn,
// one ServantManager and one ServantActivator.
// Execution options set too.
void Service::SetOptions(const bool debug,
                         const bool ProfilesOnDisk,
                         const bool ProcessPerSession,
                         const bool OnlyAdministrator,
                         const bool Authenticate,
			 const unsigned long LifeCheckDelay)
{
        debug_ = debug;
        profiles_on_disk_ = ProfilesOnDisk;
        process_per_session_ = ProcessPerSession;
        only_administrator_ = OnlyAdministrator;
        authenticate_ = Authenticate;
	life_check_delay_ = LifeCheckDelay;
}


// This method is called when the service gets a control request.  It
// handles requests for stop and shutdown by calling terminate ().
// All others get handled by calling up to inherited::handle_control.
void Service::handle_control(DWORD control_code)
{
        switch (control_code) {
        case SERVICE_CONTROL_SHUTDOWN :
        case SERVICE_CONTROL_STOP :
                report_status(SERVICE_STOP_PENDING);
                BRUTUS_LOG_INF("Service control stop requested");

                stop_ = 1;
                reactor()->notify(this, ACE_Event_Handler::EXCEPT_MASK);
                break;
        default :                
                inherited::handle_control(control_code);
                break;
        }
}


// This is just here to be the target of the notify from above... it
// doesn't do anything except aid on popping the reactor off its wait
// and causing a drop out of handle_events.
int Service::handle_exception(ACE_HANDLE)
{
        return 0;
}

// This is the main processing function for the Service.  It sets up
// the initial configuration and runs the event loop until a shutdown
// request is received.
int Service::svc(void)
{
	ACE_LOG_MSG->open("Brutus Server", ACE_Log_Msg::SYSLOG);

        // referencing the copyright string constant will prevent any 
        // compiler, however clever, to optimize it away (I hope). 
        {
                const char *dummy = copyright;
                dummy++;
        }

        // As an NT service, we come in here in a different thread than the
        // one which created the reactor.  So in order to do anything, we
        // need to own the reactor. If we are not a service, report_status
        // will return -1.
        if (!report_status(SERVICE_RUNNING)) {
                reactor()->owner(ACE_Thread::self());
        }
        stop_ = 0;
        
        // A comment because my eyes like it...
        CORBA::ORB_var server_orb = global_server_orb;
        CORBA::ORB_var internal_orb = global_internal_orb;

        // MAPI initialization
        if (!process_per_session_) {
                MAPIINIT_0 mapi_init;

                mapi_init.ulFlags = MAPI_MULTITHREAD_NOTIFICATIONS | MAPI_NT_SERVICE | (profiles_on_disk_ ? MAPI_TEMPORARY_PROFILES : 0);
                mapi_init.ulVersion = MAPI_INIT_VERSION;
                HRESULT hr = MAPIInitialize(&mapi_init);
                if (hr != S_OK) {
                        server_orb->shutdown(1);
                        internal_orb->shutdown(1);
			ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not initialize MAPI")));
                        return -1;
                }
        }

        /////////////////////////////////////////
        //     Intercepted server root POA     //
        /////////////////////////////////////////

        PortableServer::POA_var root_poa = resolve_init<PortableServer::POA>(server_orb.in(), "RootPOA");
        if (CORBA::is_nil(root_poa.in())) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create root POA")));
                return -1;
        }

        // server root poa manager
        PortableServer::POAManager_var root_poa_manager = root_poa->the_POAManager();
        if (CORBA::is_nil(root_poa_manager.in())) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create POAManager")));
                return -1;
        }
        root_poa_manager->activate();

        // brutus server poa
        PortableServer::POA_var brutus_poa = create_poa((const char*)"BRUTUS_ROOT_POA", PortableServer::POAManager::_nil(), root_poa);
        if (CORBA::is_nil(brutus_poa.in())) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create Brutus Server POA")));
                return -1;
        }

        // brutus server poa manager
        PortableServer::POAManager_var brutus_poa_manager = brutus_poa->the_POAManager();
        if (CORBA::is_nil(brutus_poa_manager.in())) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create server POAManager")));
                return -1;
        }
        brutus_poa_manager->activate();


        //////////////////////////////////////////////
        //     Non-intercepted session root POA     //
        //////////////////////////////////////////////

        PortableServer::POA_var root_session_poa = resolve_init<PortableServer::POA>(internal_orb.in(), "RootPOA");
        if (CORBA::is_nil(root_session_poa.in())) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create root session POA")));
                return -1;
        }

        // root session poa manager
        PortableServer::POAManager_var root_session_poa_manager = root_session_poa->the_POAManager();
        if (CORBA::is_nil(root_session_poa_manager.in())) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create session POAManager")));
                return -1;
        }
        root_session_poa_manager->activate();
        

        ////////////////////////////////////////////////
        //     Non-intercepted session parent POA     //
        ////////////////////////////////////////////////

        PortableServer::POA_var session_parent_poa = create_poa((const char*)"SessionParentPOA", PortableServer::POAManager::_nil(), root_session_poa);
        if (CORBA::is_nil(session_parent_poa)) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create session parent POA")));
                return -1;
        }

        // session parent POA manager
        PortableServer::POAManager_var session_parent_poa_manager = session_parent_poa->the_POAManager();
        if (CORBA::is_nil(session_parent_poa_manager.in())) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Could not create session parent POAManager")));
                return -1;
        }
        session_parent_poa_manager->activate();
        

        ////////////////////////////////////////////////////////////////////
        // create and register the servant activator for both parent POAs //
        ////////////////////////////////////////////////////////////////////

        PortableServer::ServantManager_var session_activator;
        try {
                session_activator = new BrutusActivator_i();
        }
        catch (std::bad_alloc &) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("No memory")));
                return -1;
        }
	try {
		session_parent_poa->set_servant_manager(session_activator.in());
	}
	catch (const CORBA::Exception &) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Exception caught when setting Servant Activator")));
                return -1;
	}
	try {
		brutus_poa->set_servant_manager(session_activator.in());
	}
	catch (const CORBA::Exception &) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("Exception caught when setting Servant Activator")));
                return -1;
	}


        ///////////////////////////////////////////////////////////////
        // create LogServer under non-intercepted session parent POA //
        ///////////////////////////////////////////////////////////////

        BRUTUS_LogServer_i *log_server_impl = NULL;
        try {
                log_server_impl = new BRUTUS_LogServer_i();
        }
        catch (std::bad_alloc &) {
                ACE_DEBUG((LM_CRITICAL, ACE_TEXT("No memory")));
                return -1;
        }
        PortableServer::ObjectId_var obj_oid = session_parent_poa->activate_object(log_server_impl);
        CORBA::Object_var obj = session_parent_poa->id_to_reference(obj_oid);
        BRUTUS::LogServer_var log_server = BRUTUS::LogServer::_narrow(obj);

        CORBA::String_var ior = (const char*)"";
        { // write LogServer IOR to file
                ior = internal_orb->object_to_string(log_server.in());
                ofstream ior_file(LOG_SERVER_FILE, ios::out | ios::trunc);
                if (!ior_file.is_open()) {
                        return -1;
                }
                ior_file << ior.in() << std::endl;
                ior_file.close();
        }
        // logging is now enabled!

        ACE_DEBUG((LM_INFO, ACE_TEXT("LogServer incarnated - Most logging deferred to the Brutus LogServer...")));
        BRUTUS_LOG_REF;
        BRUTUS_LOG_SUB_INF("Brutus starting");


        ///////////////////////////////////////////////////////////////////
        // create proxy manager under non-intercepted session parent POA //
        ///////////////////////////////////////////////////////////////////

        BRUTUS_ProxyManager_i *proxy_manager_impl = NULL;
        try {
                proxy_manager_impl = new BRUTUS_ProxyManager_i();
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_CRITICAL("No memory");
                return -1;
        }
        obj_oid = session_parent_poa->activate_object(proxy_manager_impl);
        obj = session_parent_poa->id_to_reference(obj_oid);
        BRUTUS::ProxyManager_var proxy_manager = BRUTUS::ProxyManager::_narrow(obj);

        ior = (const char*)"";
        { // write proxy manager IOR to file
                ior = internal_orb->object_to_string(proxy_manager.in());
                ofstream ior_file(PROXY_MANAGER_FILE, ios::out | ios::trunc);
                if (!ior_file.is_open()) {
                        BRUTUS_LOG_SUB_CRITICAL("Could not write ProxyMgr IOR to file");
                        return -1;
                }
                ior_file << ior.in() << std::endl;
                ior_file.close();
        }

        ///////////////////////////////////////////////////////
        // create BrutusSession under intercepted Brutus POA //
        ///////////////////////////////////////////////////////

        BRUTUS_BrutusSession_i *session_obj_impl = NULL; 
        try {
                session_obj_impl = new BRUTUS_BrutusSession_i(proxy_manager.in(), internal_orb.in());
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_CRITICAL("No memory");
                return -1;
        }
        obj_oid = brutus_poa->activate_object(session_obj_impl);
        obj = brutus_poa->id_to_reference(obj_oid);
        BRUTUS::BrutusSession_var session_obj = BRUTUS::BrutusSession::_narrow(obj);
        { // write BrutusSession IOR to file
                ior = internal_orb->object_to_string(session_obj.in());
                ofstream ior_file(BRUTUS_SESSION_FILE, ios::out | ios::trunc);
                if (!ior_file.is_open()) {
                        BRUTUS_LOG_SUB_CRITICAL("Could not write BrutusSession IOR to file");
                        return -1;
                }
                ior_file << ior.in() << std::endl;
                ior_file.close();
        }

        ////////////////////////////////////////////////////////////////////
        // Register the BrutusSession object in the session ORB IOR table //
        ////////////////////////////////////////////////////////////////////

        // Get a reference to the IOR Table 
        CORBA::Object_var table_obj = server_orb->resolve_initial_references("IORTable");
        IORTable::Table_var table = IORTable::Table::_narrow(table_obj.in());

        // Turn the object reference into an IOR string 
        ior = server_orb->object_to_string(session_obj.in());
        
        // Bind the stringified IOR in the IOR Table 
        char *brutus_session_name = (char*)calloc(sizeof(char) * (strlen(::BRUTUS::BrutusSession::IOR_TABLE_KEY) + 1), sizeof(char));
        if (!brutus_session_name) {
                BRUTUS_LOG_SUB_CRITICAL("No memory");
                return -1;
        }
        sprintf(brutus_session_name, ::BRUTUS::BrutusSession::IOR_TABLE_KEY);
        table->bind((const char*)brutus_session_name, ior.in());
        free(brutus_session_name);


        ///////////////////////////////////
        // create all the worker objects //
        ///////////////////////////////////


        // create the registry of served sessions
        SessionRegistry served_sessions_;

        // create session POA from the non-intercepted session parent POA
        PortableServer::POA_var session_poa_ = create_poa("SESSION_POA", PortableServer::POAManager::_nil(), session_parent_poa.in());
        if (CORBA::is_nil(session_poa_.in())) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create session root POA");
                return -1;
        }

        // session POA manager
        PortableServer::POAManager_var session_poa_manager_ = session_poa_->the_POAManager();
        if (CORBA::is_nil(session_poa_manager_.in())) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create session POAManager");
                return -1;
        }
        session_poa_manager_->activate();
        
        // set servant activator
	try {
		session_poa_->set_servant_manager(session_activator.in());
	}
	catch (const CORBA::Exception &) {
                BRUTUS_LOG_SUB_CRITICAL("Exception caught when setting Servant Activator");
                return -1;
	}

        // Create the BrutusLogOn object under the non-intercepted session parent POA.
        // Reference counting will eventually delete it.
        BRUTUS_BrutusLogOn_i *logon_obj_impl = NULL;
        try {
                logon_obj_impl = new BRUTUS_BrutusLogOn_i(session_poa_.in(), // <== session POA
                                                          session_activator.in(), 
                                                          &served_sessions_,
                                                          process_per_session_,
                                                          only_administrator_,
                                                          authenticate_,
                                                          profiles_on_disk_,
							  life_check_delay_,
                                                          internal_orb.in());
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_ERR("No memory");
                return -1;
        }
        obj_oid = session_parent_poa->activate_object(logon_obj_impl);
        obj = session_parent_poa->id_to_reference(obj_oid);
        BRUTUS::BrutusLogOn_var logon_obj = BRUTUS::BrutusLogOn::_narrow(obj);
        { // write BrutusLogon IOR to file
                ior = internal_orb->object_to_string(logon_obj.in());
                ofstream ior_file(BRUTUS_LOGON_FILE, ios::out | ios::trunc);
                if (!ior_file.is_open()) {
                        BRUTUS_LOG_SUB_CRITICAL("Could not write BrutusLogon IOR to file");
                        return -1;
                }
                ior_file << ior.in() << std::endl;
                ior_file.close();
        }

        ManagedBrutusGroup *managed_group = NULL;
        try {
                managed_group = new ManagedBrutusGroup("GROUP_POA:0", 
                                                       brutus_poa.in(),
                                                       session_parent_poa.in(),
                                                       session_activator.in(),
                                                       server_orb.in(), // <== server ORB
                                                       internal_orb.in()); // <== session ORB
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_CRITICAL("No memory");
                return -1;
        }
        catch (...) {
                BRUTUS_LOG_SUB_CRITICAL("Could not create ManagedBrutusGroup");
                return -1;
        }


        ///////////////
        // ORB loops //
        ///////////////

        // run the internal ORB loop
        ORB_Thread_Base internal_orb_thread;
        internal_orb_thread.set_orb(internal_orb.in());
        internal_orb_thread.activate();
        Sleep(1000);

        // run the server ORB loop
        ORB_Thread_Base server_orb_thread;
        server_orb_thread.set_orb(server_orb.in());
        server_orb_thread.activate();
        Sleep(1000);

        BRUTUS_LOG_SUB_INF("Brutus is running");
        ACE_DEBUG((LM_STARTUP, ACE_TEXT("Brutus is running")));
        
        while (!stop_) {
                reactor()->handle_events();
        }
        BRUTUS_LOG_SUB_INF("Brutus shutting down");

        // clean up
        delete managed_group;

        brutus_poa_manager->deactivate(1, 1);
        brutus_poa->destroy(1, 1);
        root_poa_manager->deactivate(1, 1);
        root_poa->destroy(1, 1);
        server_orb->shutdown(1);
        server_orb_thread.wait();
        server_orb->destroy();

        session_parent_poa_manager->deactivate(1, 1);
        session_parent_poa->destroy(1, 1);
        root_session_poa_manager->deactivate(1, 1);
        root_session_poa->destroy(1, 1);
        internal_orb->shutdown(1);
        internal_orb_thread.wait(); 
        internal_orb->destroy();

        if (!process_per_session_) 
                MAPIUninitialize();

        return 0;
}

#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)
template class ACE_Singleton<Service, ACE_Mutex>;
#elif defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)
#pragma instantiate ACE_Singleton<Service, ACE_Mutex>
#endif /* ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION */
