/*
 *    Brutus source file implementing POA utility functions.
 *    Copyright (C) 2004 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <TAO/tao/BiDir_GIOP/BiDirGIOP.h>
#include <TAO/tao/AnyTypeCode/Any.h>

#include <NT_Service/poa_utils.h>
#include <NT_Service/config.h>
#include <NT_Service/macros.h>
#include <objbase.h>

static bool use_bidir(void)
{
	return false;

        // configuration helper class
        Config *conf = NULL;

        // create configuration class
        try {
                conf = new Config();
        }
        catch (std::bad_alloc &) {
                goto no_mem;
        }
        if (!conf) 
                goto no_mem;

	bool retv = true;
        {
                char *bidir_str = conf->get_value("USE_BIDIR_GIOP");
                if ((bidir_str) && (!strcmp(bidir_str, "yes"))) 
                        retv = true;
		else
			retv = false;
                free(bidir_str);
        }
        delete conf;

	return retv;

no_mem:
	return false; // defaulting to the default seems to be the safe bet...
}

PortableServer::POA_ptr create_poa(const char *Name, 
                                   PortableServer::POAManager_ptr PoaManager,
                                   PortableServer::POA_ptr Poa)
{
	static bool use_bidir_giop = use_bidir();
	unsigned int pol_len = 7 + (use_bidir_giop ? 1 : 0);
        PortableServer::POA_var retval = PortableServer::POA::_nil();
	CORBA::ORB_var orb = CORBA::ORB::_nil();
	CORBA::Any bidir_pol;
        
        CORBA::PolicyList pl(pol_len);
        pl.length(pol_len);
        pl[0] = Poa->create_lifespan_policy(PortableServer::TRANSIENT);
        pl[1] = Poa->create_id_assignment_policy(PortableServer::SYSTEM_ID);
        pl[2] = Poa->create_id_uniqueness_policy(PortableServer::UNIQUE_ID);
        pl[3] = Poa->create_implicit_activation_policy(PortableServer::NO_IMPLICIT_ACTIVATION);
        pl[4] = Poa->create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
        pl[5] = Poa->create_servant_retention_policy(PortableServer::RETAIN);
        pl[6] = Poa->create_thread_policy(PortableServer::ORB_CTRL_MODEL);
	if (use_bidir_giop) {
		orb = Poa->_get_orb();

		bidir_pol <<= BiDirPolicy::BOTH;
		pl[7] = orb->create_policy(BiDirPolicy::BIDIRECTIONAL_POLICY_TYPE, bidir_pol);
	}

        try {
                retval = Poa->create_POA(Name, PoaManager, pl);
        }
        catch (PortableServer::POA::AdapterAlreadyExists &) {
                for (unsigned int n = 0; n < pol_len; n++)
                        pl[n]->destroy();

                throw; // re-throw to caller
        }
        catch (...) {
                BRUTUS_LOG_ERR("Exception caught");
                for (unsigned int n = 0; n < pol_len; n++)
                        pl[n]->destroy();

                return PortableServer::POA::_nil();
        }
        for (unsigned int n = 0; n < pol_len; n++)
                pl[n]->destroy();

        return retval._retn();
}


SessionRegistry::~SessionRegistry()
{
        ACE_Guard<ACE_RW_Mutex> entry_guard(entry_mutex_); 
        for (unsigned long n = 0; n < managed_entries_.size(); n++) {
                free(managed_entries_[n]->profile_name);

                PortableServer::POA_var poa = managed_entries_[n]->session_poa;
                PortableServer::POAManager_var poa_manager = poa->the_POAManager();

                // makes sure that all objects are closed from the world
                poa_manager->deactivate(0, 0);

                // initiate etheralize() by calling _remove_ref()
                // on all objects in this poa
                poa->destroy(1, 0);

                // remove the element
                delete managed_entries_[n];
        }
        managed_entries_.resize(0);

        ACE_Guard<ACE_RW_Mutex> block_guard(block_mutex_); 
        for (unsigned long i = 0; i < blocked_users_.size(); i++) {
                free(blocked_users_[i]);
        }
        blocked_users_.resize(0);
}


unsigned long SessionRegistry::add_entry(PortableServer::POA_ptr Poa,
                                         const char *ProfileName)
{
        BRUTUS_LOG_REF;

        if (blocked(ProfileName)) {
                msg = ProfileName;
                msg += " was blocked from logon";
                BRUTUS_LOG_SUB_PRIVATE(msg.c_str());

                return 0;
        }

        SessionEntry *reg_entry;
        try {
                reg_entry = new SessionEntry;
        }
        catch (std::bad_alloc &) {
                BRUTUS_LOG_SUB_ERR("No memory");
                return 0;
        }

        reg_entry->profile_name = strdup(ProfileName);
        if (!reg_entry->profile_name) {
                BRUTUS_LOG_SUB_ERR("No memory");
                free(reg_entry);
                return 0;
        }

        reg_entry->session_poa = PortableServer::POA::_duplicate(Poa);
        reg_entry->id = ++last_session_id;

        {
                ACE_Write_Guard<ACE_RW_Mutex> guard(entry_mutex_); 
                managed_entries_.push_back(reg_entry);
        }        

        return last_session_id;
}


void SessionRegistry::sign_out(unsigned long SessionID,
                               CORBA::Boolean wait_for_completion)
{
        ACE_Write_Guard<ACE_RW_Mutex> guard(entry_mutex_); 

        unsigned long n = managed_entries_.size();
        while (n) {
                n--;
                if (managed_entries_[n]->id == SessionID) {
                        free(managed_entries_[n]->profile_name);

                        PortableServer::POA_var poa = managed_entries_[n]->session_poa;
                        PortableServer::POAManager_var poa_manager = poa->the_POAManager();

                        // makes sure that all objects are closed from the world
                        poa_manager->deactivate(0, wait_for_completion);

                        // initiate etheralize() by calling _remove_ref()
                        // on all objects in this poa
                        poa->destroy(1, wait_for_completion);

                        // remove the element
                        delete managed_entries_[n];
                        managed_entries_.erase(managed_entries_.begin() + n);

                        return;
                }
        }
}


/* 
 * There might be multiple entries for any given ProfileName.
 * We must not unblock the user as that is an administrative
 * decision.
 */
void SessionRegistry::sign_out(const char *ProfileName,
                               CORBA::Boolean wait_for_completion)
{
        char signed_out = 0;

        ACE_Write_Guard<ACE_RW_Mutex> guard(entry_mutex_); 

        unsigned long n = managed_entries_.size();
        while (n) {
                n--;
                if (!strcmp(managed_entries_[n]->profile_name, ProfileName)) {
                        free(managed_entries_[n]->profile_name);

                        PortableServer::POA_var poa = managed_entries_[n]->session_poa;
                        PortableServer::POAManager_var poa_manager = poa->the_POAManager();

                        // makes sure that all objects are closed from the world
                        poa_manager->deactivate(0, wait_for_completion);

                        // initiate etheralize() by calling _remove_ref()
                        // on all objects in this poa
                        poa->destroy(1, wait_for_completion);

                        // remove the element
                        delete managed_entries_[n];
                        managed_entries_.erase(managed_entries_.begin() + n);

                        return;
                }
        }
}


// Any blocked user will at most get one entry in 
// the blocked users list
void SessionRegistry::block(const char *ProfileName)
{
        if (!ProfileName)
                return;

        ACE_Write_Guard<ACE_RW_Mutex> guard(block_mutex_); 
        for (unsigned long n = 0; n < blocked_users_.size(); n++) {
                if (!strcmp(blocked_users_[n], ProfileName)) {
                        return;
                }
        }
        blocked_users_.push_back(strdup(ProfileName));
}


void SessionRegistry::unblock(const char *ProfileName)
{
        if (!ProfileName)
                return;

        ACE_Write_Guard<ACE_RW_Mutex> guard(block_mutex_); 
        for (unsigned long n = 0; n < blocked_users_.size(); n++) {
                if (!strcmp(blocked_users_[n], ProfileName)) {
                        free(blocked_users_[n]);
                        blocked_users_.erase(blocked_users_.begin() + n);
                        return;
                }
        }
}


bool SessionRegistry::blocked(const char *ProfileName)
{
        if (!ProfileName)
                return false;

        ACE_Read_Guard<ACE_RW_Mutex> guard(block_mutex_); 
        for (unsigned long n = 0; n < blocked_users_.size(); n++) {
                if (!strcmp(blocked_users_[n], ProfileName)) 
                        return true;
        }

        return false;
}


char* SessionRegistry::get_profile_name(unsigned long SessionID)
{
        ACE_Read_Guard<ACE_RW_Mutex> guard(entry_mutex_); 
        for (unsigned long n = 0; n < managed_entries_.size(); n++) {
                if (managed_entries_[n]->id == SessionID) {
                        char *retval = strdup(managed_entries_[n]->profile_name);
                        if (!retval)
                                BRUTUS_LOG_ERR("No memory");

                        return retval;
                }
        }
        BRUTUS_LOG_BUG("Possible data curruption in session registry");

        return 0;
}
