/*
 *    Brutus source file for RSA helper functions.
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// Needed *RIGHT HERE* or we can't include <wincrypt.h>
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400
#endif

#include <NT_Service/rsa_utils.h>
#include <NT_Service/macros.h>

#include <openssl/err.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/rand.h>

#include <wincrypt.h>


RSA *read_pub_key(const char *pubkey_file)
{
        RSA *key = NULL;
        BIO *bp = NULL;

        ERR_load_crypto_strings();

        bp = BIO_new(BIO_s_file());
        if (BIO_read_filename(bp, pubkey_file) <= 0) 
                goto exit;

        key = PEM_read_bio_RSAPublicKey(bp, NULL, NULL, NULL);
        if (!key)
                goto exit;

exit:
        BIO_free(bp);

        return key;
}

RSA *read_priv_key(const char *privkey_file)
{
        RSA *key = NULL;
        BIO *bp = NULL;

        SSLeay_add_all_algorithms();
        ERR_load_PEM_strings();

        bp = BIO_new(BIO_s_file());
        if (BIO_read_filename(bp, privkey_file) <= 0) 
                goto exit;

        key = PEM_read_bio_RSAPrivateKey(bp, NULL, NULL, NULL);
        if (!key) 
                goto exit;

exit:
        BIO_free(bp);
        
        return key;
}

bool create_rsa_files(const char *pubkey_file,
                      const char *privkey_file)
{
        RSA *rsa = NULL;
        RSA *pub_rsa = NULL;
        RSA *priv_rsa = NULL;
        int len = 0;
        unsigned char buf[RSA_KEY_LENGTH * 2] = { 0 };
        unsigned char *p = NULL;

        HCRYPTPROV cx = NULL;
        if (!CryptAcquireContextA(&cx, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) 
                if (!CryptAcquireContextA(&cx, NULL, NULL, PROV_RSA_FULL, CRYPT_NEWKEYSET)) {
                        return false;
                }

        unsigned char seed[64];
        if (!CryptGenRandom(cx, 64, seed)) {
                return false;
        }

        RAND_seed(seed, 64);
        if (!CryptReleaseContext(cx, 0)) {
                return false;
        }

        rsa = RSA_generate_key(RSA_KEY_LENGTH, RSA_F4, NULL, (char *)stdout);
        RSA_blinding_on(rsa, NULL);

        p = buf;

        /* Save the public key into buffer, we know it will be big enough
         * but we should really check how much space we need by calling the
         * i2d functions with a NULL second parameter */
        len = i2d_RSAPublicKey(rsa, &p);
        len += i2d_RSAPrivateKey(rsa, &p);

        RSA_free(rsa);

        p = buf;
        pub_rsa = d2i_RSAPublicKey(NULL, (const unsigned char**) &p, (long)len);
        len -= (p-buf);
        priv_rsa = d2i_RSAPrivateKey(NULL, (const unsigned char**)&p, (long)len);

        if ((pub_rsa == NULL) || (priv_rsa == NULL)) {
                return false;
        }

        if (!RSA_check_key(priv_rsa)) {
                return false;
        } 

        // store public key 
        BIO *pub = BIO_new_file(pubkey_file, "w");
        PEM_write_bio_RSAPublicKey(pub, pub_rsa); 
        BIO_free(pub);
        RSA_free(pub_rsa);

        // store private key
        BIO *priv = BIO_new_file(privkey_file, "w");
        PEM_write_bio_RSAPrivateKey(priv, priv_rsa, NULL, NULL, 0, NULL, NULL); 
        BIO_free(priv);
        RSA_free(priv_rsa);

        return true;
}
