/*
 *    Brutus source file implementing a ServerRequestInterceptor.
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <NT_Service/server_request_interceptor.h>
#include <NT_Service/corbaloc.h>
#include <NT_Service/defines.h>
#include <NT_Service/config.h>

#include <string.h>
#include <stdio.h>


ServerRequestInterceptor::ServerRequestInterceptor(bool forward,
                                                   CORBA::ORB_ptr ORB)
        : forward_(forward),
          orb_(ORB)
{
        if (!forward_) {
                session_objs_.resize(1);
                logon_objs_.resize(1);

                // local objects always at index 0
                session_objs_[0].location = strdup(BRUTUS_SESSION_IOR);
                logon_objs_[0].location =  strdup(BRUTUS_LOGON_IOR);
                session_objs_[0].obj = BRUTUS::BrutusSession::_nil();
                logon_objs_[0].obj = BRUTUS::BrutusLogOn::_nil();
                
                return;
        } 
        session_objs_.resize(10);
        logon_objs_.resize(10);
        
        // configuration helper class
        Config *conf = new Config(SERVER_ORB_STACK); // stack name
        if (!conf) 
                throw 0;

        // load port configuration
        int retv = conf->load(BRUTUS_CONF);
        if (retv)
                throw 0;

        // get endpoint port numbers
        char *orb_cmd_line_option = conf->get_value(SERVER_ORB_STACK);
        while (orb_cmd_line_option) {
                if (!strcmp(orb_cmd_line_option, "-ORBListenEndpoints"))
                        break;

                free(orb_cmd_line_option);
                orb_cmd_line_option = conf->get_value(SERVER_ORB_STACK);
                if (!orb_cmd_line_option)
                        throw 0;
        }
        free(orb_cmd_line_option);
        orb_cmd_line_option = conf->get_value(SERVER_ORB_STACK);
        unsigned short iiop_port = 0;
        unsigned short ssl_port = 0; // 0 if not using SSL
        if (!get_ports(orb_cmd_line_option, iiop_port, ssl_port))
                throw 0;
        free(orb_cmd_line_option);
        unsigned short port = ssl_port ? ssl_port : iiop_port;
        delete conf;

        // load cluster configuration
        conf = new Config();
        if (!conf)
                throw 0;
        retv = conf->load(BRUTUS_SERVERS);
        if (retv)
                throw 0;
        const char *name = NULL;
        const char *ipaddr = NULL;

        unsigned long n = 1;
        unsigned char c = 1;
        while (conf->get_token_value_pair(name, ipaddr)) {

                if (c == 9) {
                        session_objs_.resize(n + 10);
                        logon_objs_.resize(n + 10);
                        c = 0;
                }

                if (!strcmp("localhost", name) && (strcmp("yes", ipaddr))) {
                        free(session_objs_[0].location);
                        free(logon_objs_[0].location);
                        session_objs_[0].location = NULL;
                        logon_objs_[0].location = NULL;

                        continue;
                }

                session_objs_[n].location = corbaloc_str(::BRUTUS::BrutusSession::IOR_TABLE_KEY, ipaddr, port);
                logon_objs_[n].location = corbaloc_str(::BRUTUS::BrutusLogOn::IOR_TABLE_KEY, ipaddr, port);
                session_objs_[n].obj = BRUTUS::BrutusSession::_nil();
                logon_objs_[n].obj = BRUTUS::BrutusLogOn::_nil();

                c++;
                n++;
        }
        session_objs_.resize(n);
        logon_objs_.resize(n);

        delete conf;
}

char *ServerRequestInterceptor::name(ACE_ENV_SINGLE_ARG_DECL_NOT_USED)
        ACE_THROW_SPEC((CORBA::SystemException))
{
        return CORBA::string_dup("Brutus ServerRequestInterceptor");
}

void ServerRequestInterceptor::destroy(ACE_ENV_SINGLE_ARG_DECL_NOT_USED)
        ACE_THROW_SPEC((CORBA::SystemException))
{
}

void ServerRequestInterceptor::receive_request_service_contexts(PortableInterceptor::ServerRequestInfo_ptr ri
                                                                ACE_ENV_ARG_DECL)
        ACE_THROW_SPEC((CORBA::SystemException,
                        PortableInterceptor::ForwardRequest))
{
        // Throw BAD_INV_ORDER for all other operations on intercepted BrutusLogOn interface. 
        // One of Logon() or GetPublicKey() must be called first. The exception is thrown by 
        // the servant.
        if (strcmp("Logon", (const char*)ri->operation()) 
            && strcmp("GetPublicKey", (const char*)ri->operation()))
                return;

        /* 
         * Forward invocations on exposed Logon() and GetPublicKey() methods
         */

        // don't forward to other servers, use local object
        if (!forward_) {
                ACE_Write_Guard<ACE_RW_Mutex> guard(mutex_);
                if (CORBA::is_nil(logon_objs_[0].obj.in())) {
                        CORBA::Object_var obj = orb_->string_to_object(logon_objs_[0].location);
                        logon_objs_[0].obj = BRUTUS::BrutusLogOn::_narrow(obj.in());
                }
                // permanent forwarding
                throw (PortableInterceptor::ForwardRequest(logon_objs_[0].obj.in()));
        } 

        // forward to least loaded server, possibly the local one
        {
                ACE_Write_Guard<ACE_RW_Mutex> guard(mutex_);

                bool found = false;
                unsigned long least_loaded_proxy = 0;
                unsigned long proxy_load = 0;
                unsigned long proxy_load_min = 0;
                CORBA::Object_var obj = CORBA::Object::_nil();                

                for (unsigned long n=0; n<session_objs_.size(); n++) {

                        // get reference
                        try {
                                if (CORBA::is_nil(session_objs_[n].obj.in())) {
                                        obj = orb_->string_to_object(session_objs_[n].location);
                                        session_objs_[n].obj = ::BRUTUS::BrutusSession::_narrow(obj.in());
                                }
                                if (CORBA::is_nil(logon_objs_[n].obj.in())) {
                                        obj = orb_->string_to_object(logon_objs_[n].location);
                                        logon_objs_[n].obj = ::BRUTUS::BrutusLogOn::_narrow(obj.in());
                                }
                        }
                        catch (...) {
                                session_objs_[n].obj = BRUTUS::BrutusSession::_nil();
                                logon_objs_[n].obj = BRUTUS::BrutusLogOn::_nil();
                                continue;
                        }

                        // check reference
                        if (CORBA::is_nil(session_objs_[n].obj.in()) 
                            || (CORBA::is_nil(logon_objs_[n].obj.in()))) {
                                session_objs_[n].obj = BRUTUS::BrutusSession::_nil();
                                logon_objs_[n].obj = BRUTUS::BrutusLogOn::_nil();
                                continue;
                        }

                        // get proxy load
                        try {
                                proxy_load = session_objs_[n].obj->getProxyCount();
                                if (!found) 
                                        found = true;
                                if (proxy_load < proxy_load_min) 
                                        least_loaded_proxy = n;
                        } catch (...) {
                                session_objs_[n].obj = BRUTUS::BrutusSession::_nil();
                                logon_objs_[n].obj = BRUTUS::BrutusLogOn::_nil();
                                continue;
                        }
                }
                if (!found)
                        throw (CORBA::INITIALIZE());
                else // permanent forwarding
                        throw (PortableInterceptor::ForwardRequest(logon_objs_[least_loaded_proxy].obj.in()));
        }
}

void ServerRequestInterceptor::receive_request(PortableInterceptor::ServerRequestInfo_ptr ri
                                               ACE_ENV_ARG_DECL_NOT_USED)
        ACE_THROW_SPEC((CORBA::SystemException,
                        PortableInterceptor::ForwardRequest))
{
}

void ServerRequestInterceptor::send_reply(PortableInterceptor::ServerRequestInfo_ptr ri
                                          ACE_ENV_ARG_DECL_NOT_USED)
        ACE_THROW_SPEC((CORBA::SystemException))
{
}

void ServerRequestInterceptor::send_exception(PortableInterceptor::ServerRequestInfo_ptr ri
                                              ACE_ENV_ARG_DECL_NOT_USED)
        ACE_THROW_SPEC((CORBA::SystemException,
                        PortableInterceptor::ForwardRequest))
{
}

void ServerRequestInterceptor::send_other(PortableInterceptor::ServerRequestInfo_ptr ri
                                          ACE_ENV_ARG_DECL)
        ACE_THROW_SPEC((CORBA::SystemException,
                        PortableInterceptor::ForwardRequest))
{
}

        
        
