/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Authors: Jules Colding <colding@omesc.com>
 *
 *    Camel Brutus store class.
 *
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of
 *    the License, or (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* system includes */
#include <string.h>
#include <uuid/uuid.h>
#include <orbit/orbit.h>
#include <sys/time.h>
#include <glib.h>

/* Camel includes */
#include <camel/camel-i18n.h>
#include <camel/camel-session.h>
#include <camel/camel-store.h>
#include <camel/camel-file-utils.h>
#include <camel/camel-exception.h>

/* camel-brutus includes */
#include "camel-brutus-private.h"
#include "camel-brutus-store.h"
#include "camel-brutus-folder.h"
#include "camel-brutus-folder-summary.h"

/* pure brutus includes */
#include <idl_output/BrutusLogOn.h>
#include <idl_output/return_codes.h>
#include <idl_output/types.h>
#include <idl_output/uid.h>
#include <servant_impl/BrutusCheck_impl.c>

/* evolution-brutus includes */
#include <server/brutus_conv.h>
#include <server/brutus_mapi.h>
#include <server/brutus_corba.h>
#include <server/brutus_string.h>
#include <server/brutus.h>
#include <session/brutusd.h>

static CamelOfflineStoreClass *parent_class = NULL;
extern CamelServiceAuthType camel_brutus_password_authtype;
#ifdef BRUTUS_HAVE_ENCRYPTED_AUTH
extern CamelServiceAuthType camel_brutus_encrypted_authtype;
#endif

/* ORB run thread arguments */
struct orb_thread_arg {
	struct _CamelBrutusStorePrivate *priv;
	char *brutus_server;
	const guint16 brutus_port;
};

/******************************
 *   static helper functions  *
 *****************************/

#ifdef BRUTUS_DEBUG
static void
brutus_print_folder_tree(CamelFolderInfo *top)
{
	int n = 0;
	int level = 0;
	CamelFolderInfo *fi = NULL;

	if (!top)
		return;

	if (!getenv(ENV_DEBUG))
		return;

	fi = top;
	while (TRUE) {
		fi = fi->parent;
		if (fi)
			level++;
		else break;
	}

	fi = top;
	printf("\n");
	while (TRUE) {
		if (level >= 1) {
			for (n = 0; n < level-1; n++)
				printf("\t");
			printf("|----- ");
		}

		if (!fi->child)
			printf("%s\n", fi->name);
		else {
			printf("%s", fi->name);
			brutus_print_folder_tree(fi->child);
		}

		fi = fi->next;
		if (!fi) {
			printf("\n");
			break;
		}
	}
}
#else
#define brutus_print_folder_tree(__top__) G_STMT_START { } G_STMT_END
#endif

/*
 * Will add a CamelFolderInfo to the hash tables.
 * Must be called with the folder_tree_mutex hold.
 */
static void
brutus_add_to_hash_tables(CamelBrutusStorePrivate *priv,
			  CamelFolderInfo *info,
			  BRUTUS_ENTRYID *eid)
{

	g_hash_table_insert(priv->fi_from_entryid_hash,
			    (gpointer)eid,
			    (gpointer)info);
	g_hash_table_insert(priv->fi_from_full_name_hash,
			    (gpointer)info->full_name,
			    (gpointer)info);
	g_hash_table_insert(priv->entryid_from_full_name_hash,
			    (gpointer)info->full_name,
			    (gpointer)brutus_entryid_dup(eid));
}

static void
brutus_hash_table_data_free(gpointer data)
{
	CORBA_free(data);
}

/*
 * Will free top and all sub-folders.
 * Must be called with the folder_tree_mutex hold.
 */
static void
brutus_free_folder_tree(CamelFolderInfo *top,
			GHashTable *fi_from_entryid,
			GHashTable *fi_from_full_name,
			GHashTable *entryid_from_full_name)
{
}

/*
 * Builds a store URI
 */
static gchar*
brutus_build_uri(const gchar *base_url,
		 const gchar *full_name,
		 CamelException *ex)
{
	CamelURL *url = NULL;
	gchar *uri = NULL;

	url = camel_url_new(base_url, ex);
	if (!url)
		return NULL;

	g_free(url->path);
	url->path = g_strdup_printf("/%s", full_name ? full_name : "");
	uri = camel_url_to_string(url, CAMEL_URL_HIDE_ALL);
	camel_url_free(url);

	return uri;
}

/* no lock needed by caller */
static gboolean
brutus_is_connected(CamelBrutusStorePrivate *priv)
{
	gboolean retv = FALSE;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil((CORBA_Object)priv->mapi_session, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (!cb) {
		BRUTUS_IMAPISession_ping(priv->mapi_session, ev);
		if (!ORBIT2_EX(ev))
			retv = TRUE; // we are connected
	}

out:
	CORBA_exception_free(ev);
	return retv;
}

/*
 * Must be called with the folder_tree_mutex lock acquired
 */
static CamelFolderInfo*
brutus_build_root_folder_info(CamelStore *store,
			      BRUTUS_IMAPIFolder folder,
			      CamelFolderInfo *parent,
			      CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelFolderInfo *info = g_new0(CamelFolderInfo, 1);

	info->next = NULL; // there is no next for root folders
	info->parent = parent;
	info->child = NULL; // to be created later

	info->name = brutus_get_property_as_string((BRUTUS_IMAPIProp)folder, BRUTUS_BRUTUS_PR_DISPLAY_NAME);
	if (!info->name || (*info->name == '\0')) {
		free(info->name);
		info->name = g_strdup("Root Folder");
	}
	info->full_name = g_strdup_printf("%s/%s", parent->full_name, info->name);
	info->uri = brutus_build_uri(priv->base_url, info->full_name, ex);

	info->flags = CAMEL_FOLDER_CHILDREN | CAMEL_FOLDER_SYSTEM;
	info->unread = 0;
	info->total = 0;

	return info;
}

/*
 * Must be called with the folder_tree_mutex lock acquired. The child sub-tree
 * is populated by this function. Everything else must be set prior to this
 * function. "parent_folder" is the parent to all folders added here. "this" is the parent
 * CamelFolderInfo to all CamelFolderInfos created here.
 */
static void
brutus_build_folder_hierachy(CamelStore *store,
			     BRUTUS_IMAPIFolder parent_folder,
			     CamelFolderInfo *this,
			     gboolean release_folder,
			     CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelFolderInfo *previous_info = NULL;
	CamelFolderInfo *current_info = NULL;
	char *tmp = NULL;
	gchar *tmp_name = NULL;
	gboolean name_mangled = FALSE;
	int mangle_level = 1;
	gchar *mangled_name = NULL;
	CORBA_boolean cb = CORBA_FALSE;
	CORBA_long n = 0;
	CORBA_unsigned_long r = 0;
	BRUTUS_SRowSet *row_set = NULL;
	BRUTUS_SPropValue *pv = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_IMAPITable hierachy = CORBA_OBJECT_NIL;
	BRUTUS_IMAPIFolder tmp_folder = CORBA_OBJECT_NIL;
	BRUTUS_BDEFINE obj_type;
	CORBA_string iid = CORBA_string_dup("");
	BRUTUS_ENTRYID *eid = NULL;
	CORBA_Environment ev[1];
	static BRUTUS_BDEFINE props[7] = {
		BRUTUS_BRUTUS_PR_ENTRYID,
		BRUTUS_BRUTUS_PR_DISPLAY_NAME,
		BRUTUS_BRUTUS_PR_SUBFOLDERS,
		BRUTUS_BRUTUS_PR_FOLDER_TYPE,
		BRUTUS_BRUTUS_PR_STATUS,
		BRUTUS_BRUTUS_PR_CONTENT_COUNT,
		BRUTUS_BRUTUS_PR_CONTENT_UNREAD
	};
	static BRUTUS_SPropTagArray proptag_array = {
		._maximum = 7,
		._length = 7,
		._buffer = props,
	};
	static BRUTUS_SSortOrderSet sort_order = {
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};

	if (!this || !store)
		goto out;
	this->child = NULL;

	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil(parent_folder, ev);
	if (!ORBIT2_EX(ev) && cb)
		goto out;

	// get hierachy table
	br = BRUTUS_IMAPIFolder_GetHierarchyTable(parent_folder, 0, &hierachy, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto out;

	// get all rows
	br = brutus_QueryAllRows(hierachy,
				 &proptag_array,
				 CORBA_OBJECT_NIL,
				 &sort_order,
				 0,
				 &row_set,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto out;
	cb = CORBA_Object_is_nil(hierachy, ev);
	if (!ORBIT2_EX(ev) && !cb) {
		BRUTUS_IMAPITable_Destroy(hierachy, 0, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_Object_release(hierachy, ev);
		hierachy = CORBA_OBJECT_NIL;
	}
	if (!row_set->_length)
		goto out;

	// get data for all the parent_folder children
	previous_info = NULL;
	current_info = g_new0(CamelFolderInfo, 1);
	for (r = 0; r < row_set->_length; r++) {

		// in case of a previous rollback
		if (!current_info)
			current_info = g_new0(CamelFolderInfo, 1);

		/*
		 * link the tree
		 */

		if (!previous_info)
			this->child = current_info;
		current_info->parent = this;
		current_info->child = NULL;
		current_info->next = NULL;

		/*
		 * strings
		 */

		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_DISPLAY_NAME);
		if (-1 == n) {
			d("rollback");
			goto rollback;
		}
		current_info->name = g_strdup(row_set->_buffer[r]._buffer[n].Value._u.lpszA);

		/* mangle names containing '/' */
		tmp = strchr(current_info->name, '/');
		if (tmp) {
			name_mangled = TRUE;
			tmp_name = g_strdup(current_info->name);
			tmp = strchr(tmp_name, '/');
			*(tmp++) = '~';
			*tmp = '\0';
			mangled_name = g_strdup_printf("%s%d", tmp_name, mangle_level++); 
			g_free(tmp_name);
		} else {
			name_mangled = FALSE;
			mangled_name = current_info->name;
		}
		current_info->full_name = g_strdup_printf("%s/%s", current_info->parent->full_name, mangled_name);
		if (name_mangled)
			g_free(mangled_name);
		current_info->uri = brutus_build_uri(priv->base_url, current_info->full_name, ex);

		/*
		 * counts
		 */

		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_CONTENT_UNREAD);
		if (-1 == n)
			current_info->unread = -1;
		else
			current_info->unread = (guint32)row_set->_buffer[r]._buffer[n].Value._u.l;

		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_CONTENT_COUNT);
		if (-1 == n)
			current_info->total = -1;
		else
			current_info->total = (guint32)row_set->_buffer[r]._buffer[n].Value._u.l;

		/*
		 * flags
		 */

		current_info->flags = 0;

		/* mark folders containing '/' characters */
		if (name_mangled)
			current_info->flags |= CAMEL_FOLDER_NOSELECT | CAMEL_FOLDER_SYSTEM;

		/* children? */
		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_SUBFOLDERS);
		if (-1 == n) {
			d("rollback");
			goto rollback;
		}
		if ((guint32)row_set->_buffer[r]._buffer[n].Value._u.b)
			current_info->flags |= CAMEL_FOLDER_CHILDREN;
		else
			current_info->flags |= CAMEL_FOLDER_NOCHILDREN;

		/* search result folder? */
		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_FOLDER_TYPE);
		if (-1 == n) {
			d("rollback");
			goto rollback;
		}
		if (row_set->_buffer[r]._buffer[n].Value._u.l & BRUTUS_BRUTUS_FOLDER_SEARCH)
			current_info->flags |= CAMEL_FOLDER_VIRTUAL;

		/* hidden system folder? */
		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_STATUS);
		if (-1 != n) {
			if (row_set->_buffer[r]._buffer[n].Value._u.l & BRUTUS_BRUTUS_FLDSTATUS_HIDDEN) {
#ifdef BRUTUS_SPY
				current_info->flags |= CAMEL_FOLDER_SYSTEM;
#else
				d("rollback");
				goto rollback; // do not show hidden folders unless spying is on
#endif
			}
		}

		/*
		 * hash tables
		 */

		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ENTRYID);
		if (-1 == n) {
			d("rollback");
			goto rollback;
		}

		// get long-term ENTRYID and register it in the hash table
		brutus_sbinary_to_entryid(&row_set->_buffer[r]._buffer[n].Value._u.bin, &eid);
		pv = brutus_open_object_get_one_prop(CORBA_OBJECT_NIL,
						     parent_folder,
						     eid,
						     BRUTUS_BRUTUS_PR_ENTRYID);
		CORBA_free(eid);
		brutus_sbinary_to_entryid(&pv->Value._u.bin, &eid);
		CORBA_free(pv);
		brutus_add_to_hash_tables(priv, current_info, eid);

		// add new info struct
		if (previous_info)
			previous_info->next = current_info;

		if (r == (row_set->_length - 1))
			break;

		previous_info = current_info;
		current_info->next = g_new0(CamelFolderInfo, 1);
		current_info = current_info->next;


		continue;

	rollback:
		d("*************  ROLLBACK ROLLBACK ROLLBACK  **************");
		g_free(current_info->name);
		g_free(current_info->full_name);
		g_free(current_info->uri);
		g_free(current_info);
		current_info = NULL;
		if (!previous_info)
			this->child = NULL;
	}
	CORBA_free(row_set);
	row_set = NULL;

	/* recursively build the child nodes */
	current_info = this->child;
	while (current_info) {
		if (current_info->flags & CAMEL_FOLDER_NOCHILDREN)
			goto next_sibling;

		eid = (BRUTUS_ENTRYID*)g_hash_table_lookup(priv->entryid_from_full_name_hash, current_info->full_name);
		br = BRUTUS_IMAPIFolder_OpenEntry(parent_folder,
						  eid,
						  iid,
						  BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						  &obj_type,
						  &tmp_folder,
						  ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto next_sibling;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			goto next_sibling;

		// will relase tmp_folder
		brutus_build_folder_hierachy(store, tmp_folder, current_info, TRUE, ex);

	next_sibling:
		current_info = current_info->next;
	}

out:
	CORBA_free(row_set);

	/* release hierachy table */
	cb = CORBA_Object_is_nil(hierachy, ev);
	if (!ORBIT2_EX(ev) && !cb) {
		BRUTUS_IMAPITable_Destroy(hierachy, 0, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_Object_release(hierachy, ev);
	}

	/* release input folder */
	if (release_folder) {
		cb = CORBA_Object_is_nil(parent_folder, ev);
		if (!ORBIT2_EX(ev) && !cb) {
			BRUTUS_IMAPIFolder_Destroy(parent_folder, 0, ev);
			PRINT_IF_ORBIT2_EX(ev);
			CORBA_Object_release(parent_folder, ev);
		}
	}
	CORBA_exception_free(ev);

	return;
}


/*
 * Must be called with the folder_tree_mutex lock acquired
 */
static CamelFolderInfo*
brutus_build_folder_tree(CamelStore *store,
			 CamelException *ex)
{
	guint n = 0;
	char *tmp = NULL;
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelFolderInfo *top;
	CamelFolderInfo *current = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_IMAPIFolder folder;
	BRUTUS_IMsgStore msg_store = CORBA_OBJECT_NIL;
	BRUTUS_ENTRYID *eid = NULL;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	top = NULL;

	cb = brutus_is_connected(priv);
	if (!cb) {
		camel_exception_set(ex,
				    CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
				    _("Not connected."));
		goto out;
	}
	if (!priv->msg_stores->len)
		goto out;

	top = g_new0(CamelFolderInfo, 1);
	priv->folder_tree = top;

	/* initialize top */
	top->next = NULL;
	top->parent = NULL;
	top->child = NULL;
	top->uri = brutus_build_uri(priv->base_url, NULL, ex);
	top->name = g_strdup(top->uri);
	top->full_name = g_strdup(top->uri);
	top->flags = CAMEL_FOLDER_NOSELECT | CAMEL_FOLDER_SYSTEM;
	top->unread = -1;
	top->total = -1;

	/* start by populating the message store infos */
	top->child = g_new0(CamelFolderInfo, 1);
	current = top->child;
	for (n = 0; n < priv->msg_stores->len; n++) {
		current->next = NULL;
		current->parent = top;
		current->child = NULL;

		// do not use unappropiate message stores unless we are in spy mode
#ifndef BRUTUS_SPY
		// we really need a unique ENTRYID
		if (!(BRUTUS_BRUTUS_STORE_ENTRYID_UNIQUE & ((struct BrutusMsgStore*)g_ptr_array_index(priv->msg_stores, n))->support_mask)) {
			d("ENTRYIDs not unique in this store");
			continue;
		}
#endif

		// the store
		msg_store = ((struct BrutusMsgStore*)g_ptr_array_index(priv->msg_stores, n))->msg_store;

		current->name = brutus_get_property_as_string((BRUTUS_IMAPIProp)msg_store, BRUTUS_BRUTUS_PR_DISPLAY_NAME);
		current->full_name = g_strdup(current->name);
		current->uri = brutus_build_uri(priv->base_url, current->full_name, ex);

		/* determine message store type */
		current->flags = CAMEL_FOLDER_CHILDREN | CAMEL_FOLDER_SYSTEM;
		tmp = brutus_get_guid_str((BRUTUS_IMAPIProp)msg_store, BRUTUS_BRUTUS_PR_MDB_PROVIDER);
		do {
			if (!g_ascii_strncasecmp(tmp, BRUTUS_BRUTUS_pbExchangeProviderPrimaryUserGuid, strlen(BRUTUS_BRUTUS_pbExchangeProviderPrimaryUserGuid)))
				break;

			if (!g_ascii_strncasecmp(tmp, BRUTUS_BRUTUS_pbExchangeProviderDelegateGuid, strlen(BRUTUS_BRUTUS_pbExchangeProviderDelegateGuid))) {
				current->flags |= CAMEL_FOLDER_SHARED_TO_ME;
				break;
			}
			if (!g_ascii_strncasecmp(tmp, BRUTUS_BRUTUS_pbExchangeProviderPublicGuid, strlen(BRUTUS_BRUTUS_pbExchangeProviderPublicGuid))) {
				current->flags |= CAMEL_FOLDER_SHARED_TO_ME;
				break;
			}
			if (!g_ascii_strncasecmp(tmp, BRUTUS_BRUTUS_pbExchangeProviderXportGuid, strlen(BRUTUS_BRUTUS_pbExchangeProviderXportGuid))) {
				current->flags = CAMEL_FOLDER_NOSELECT;
				break;
			}
			d("unknown message store type");
			break;
		} while (0);
		free(tmp);

		current->unread = 0;
		current->total = 0;

		if (n == (priv->msg_stores->len - 1))
			break;

		/* add to the hash tables */
		eid = brutus_get_entryid((BRUTUS_IMAPIProp)msg_store, BRUTUS_BRUTUS_PR_ENTRYID);
		if (!eid) {
			g_free(current->uri);
			g_free(current->name);
			g_free(current->full_name);
			continue;
		}
		brutus_add_to_hash_tables(priv, current, eid);

		current->next = g_new0(CamelFolderInfo, 1);
		current = current->next;
	}

	/* now get all root folders, and their folder hierachy, for all message stores, starting with the first child to top */
	current = top->child;
	for (n = 0; n < priv->msg_stores->len; n++) {

#ifndef BRUTUS_SPY
		if (!(BRUTUS_BRUTUS_STORE_ENTRYID_UNIQUE & ((struct BrutusMsgStore*)g_ptr_array_index(priv->msg_stores, n))->support_mask))
			continue;
#endif
		msg_store = ((struct BrutusMsgStore*)g_ptr_array_index(priv->msg_stores, n))->msg_store;

		br = brutus_get_root_folder(msg_store, &folder, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			continue;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			continue;

		eid = brutus_get_entryid((BRUTUS_IMAPIProp)folder, BRUTUS_BRUTUS_PR_ENTRYID);
		if (!eid)
			continue;

		current->child = brutus_build_root_folder_info(store, folder, current, ex);

		/* add to the hash tables */
		brutus_add_to_hash_tables(priv, current->child, eid);

		/* will release the folder */
		brutus_build_folder_hierachy(store, folder, current->child, TRUE, ex);

		current = current->next;
	}

out:
	CORBA_exception_free(ev);
	brutus_print_folder_tree(priv->folder_tree);

	return top;
}

/* only invoked from brutus_connect(), when the store is connect locked */
static void
brutus_load_mapi_stores(CamelBrutusStorePrivate *priv)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_SRowSet *row_set = NULL;
	BRUTUS_ENTRYID *eid = NULL;
	CORBA_unsigned_long n = 0;
	struct BrutusMsgStore *msg_store_struct = NULL;
	BRUTUS_SPropValue *pv = NULL;
	CORBA_string iid = CORBA_string_dup(""); // use default interface (IMsgStore)
	CORBA_Environment ev[1];
	BRUTUS_BDEFINE props[1] = { // Only retrieve entry id.
		BRUTUS_BRUTUS_PR_ENTRYID,
	};
	BRUTUS_SPropTagArray proptag_array = {
		._maximum = 1,
		._length = 1,
		._buffer = props,
	};
	BRUTUS_SSortOrderSet sort_order = { 
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};

	CORBA_exception_init(ev);

	// get the single row
	br = brutus_QueryAllRows(priv->mapi_msg_store_table,
				 &proptag_array,
				 CORBA_OBJECT_NIL,
				 &sort_order,
				 0,
				 &row_set,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	priv->msg_stores = g_ptr_array_new();
	for (n = 0; n < row_set->_length; n++) {
		msg_store_struct = g_new(struct BrutusMsgStore, 1);
		if (!msg_store_struct) {
			d("No memory");
			goto exit;
		}
		msg_store_struct->msg_store = CORBA_OBJECT_NIL;
		msg_store_struct->support_mask = 0;

		brutus_sbinary_to_entryid(&row_set->_buffer[n]._buffer[0].Value._u.bin, &eid);

		// Open the first returned (default) message store
		br = BRUTUS_IMAPISession_OpenMsgStore(priv->mapi_session,
						      (const BRUTUS_ENTRYID*)eid,
						      iid,
						      BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						      &msg_store_struct->msg_store,
						      ev);
		CORBA_free(eid);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
		} else if (BRUTUS_BRUTUS_S_OK == br) {
			br = brutus_GetOneProp(msg_store_struct->msg_store,
					       BRUTUS_BRUTUS_PR_STORE_SUPPORT_MASK,
					       &pv,
					       ev);
			if (BRUTUS_BRUTUS_S_OK == br) 
				msg_store_struct->support_mask = pv->Value._u.l;
			if (pv) {
				CORBA_free(pv);
				pv = NULL;
			}
			g_ptr_array_add(priv->msg_stores, (gpointer)msg_store_struct);
		}
	}

exit:
	CORBA_exception_free(ev);
	CORBA_free(iid);
	CORBA_free(row_set);

	return;
}

/* only invoked from brutus_(dis)connect() and from finalize() when the store is connect locked */
static void
brutus_mapi_shutdown(CamelBrutusStore *brutus_store)
{
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	/* message store table */
	cb = CORBA_Object_is_nil(priv->mapi_msg_store_table, ev);
	if (!ORBIT2_EX(ev) && !cb) {
		BRUTUS_IMAPITable_Destroy(priv->mapi_msg_store_table, 0, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_Object_release(priv->mapi_msg_store_table, ev);
	}
	priv->mapi_msg_store_table = CORBA_OBJECT_NIL;

	/* message stores */
	if (priv->msg_stores) {
		guint n = 0;
		BRUTUS_IMsgStore imsg_store = CORBA_OBJECT_NIL;
		for (n = 0; n < priv->msg_stores->len; n++) {
			imsg_store = ((struct BrutusMsgStore*)g_ptr_array_index(priv->msg_stores, n))->msg_store;
			cb = CORBA_Object_is_nil(imsg_store, ev);
			if (!ORBIT2_EX(ev) && !cb) {
				BRUTUS_IMsgStore_Destroy(imsg_store, 0, ev);
				PRINT_IF_ORBIT2_EX(ev);

				CORBA_Object_release(imsg_store, ev);
			}
			g_free(g_ptr_array_index(priv->msg_stores, n));
		}
		g_ptr_array_free(priv->msg_stores, TRUE);
		priv->msg_stores= NULL;
	}

	/* must reload the entire tree when getting folder info */
	BRUTUS_PRIV_LOCK(priv, folder_tree_mutex);
	if (priv->folder_tree)
		brutus_free_folder_tree(priv->folder_tree,
					priv->fi_from_entryid_hash,
					priv->fi_from_full_name_hash,
					priv->entryid_from_full_name_hash);
	g_free(priv->folder_tree);
	priv->folder_tree = NULL;
	BRUTUS_PRIV_UNLOCK(priv, folder_tree_mutex);

	/* mapi session and instance ID */
	cb = CORBA_Object_is_nil(priv->mapi_session, ev);
	if (!ORBIT2_EX(ev) && !cb) {
		BRUTUS_IMAPISession_Destroy(priv->mapi_session, priv->instance_id, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_Object_release(priv->mapi_session, ev);
	}
	priv->mapi_session = CORBA_OBJECT_NIL;
	priv->instance_id = 0;

	CORBA_exception_free(ev);
}

static void
brutus_free_lifeline(BRUTUS_BrutusProxy proxy,
		     struct BrutusCheckProxy *lifeline)
{
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil((CORBA_Object)lifeline->wrapper, ev);
	if (!ORBIT2_EX(ev) && !cb) {
		BRUTUS_BrutusProxy_killWrapper(proxy, lifeline->wrapper, ev);
		CORBA_Object_release((CORBA_Object)lifeline->wrapper, ev);
	}

	cb = CORBA_Object_is_nil((CORBA_Object)lifeline->filling, ev);
	if (!ORBIT2_EX(ev) && !cb)
		CORBA_Object_release((CORBA_Object)lifeline->filling, ev);

	CORBA_exception_free(ev);
	lifeline->wrapper = CORBA_OBJECT_NIL;
	lifeline->filling = CORBA_OBJECT_NIL;
}

// ORB run thread. MUST call ORB_init()!!
static gpointer
orb_run_thread_func(gpointer data)
{
	CORBA_boolean cb = CORBA_TRUE;
	CamelBrutusStorePrivate *priv = ((struct orb_thread_arg*)data)->priv;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	d("Invoking ORB_Init()");
	priv->orb = brutus_create_orb(0,
				      NULL,
				      priv->use_ssliop,
				      "BrutusLogOn",
				      BRUTUS_BrutusLogOn_IOR_TABLE_KEY,
				      ((struct orb_thread_arg*)data)->brutus_server,
				      ((struct orb_thread_arg*)data)->brutus_port,
				      "Brutus-CamelStore-ORB_orbit-io-thread",
				      ev);
	g_free(((struct orb_thread_arg*)data)->brutus_server);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out_no_orb;
	}
	cb = CORBA_Object_is_nil((CORBA_Object) priv->orb, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out_no_orb;
	}
	if (cb) {
		d("Could not initialize the ORB");
		goto out_no_orb;
	}

	d("Getting root POA");
	priv->root_poa = brutus_get_root_poa(priv->orb, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	cb = CORBA_Object_is_nil((CORBA_Object) priv->root_poa, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (cb) {
		d("Could not get the root POA");
		goto out;
	}

	// signal to the main thread that the ORB is initiated and the Root POA aquired
	g_mutex_lock(priv->proceed_orb_ok_mutex);
	priv->corba_ready = TRUE;
	g_cond_signal(priv->proceed_orb_ok);
	g_mutex_unlock(priv->proceed_orb_ok_mutex);

	d("Running the ORB event loop");
	CORBA_ORB_run(priv->orb, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}

out:
	d("ORB shutdown in progress");

	// wait for destroy signal
	g_mutex_lock(priv->proceed_orb_destroy_mutex);
	while (!priv->orb_destroy)
		g_cond_wait(priv->proceed_orb_destroy, priv->proceed_orb_destroy_mutex);
	g_mutex_unlock(priv->proceed_orb_destroy_mutex);

	CORBA_Object_release((CORBA_Object)priv->orb, ev);
	PRINT_IF_ORBIT2_EX(ev);

out_no_orb:
	CORBA_exception_free(ev);

	d("ORB is down");
	return NULL;
}

static void
brutus_get_password(CamelService *service,
		    const char *msg,
		    CamelException *ex)
{
	CamelSession *session = camel_service_get_session(service);
	char *prompt = NULL;

	/* forget any previously cached password before prompting again */
	if (service->url->passwd) {
		camel_session_forget_password(session, service, "Brutus", "password", ex);
		g_free(service->url->passwd);
		service->url->passwd = NULL;
	}

	/* make a nifty prompt string */
	prompt = g_strdup_printf(_("%sPlease enter password for user \"%s\" at the \"%s\" Exchange mailbox "),
				 msg ? msg : "",
				 service->url->user,
				 camel_url_get_param(service->url, "exchange_mailbox"));

	service->url->passwd = camel_session_get_password(session,
							  service,
							  "Brutus",
							  prompt,
							  "password",
							  CAMEL_SESSION_PASSWORD_SECRET,
							  ex);
	if (camel_exception_is_set(ex))
		dfmt("Camel exception: %s", camel_exception_get_description(ex));

	if (!service->url->passwd) {
		if (!camel_exception_is_set(ex)) {
			camel_exception_set(ex, CAMEL_EXCEPTION_USER_CANCEL,
					    _("User canceled"));
		}
	}

	g_free(prompt);

	return;
}

/* The store must be locked by caller */
static gboolean
brutus_logon_loop(BRUTUS_BrutusLogOn brutus_logon,
		  char **mapi_profile,
		  const char *windows_user,
		  const char *windows_domain,
		  const char *exchange_mailbox,
		  const char *exchange_server,
		  CamelService *service,
		  CamelException *ex)
{
	CamelBrutusStore *store = CAMEL_BRUTUS_STORE(service);
	CamelBrutusStorePrivate *priv = store->priv;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_MAPI_E_LOGON_FAILED;
	const char *password = NULL;
	char *err = g_strdup(_("Authentication attempt failed - "));
	int c = 0;
	gboolean retv = FALSE;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	/* get the password */
	password = service->url->passwd;
	if (!password) {
		brutus_get_password(service, NULL, ex);
		if (camel_exception_is_set(ex))
			goto out;
		password = service->url->passwd;
	}

	do { // loop indefinitely if password is bad, break for other errors
		br = BRUTUS_BrutusLogOn_Logon(brutus_logon,
					      priv->lifeline.wrapper,
					      (const char*)(*mapi_profile),
					      "",
					      windows_user,
					      windows_domain,
					      password,
					      0,
					      BRUTUS_BRUTUS_MAPI_EXPLICIT_PROFILE,
					      &priv->instance_id,
					      &priv->mapi_session,
					      ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
		if (BRUTUS_BRUTUS_MAPI_E_LOGON_FAILED == br) {
			brutus_get_password(service, err, ex);
			if (camel_exception_is_set(ex))
				goto out;
			password = service->url->passwd;
			continue;
		}
		if (BRUTUS_BRUTUS_MAPI_E_NOT_FOUND == br) {
			d("Could not logon with provided profile - creating new MAPI profile");
			goto create_profile;
		}

		break;
	} while (TRUE);
	if (BRUTUS_BRUTUS_S_OK == br) {
		retv = TRUE;
		goto out;
	}

	dfmt("BrutusLogOn->Logon() returned %s", brutus_bresult_to_str(br));
	if (BRUTUS_BRUTUS_MAPI_E_NOT_INITIALIZED == br)
		goto out;

	c = 0;
	do {
	create_profile:
		br = BRUTUS_BrutusLogOn_CreateProfile(brutus_logon,
						      (const char*)(*mapi_profile),
						      "",
						      exchange_mailbox,
						      exchange_server,
						      0,
						      ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
		dfmt("BrutusLogOn->CreateProfile() returned %s", brutus_bresult_to_str(br));
		if (BRUTUS_BRUTUS_S_OK == br)
			break;

		if (BRUTUS_BRUTUS_MAPI_E_NO_ACCESS == br) { // should never happen...
			char guid_str[37] = { '\0' };
			uuid_t guid = { 0 };

			d("MAPI profile name collision, creating a new one...");
			uuid_generate(guid);
			uuid_unparse(guid, guid_str);
			CORBA_free(*mapi_profile);
			*mapi_profile = CORBA_string_dup(guid_str);

			dfmt("The new MAPI profile is \"%s\"", *mapi_profile);
		}

		c++;
	} while (c < 2);
	if (BRUTUS_BRUTUS_S_OK != br) {
		d("No MAPI profile could be created - check server permissions");
		goto out;
	}

	dfmt("Attempting logon with MAPI profile \"%s\"", *mapi_profile);
	do {
		br = BRUTUS_BrutusLogOn_Logon(brutus_logon,
					      priv->lifeline.wrapper,
					      (const char*)(*mapi_profile),
					      "",
					      windows_user,
					      windows_domain,
					      password,
					      0,
					      BRUTUS_BRUTUS_MAPI_EXPLICIT_PROFILE,
					      &priv->instance_id,
					      &priv->mapi_session,
					      ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
		if (BRUTUS_BRUTUS_MAPI_E_LOGON_FAILED == br) {
			brutus_get_password(service, err, ex);
			if (camel_exception_is_set(ex))
				goto out;
			password = service->url->passwd;
		} else
			break;
	} while (TRUE);
	if (BRUTUS_BRUTUS_S_OK == br) {
		dfmt("\"%s\" has logged on", windows_user);
		retv = TRUE;
	}

	dfmt("Connected to IMAPISession: %s", brutus_is_connected(priv)? "YES" : "NO");

out:
	g_free(err);
	CORBA_exception_free(ev);

	if (!retv && !camel_exception_is_set(ex)) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not connect to Brutus server"));
	}

	return !camel_exception_is_set(ex);
}

/****************************************
 *   service virtual method overloads   *
 ***************************************/

static void brutus_construct(CamelService *service,
			     CamelSession *session,
			     CamelProvider *provider,
			     CamelURL *url,
			     CamelException *ex);

static GList *brutus_query_auth_types(CamelService *service,
				      CamelException *ex);

static char *brutus_get_name(CamelService *service,
			     gboolean brief);

static gboolean brutus_connect(CamelService *service,
			       CamelException *ex);

static gboolean brutus_disconnect(CamelService *service,
				  gboolean clean,
				  CamelException *ex);

/**************************************
 *   store virtual method overloads   *
 *************************************/

static CamelFolder *brutus_get_folder(CamelStore *store,
				      const char *folder_name,
				      guint32 flags,
				      CamelException *ex);

static CamelFolder *brutus_get_inbox(CamelStore *store,
				     CamelException *ex);

static CamelFolder *brutus_get_trash(CamelStore *store,
				     CamelException *ex);

static CamelFolder *brutus_get_junk(CamelStore *store,
				    CamelException *ex);

static CamelFolderInfo *brutus_create_folder(CamelStore *store,
					     const char *parent_name,
					     const char *folder_name,
					     CamelException *ex);

static void brutus_delete_folder(CamelStore *store,
				 const char *folder_name,
				 CamelException *ex);

static void brutus_rename_folder(CamelStore *store,
				 const char *old_name,
				 const char *new_name,
				 CamelException *ex);

static void brutus_sync(CamelStore *store,
			int expunge,
			CamelException *ex);

static CamelFolderInfo *brutus_get_folder_info(CamelStore *store,
					       const char *top,
					       guint32 flags,
					       CamelException *ex);

static gboolean brutus_folder_subscribed(CamelStore *store,
					 const char *folder_name);

static void brutus_subscribe_folder(CamelStore *store,
				    const char *folder_name,
				    CamelException *ex);

static void brutus_unsubscribe_folder(CamelStore *store,
				      const char *folder_name,
				      CamelException *ex);

/**********************************************
 ************ IMPLEMENTATIONS *****************
 *********************************************/

static void
camel_brutus_store_class_init(CamelBrutusStoreClass *camel_brutus_store_class)
{
	CamelServiceClass *camel_service_class = CAMEL_SERVICE_CLASS(camel_brutus_store_class);
	CamelStoreClass *camel_store_class = CAMEL_STORE_CLASS(camel_brutus_store_class);

	// initiate gthreads
	if (!g_thread_supported())
		g_thread_init(NULL);

	parent_class = CAMEL_OFFLINE_STORE_CLASS(camel_type_get_global_classfuncs(camel_offline_store_get_type()));

	/* service virtual method overloading */
	camel_service_class->construct = brutus_construct;
	camel_service_class->query_auth_types = brutus_query_auth_types;
	camel_service_class->get_name = brutus_get_name;
	camel_service_class->connect = brutus_connect;
	camel_service_class->disconnect = brutus_disconnect;

	/* store virtual method overloading */
	camel_store_class->get_folder = brutus_get_folder;
	camel_store_class->get_inbox = brutus_get_inbox;
	camel_store_class->get_trash = brutus_get_trash;
	camel_store_class->get_junk = brutus_get_junk;

	camel_store_class->create_folder = brutus_create_folder;
	camel_store_class->delete_folder = brutus_delete_folder;
	camel_store_class->rename_folder = brutus_rename_folder;

	camel_store_class->sync = brutus_sync;

	camel_store_class->get_folder_info = brutus_get_folder_info;

	camel_store_class->folder_subscribed = brutus_folder_subscribed;
	camel_store_class->subscribe_folder = brutus_subscribe_folder;
	camel_store_class->unsubscribe_folder = brutus_unsubscribe_folder;
}

/* TODO: implement store summary - see the groupwise store */
/* TODO: Enable Exchange Trash and Junk if needed */
static void
brutus_construct(CamelService *service,
		 CamelSession *session,
		 CamelProvider *provider,
		 CamelURL *url,
		 CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(service);
	CamelStore *store = CAMEL_STORE(service);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CORBA_boolean cb = CORBA_TRUE;
	gchar *path = NULL;

	d("in brutus store construct\n");

	CAMEL_SERVICE_CLASS(parent_class)->construct(service, session, provider, url, ex);
	if (camel_exception_is_set(ex))
		return;

	if (!(url->host
	      && url->user
	      && camel_url_get_param(url, "iiop_port")
	      && camel_url_get_param(url, "windows_domain")
	      && camel_url_get_param(url, "exchange_mailbox")
	      && camel_url_get_param(url, "brutus_server"))) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_INVALID,
				    _("Brutus Account not configured - required data not available in url"));
		return;
	}

	/* SSLIOP */
#ifdef BRUTUS_HAVE_SSL
	priv->use_ssliop = camel_url_get_param(url, "use_ssliop") ? TRUE : FALSE;
#endif

	/* base URL */
	priv->base_url = camel_url_to_string(service->url, (CAMEL_URL_HIDE_PASSWORD | CAMEL_URL_HIDE_PARAMS | CAMEL_URL_HIDE_AUTH));
	if (!priv->base_url)
		goto out_no_mem;

	/* storage path */
	priv->storage_path = camel_session_get_storage_path(session, service, ex);
	if (!priv->storage_path)
		goto out_no_mem;

	/* create evolution-brutus subdirectories */	
	path = g_strdup_printf("%s/%s", priv->storage_path, FOLDER_SUMMARIES);
	if (!path) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("No Camel storage path"));
		return;
	}
	if (-1 == camel_mkdir(path, 0700)) { // for folder summaries
		g_free(path);
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not create Camel Folder Summary directory"));
		return;
	}
	g_free(path);
	path = g_strdup_printf("%s/%s", priv->storage_path, FOLDER_MESSAGE_CACHE);
	if (!path) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("No Camel storage path"));
		return;
	}
	if (-1 == camel_mkdir(path, 0700)) { // for the folder message cache
		g_free(path);
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not create Camel Folder Summary directory"));
		return;
	}
	g_free(path);

	/* filter */
	if (camel_url_get_param(url, "filter"))
		store->flags |= CAMEL_STORE_FILTER_INBOX;

	/* hash tables */

	// key is BRUTUS_ENTRYID* - value is CamelFolderInfo*
	priv->fi_from_entryid_hash = g_hash_table_new_full(g_direct_hash,
							   brutus_entryid_cmp,
							   brutus_hash_table_data_free, // key destroy
							   NULL);
	if (!priv->fi_from_entryid_hash)
		goto out_no_mem;

	// key is string - value is CamelFolderInfo*
	priv->fi_from_full_name_hash = g_hash_table_new_full(g_str_hash,
							     g_str_equal,
							     NULL,
							     NULL);
	if (!priv->fi_from_full_name_hash)
		goto out_no_mem;


	// key is string - value is BRUTUS_ENTRYID*
	priv->entryid_from_full_name_hash = g_hash_table_new_full(g_str_hash,
								  g_str_equal,
								  NULL,
								  brutus_hash_table_data_free); // value destroy
	if (!priv->entryid_from_full_name_hash)
		goto out_no_mem;


	/* TODO - Enable Exchange Junk and Trash */
	store->flags &= ~CAMEL_STORE_VJUNK;
	store->flags &= ~CAMEL_STORE_VTRASH;

	d("Initializing CORBA");
	{
		guint64 iiop_port;
		guint64 ssliop_port;

		if (!brutus_get_xbit_number(16, camel_url_get_param(url, "ssliop_port"), &ssliop_port)) {
			camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_INVALID,
					    _("Invalid SSLIOP port in url"));
			return;
		}
		if (!brutus_get_xbit_number(16, camel_url_get_param(url, "iiop_port"), &iiop_port)) {
			camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_INVALID,
					    _("Invalid IIOP port in url"));
			return;
		}


		{
			struct orb_thread_arg thread_args = {
				.priv = priv,
				.brutus_server = g_strdup(camel_url_get_param(url, "brutus_server")),
				.brutus_port = priv->use_ssliop ? (guint16)ssliop_port : (guint16)iiop_port,
			};

			priv->brutus_logon_str = brutus_build_corbaloc_string(thread_args.brutus_server,
									      BRUTUS_BrutusLogOn_IOR_TABLE_KEY,
									      priv->use_ssliop,
									      thread_args.brutus_port);

			priv->orb_thread = g_thread_create((GThreadFunc)orb_run_thread_func,
							   (gpointer)&thread_args,
							   TRUE,
							   NULL);
			if (!priv->orb_thread) {
				camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
						    _("Could not initiate ORB"));
				return;
			}
		}
		// wait for the ORB to initialize
		g_mutex_lock(priv->proceed_orb_ok_mutex);
		while (!priv->corba_ready)
			g_cond_wait(priv->proceed_orb_ok, priv->proceed_orb_ok_mutex);
		g_mutex_unlock(priv->proceed_orb_ok_mutex);
	}

	d("Getting root POAManager, activating the RootPOA and instantiating the lifeline");
	{
		PortableServer_POAManager root_poa_manager = CORBA_OBJECT_NIL;
		gboolean use_local_proxy = !(ORBit_proto_use("IPv4") || ORBit_proto_use("IPv6"));
		CORBA_Environment ev[1];
		CORBA_exception_init(ev);

		root_poa_manager = PortableServer_POA__get_the_POAManager(priv->root_poa, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}

		cb = CORBA_Object_is_nil((CORBA_Object)root_poa_manager, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}
		if (cb) {
			d("Could not get the root POAManager");
			goto out_ex;
		}
		PortableServer_POAManager_activate(root_poa_manager, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}

		CORBA_Object_release((CORBA_Object)root_poa_manager, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}

		// get brutus proxy
		{
			char *home_dir = getenv("HOME");
			gchar *ref_path = g_strconcat(home_dir, "/", BRUTUSD_REF_FILE_NAME, NULL);

			priv->brutus_proxy = brutus_object_ref_from_file(priv->orb, ref_path, ev);
			g_free(ref_path);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_ex;
			}
			cb = CORBA_Object_is_nil((CORBA_Object)priv->brutus_proxy, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_ex;
			}
			if (cb) {
				d("Could not get the Brutus proxy reference");
				goto out_ex;
			}
		}

		// the lifeline
		priv->lifeline.filling = impl_BRUTUS_BrutusCheck__create(priv->root_poa, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto out_ex;
		}
		if (use_local_proxy) {
			d("Using local network capable Brutus proxy");
			BRUTUS_BrutusProxy_wrapObject(priv->brutus_proxy,
						      BRUTUS_BrutusCheck_tc,
						      priv->lifeline.filling,
						      &priv->lifeline.wrapper,
						      ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_ex;
			}
		} else {
			/* this will enable us to use "wrapper" unconditionally */
			priv->lifeline.wrapper = CORBA_Object_duplicate(priv->lifeline.filling, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_ex;
			}
		}
		goto out;

	out_ex:
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not construct the store - CORBA exception caught"));
		goto out;

	out_no_mem:
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not construct the store - No memory"));

	out:
		CORBA_exception_free(ev);
	}
}

static GList*
brutus_query_auth_types(CamelService *service,
			CamelException *ex)
{
	GList *auth_types = NULL;

	d("");
	auth_types = g_list_prepend(auth_types, &camel_brutus_password_authtype
#ifdef BRUTUS_HAVE_ENCRYPTED_AUTH
				    , &camel_brutus_encrypted_authtype
#endif
		);

	return auth_types;
}

static char*
brutus_get_name(CamelService *service,
		gboolean brief)
{
	if (brief) {
		return g_strdup_printf(_("Exchange server %s"),
				       service->url->host);
	} else {
		return g_strdup_printf(_("Exchange account for %s on %s"),
				       service->url->user,
				       service->url->host);
	}
}

static gboolean
brutus_connect(CamelService *service,
	       CamelException *ex)
{
	static gboolean got_time_zones = FALSE;
	CamelBrutusStore *store = CAMEL_BRUTUS_STORE(service);
	CamelBrutusStorePrivate *priv = store->priv;
	BRUTUS_BrutusLogOn brutus_logon = CORBA_OBJECT_NIL;
	CORBA_boolean cb = CORBA_TRUE;
	gboolean retv = FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	BRUTUS_STORE_LOCK(store, connect_mutex);

	/* check connection status */
	retv = brutus_is_connected(priv);
	if (retv)
		goto out; // connected
	else
		brutus_mapi_shutdown(store); // reset connection

	/* we are not connected - proceed */
	d("Attempting to query a session from the proxy");
	{
		d(camel_url_get_param(service->url, "exchange_mailbox"));
		d(service->url->host);
		BRUTUS_BrutusProxy_querySession(priv->brutus_proxy,
						camel_url_get_param(service->url, "exchange_mailbox"),
						service->url->host,
						&priv->mapi_session,
						ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto proceed;
		}

		cb = CORBA_Object_is_nil((CORBA_Object)priv->mapi_session, ev);
		if (!ORBIT2_EX(ev) && !cb) {
			BRUTUS_IMAPISession_ping(priv->mapi_session, ev);
			if (!ORBIT2_EX(ev)) {
				priv->instance_id = BRUTUS_IMAPISession_addLifeline(priv->mapi_session, priv->lifeline.wrapper, ev);
				if (ORBIT2_EX(ev)) {
					d("Could not add lifeline");
					goto proceed;
				}

				d("Query successful");
				retv = TRUE; // we are connected, no query lock to release
				goto out_get_msgstore_tables;
			} else { // could not ping
				CORBA_Object_release((CORBA_Object)priv->mapi_session, ev);
				priv->mapi_session = CORBA_OBJECT_NIL;
			}
		}
	}

proceed:
	d("Could not get session from proxy, must log on explicitly");
	BRUTUS_BrutusProxy_getProfile(priv->brutus_proxy,
				      camel_url_get_param(service->url, "exchange_mailbox"),
				      service->url->host,
				      &priv->mapi_profile,
				      ev);
	if (ORBIT2_EX(ev))
		PRINT_ORBIT2_EX(ev);
	else if (!strlen(priv->mapi_profile)) {
		d("No stored MAPI Profile - generating a new one");
		char guid_str[37] = { 0 };
		uuid_t guid = { 0 };

		CORBA_free(priv->mapi_profile);

		uuid_generate(guid);
		uuid_unparse(guid, guid_str);
		priv->mapi_profile = CORBA_string_dup(guid_str);
	}
	dfmt("Using this MAPI Profile: \"%s\"", priv->mapi_profile);

	d("Getting reference to BrutusLogOn object");
	{
		brutus_logon = (BRUTUS_BrutusLogOn)CORBA_ORB_resolve_initial_references(priv->orb, "BrutusLogOn", ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			brutus_logon = CORBA_OBJECT_NIL;
		}

		cb = CORBA_Object_is_nil(brutus_logon, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			cb = TRUE;
		}
		if (cb) {
			d("Could not resolve (resolve_initial_references) BrutusLogOn, falling back to corbaloc...");
			brutus_logon = (BRUTUS_BrutusLogOn)CORBA_ORB_string_to_object(priv->orb, priv->brutus_logon_str, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto out_release_lock;
			}

			cb = CORBA_Object_is_nil(brutus_logon, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				cb = TRUE;
			}
			if (cb) {
				d("Could not resolve (string_to_object) BrutusLogOn, NIL object returned");
				goto out_release_lock;
			} else
				d("BrutusLogOn resolved");
		}
		d("BrutusLogOn reference has been acquired");
	}

	retv = brutus_logon_loop(brutus_logon,
				 &priv->mapi_profile,
				 service->url->user,
				 camel_url_get_param(service->url, "windows_domain"),
				 camel_url_get_param(service->url, "exchange_mailbox"),
				 service->url->host,
				 service,
				 ex);

out_release_lock:
	if (retv) {
		d("Offering IMAPISession to Brutus proxy");
		BRUTUS_BrutusProxy_offerSession(priv->brutus_proxy,
						priv->mapi_session,
						camel_url_get_param(service->url, "exchange_mailbox"),
						service->url->host,
						priv->mapi_profile,
						ev);
		PRINT_IF_ORBIT2_EX(ev);

	out_get_msgstore_tables:
		d("Getting message store table");
		br = BRUTUS_IMAPISession_GetMsgStoresTable(priv->mapi_session, 0, &priv->mapi_msg_store_table, ev);
		if (!ORBIT2_EX(ev)) {
			if (BRUTUS_BRUTUS_S_OK != br) {
				d("Could not get message store table");
				priv->mapi_msg_store_table = CORBA_OBJECT_NIL;
			} else 
				brutus_load_mapi_stores(priv);
		} else
			PRINT_ORBIT2_EX(ev);
	} else {

		d("No IMAPISession offer - releasing query lock");
		BRUTUS_BrutusProxy_offerSession(priv->brutus_proxy,
						CORBA_OBJECT_NIL,
						"",
						"",
						"",
						ev);
		PRINT_IF_ORBIT2_EX(ev);
	}

out:
	cb = CORBA_Object_is_nil(brutus_logon, ev);
	if (!ORBIT2_EX(ev) && !cb)
		CORBA_Object_release(brutus_logon, ev);

	if (!retv && !camel_exception_is_set(ex)) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not connect to Brutus server"));
	}

	do { // get time zone information - needed at all?
		if (!got_time_zones && !camel_exception_is_set(ex)) {
			struct timeval tv;
			struct timezone tz;
			gchar *tmp_exchange_server = NULL;
			BRUTUS_TIME_OF_DAY_INFO exchange_tod;

			tmp_exchange_server = g_strjoin(NULL, "\\\\", service->url->host, NULL);
			d(tmp_exchange_server);
			if (!tmp_exchange_server)
				break;

			cb = BRUTUS_IMAPISession_GetServerTime(priv->mapi_session, tmp_exchange_server, &exchange_tod, ev);
			g_free(tmp_exchange_server);

			if (cb && (exchange_tod.tod_timezone != -1)) {
				if (gettimeofday(&tv, &tz)) // must not fail!
					break;

				priv->exchange_time_zone = (long)exchange_tod.tod_timezone;
				priv->client_time_zone = (long)tz.tz_minuteswest;
				got_time_zones = TRUE;
			}
		}
	} while (0);

	BRUTUS_STORE_UNLOCK(store, connect_mutex);

	CORBA_exception_free(ev);


	return !camel_exception_is_set(ex);
}

static gboolean
brutus_disconnect(CamelService *service,
		  gboolean clean,
		  CamelException *ex)
{
	CORBA_boolean cb = CORBA_TRUE;
	CamelBrutusStore *store = CAMEL_BRUTUS_STORE(service);
	CamelBrutusStorePrivate *priv = store->priv;

	d("");
	cb = brutus_is_connected(priv);
	if (cb) {
		BRUTUS_STORE_LOCK(store, connect_mutex);
		brutus_mapi_shutdown(store);
		BRUTUS_STORE_UNLOCK(store, connect_mutex);
	}

	return TRUE;
}

/* store virtual method overloads */
static CamelFolder*
brutus_get_folder(CamelStore *store,
		  const char *folder_name,
		  guint32 flags,
		  CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d(folder_name);
	if (!folder_name || !strlen(folder_name))
		goto out;

	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		goto out;

	return camel_brutus_folder_new(store, folder_name, flags, ex);
	d("");

out:
	return NULL;
}

static CamelFolder*
brutus_get_inbox(CamelStore *store,
		 CamelException *ex)
{
	CamelFolder *retv = NULL;
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		goto out;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No support"));
out:
	return retv;
}

static CamelFolder*
brutus_get_trash(CamelStore *store,
		 CamelException *ex)
{
	CamelFolder *retv = NULL;
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		goto out;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No Brutus Trash folder yet"));

out:
	return retv;
}

static CamelFolder*
brutus_get_junk(CamelStore *store,
		CamelException *ex)
{
	CamelFolder *retv = NULL;
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		goto out;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No Brutus Junk folder yet"));

out:
	return retv;
}

static CamelFolderInfo*
brutus_create_folder(CamelStore *store,
		     const char *parent_name,
		     const char *folder_name,
		     CamelException *ex)
{
	CamelFolderInfo *retv = NULL;
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		goto out;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No support"));

out:
	return retv;
}

static void
brutus_delete_folder(CamelStore *store,
		     const char *folder_name,
		     CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		return;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No support"));
}

static void
brutus_rename_folder(CamelStore *store,
		     const char *old_name,
		     const char *new_name,
		     CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		return;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No support"));
}

static void
brutus_sync(CamelStore *store,
	    int expunge,
	    CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		return;
}

static CamelFolderInfo*
brutus_get_folder_info(CamelStore *store,
		       const char *top,
		       guint32 flags,
		       CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;

	d(top);

	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		goto out;

	BRUTUS_PRIV_LOCK(priv, folder_tree_mutex);
	if (!priv->folder_tree)
		brutus_build_folder_tree(store, ex);
	BRUTUS_PRIV_UNLOCK(priv, folder_tree_mutex);

	if (!priv->folder_tree)
		goto out;

	if (!top || (*top == '\0'))
		return priv->folder_tree;

	return (CamelFolderInfo*)g_hash_table_lookup(priv->fi_from_full_name_hash, (gpointer)top);

out:
	d("");
	return NULL;
}

static gboolean
brutus_folder_subscribed(CamelStore *store,
			 const char *folder_name)
{
	gboolean retv = FALSE;
	CamelException *ex = camel_exception_new();
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		goto out;

out:
	camel_exception_free(ex);

	return retv;
}

static void
brutus_subscribe_folder(CamelStore *store,
			const char *folder_name,
			CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		return;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No support"));
}

static void
brutus_unsubscribe_folder(CamelStore *store,
			  const char *folder_name,
			  CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);

	d("");
	brutus_connect((CamelService *)brutus_store, ex);
	if (camel_exception_is_set(ex))
		return;

	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No support"));
}

/* Frees the private structure */
static void
camel_brutus_store_finalize(CamelObject *object)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(object);
	CamelBrutusStorePrivate *priv = brutus_store->priv;


	d("");
	if (priv) {
		CORBA_boolean cb = CORBA_TRUE;
		CORBA_Environment ev[1];
		CORBA_exception_init(ev);

		if (priv->mapi_profile)
			CORBA_free(priv->mapi_profile);

		BRUTUS_PRIV_LOCK(priv, folder_tree_mutex);
		if (priv->folder_tree)
			brutus_free_folder_tree(priv->folder_tree,
						priv->fi_from_entryid_hash,
						priv->fi_from_full_name_hash,
						priv->entryid_from_full_name_hash);
		if (priv->fi_from_entryid_hash)
			g_hash_table_destroy(priv->fi_from_entryid_hash);

		if (priv->fi_from_full_name_hash)
			g_hash_table_destroy(priv->fi_from_full_name_hash);

		if (priv->entryid_from_full_name_hash)
			g_hash_table_destroy(priv->entryid_from_full_name_hash);
		BRUTUS_PRIV_UNLOCK(priv, folder_tree_mutex);

		// free the MAPI session
		d("Shutting down MAPI");
		cb = brutus_is_connected(priv);
		if (cb)
			brutus_mapi_shutdown(brutus_store);

		// destroy the lifeline
		brutus_free_lifeline(priv->brutus_proxy,
				     &priv->lifeline);

		// Deactivate RootPOA
		cb = CORBA_Object_is_nil((CORBA_Object)priv->root_poa, ev);
		if (!ORBIT2_EX(ev) && !cb) {
			brutus_deactivate_poa(priv->root_poa, ev);
			PRINT_IF_ORBIT2_EX(ev);
		}

		/* make the ORB run thread proceed towards ORB destruction */
		cb = CORBA_Object_is_nil((CORBA_Object)priv->orb, ev);
		if (!ORBIT2_EX(ev) && !cb) {
			CORBA_ORB_shutdown(priv->orb, TRUE, ev);
			PRINT_IF_ORBIT2_EX(ev);
		}

		// destroy RootPOA
		cb = CORBA_Object_is_nil((CORBA_Object)priv->root_poa, ev);
		if (!ORBIT2_EX(ev) && !cb) {
			PortableServer_POA_destroy(priv->root_poa, TRUE, FALSE, ev);
			PRINT_IF_ORBIT2_EX(ev);

			CORBA_Object_release((CORBA_Object)priv->root_poa, ev);
		}
		CORBA_exception_free(ev);

		// signal to the ORB_run() thread that the ORB can be destroyed
		g_mutex_lock(priv->proceed_orb_destroy_mutex);
		priv->orb_destroy = TRUE;
		g_cond_signal(priv->proceed_orb_destroy);
		g_mutex_unlock(priv->proceed_orb_destroy_mutex);

		// wait for ORB to destroy cleanly
		if (priv->orb_thread)
			g_thread_join(priv->orb_thread);
		d("CORBA is down");

		// mutexes
		g_mutex_free(priv->proceed_orb_ok_mutex);
		g_mutex_free(priv->proceed_orb_destroy_mutex);
		g_mutex_free(priv->connect_mutex);
		g_mutex_free(priv->folder_tree_mutex);

		// conditionals
		g_cond_free(priv->proceed_orb_ok);
		g_cond_free(priv->proceed_orb_destroy);

		g_free(priv->base_url);
		g_free(priv->storage_path);
		g_free(priv->brutus_logon_str);

		g_free(priv);
		brutus_store->priv = NULL;
	}
}

static void
camel_brutus_store_init(gpointer object,
			gpointer klass)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(object);
	CamelBrutusStorePrivate *priv = g_new0(CamelBrutusStorePrivate, 1);

	d("");
	priv->orb_thread = NULL;
	priv->orb = CORBA_OBJECT_NIL;
	priv->root_poa = CORBA_OBJECT_NIL;

	priv->use_ssliop = FALSE;

	priv->mapi_profile = NULL;
	priv->mapi_session = CORBA_OBJECT_NIL;
	priv->instance_id = 0;

	priv->brutus_proxy = CORBA_OBJECT_NIL;
	priv->lifeline.wrapper = CORBA_OBJECT_NIL;
	priv->lifeline.filling = CORBA_OBJECT_NIL;

	priv->exchange_time_zone = 0;
	priv->client_time_zone = 0;

	priv->mapi_msg_store_table = CORBA_OBJECT_NIL;
	priv->msg_stores = NULL;

	priv->folder_tree = NULL;
	priv->entryid_from_full_name_hash = NULL;
	priv->fi_from_entryid_hash = NULL;
	priv->fi_from_full_name_hash = NULL;

	priv->base_url = NULL;
	priv->storage_path = NULL;
	priv->brutus_logon_str = NULL;

	priv->proceed_orb_ok_mutex = g_mutex_new();
	priv->proceed_orb_ok = g_cond_new();
	priv->corba_ready = FALSE;

	priv->proceed_orb_destroy_mutex = g_mutex_new();
	priv->proceed_orb_destroy = g_cond_new();
	priv->orb_destroy = FALSE;

	priv->connect_mutex = g_mutex_new();
	priv->folder_tree_mutex = g_mutex_new();

	brutus_store->priv = priv;
}

CamelType
camel_brutus_store_get_type(void)
{
	static CamelType camel_brutus_store_type = CAMEL_INVALID_TYPE;

	if (camel_brutus_store_type == CAMEL_INVALID_TYPE) {
		camel_brutus_store_type =
			camel_type_register(camel_offline_store_get_type(),
					    "CamelBrutusStore",
					    sizeof(CamelBrutusStore),
					    sizeof(CamelBrutusStoreClass),
					    (CamelObjectClassInitFunc)camel_brutus_store_class_init,
					    NULL,
					    (CamelObjectInitFunc)camel_brutus_store_init,
					    (CamelObjectFinalizeFunc)camel_brutus_store_finalize);
	}

	return camel_brutus_store_type;
}

gboolean
brutus_insert_in_folder_tree(CamelStore *store,
			     const char *parent_name,
			     const char *full_name,
			     BRUTUS_IMAPIFolder mapi_folder,
			     CamelException *ex)
{
	gboolean retv = FALSE;
	char *name = NULL; // PR_ENTRYID by definition
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelFolderInfo *fi = NULL;
	CamelFolderInfo *self = NULL;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_long n = 0;
	CORBA_unsigned_long r = 0;
	CORBA_string iid = CORBA_string_dup("");
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_BDEFINE obj_type;
	BRUTUS_ENTRYID *parent_eid = NULL;
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_IMAPIFolder parent_folder;
	BRUTUS_IMAPITable hierachy = CORBA_OBJECT_NIL;
	BRUTUS_SRowSet *row_set = NULL;
	CORBA_Environment ev[1];
	static BRUTUS_BDEFINE props[7] = {
		BRUTUS_BRUTUS_PR_ENTRYID,
		BRUTUS_BRUTUS_PR_DISPLAY_NAME,
		BRUTUS_BRUTUS_PR_SUBFOLDERS,
		BRUTUS_BRUTUS_PR_FOLDER_TYPE,
		BRUTUS_BRUTUS_PR_STATUS,
		BRUTUS_BRUTUS_PR_CONTENT_COUNT,
		BRUTUS_BRUTUS_PR_CONTENT_UNREAD
	};
	static BRUTUS_SPropTagArray proptag_array = {
		._maximum = 7,
		._length = 7,
		._buffer = props,
	};
	static BRUTUS_SSortOrderSet sort_order = {
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};

	CORBA_exception_init(ev);

	cb = brutus_is_connected(priv);
	if (!cb) {
		camel_exception_set(ex,
				    CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
				    _("Not connected."));
		goto out_no_lock;
	}

	name = strrchr(full_name, '/');
	if (name)
		name++;
	else
		name = (char*)full_name;


	BRUTUS_PRIV_LOCK(priv, folder_tree_mutex);
	fi = (CamelFolderInfo*)g_hash_table_lookup(priv->fi_from_full_name_hash, parent_name);
	if (!fi) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Internal data corruption - Restart Evolution"));
		goto out;
	}

	parent_eid = (BRUTUS_ENTRYID*)g_hash_table_lookup(priv->entryid_from_full_name_hash, parent_name);
	if (!parent_eid) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Internal data corruption - Restart Evolution"));
		goto out;
	}

	// open parent folder
	br = BRUTUS_IMAPISession_OpenEntry(priv->mapi_session,
					   eid,
					   iid,
					   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
					   &obj_type,
					   &parent_folder,
					   ev);
	CORBA_free(parent_eid);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto out;

	// get hierachy table
	br = BRUTUS_IMAPIFolder_GetHierarchyTable(parent_folder, 0, &hierachy, ev);
	{
		CORBA_Environment ev2[1];
		CORBA_exception_init(ev2);

		BRUTUS_IMAPIFolder_Destroy(parent_folder, 0, ev2);
		PRINT_IF_ORBIT2_EX(ev2);

		CORBA_Object_release(parent_folder, ev2);
		CORBA_exception_free(ev2);
	}
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto out;

	// get all rows
	br = brutus_QueryAllRows(hierachy,
				 &proptag_array,
				 CORBA_OBJECT_NIL,
				 &sort_order,
				 0,
				 &row_set,
				 ev);
	{
		CORBA_Environment ev3[1];
		CORBA_exception_init(ev3);

		BRUTUS_IMAPITable_Destroy(hierachy, 0, ev3);
		PRINT_IF_ORBIT2_EX(ev3);

		CORBA_Object_release(hierachy, ev3);
		CORBA_exception_free(ev3);
	}
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto out;


	if (!row_set->_length)
		goto out;

	// locate the correct child
	for (r = 0;  r <= row_set->_length; r++) {
		if (r == row_set->_length)
			goto out; // not found

		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_DISPLAY_NAME);
		if (-1 == n)
			continue;

		if (!g_ascii_strncasecmp(name, row_set->_buffer[r]._buffer[n].Value._u.lpszA, strlen(name)))
			break;
	}

	/* r is now pointing at the correct entry in the hierarchy table */

	/* hidden system folder? */
	n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_STATUS);
	if (-1 != n) {
		if (row_set->_buffer[r]._buffer[n].Value._u.l & BRUTUS_BRUTUS_FLDSTATUS_HIDDEN) {
#ifdef BRUTUS_SPY
			self = g_new0(CamelFolderInfo, 1);
			self->flags = CAMEL_FOLDER_SYSTEM;
#else
			goto out; // do not show hidden folders unless spying is on
#endif
		}
	} else {
		self = g_new0(CamelFolderInfo, 1);
		self->flags = 0;
	}

	self->next = fi->child;
	fi->child = self;
	self->parent = fi;
	self->uri = brutus_build_uri(priv->base_url, full_name, ex);
	self->name = g_strdup(name);
	self->full_name = g_strdup(full_name);
	self->child = NULL;

	n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_CONTENT_UNREAD);
	if (-1 == n)
		self->unread = -1;
	else
		self->unread = (guint32)row_set->_buffer[r]._buffer[n].Value._u.l;

	n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_CONTENT_COUNT);
	if (-1 == n)
		self->total = -1;
	else
		self->total = (guint32)row_set->_buffer[r]._buffer[n].Value._u.l;

	n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_FOLDER_TYPE);
	if (-1 != n) {
		if (row_set->_buffer[r]._buffer[n].Value._u.l & BRUTUS_BRUTUS_FOLDER_SEARCH)
			self->flags |= CAMEL_FOLDER_VIRTUAL;
	}

	n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ENTRYID);
	if (-1 == n) {
		g_free(self->uri);
		g_free(self->name);
		g_free(self->full_name);
		g_free(self);
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Communication error with server"));
		goto out;
	}
	brutus_sbinary_to_entryid(&row_set->_buffer[r]._buffer[n].Value._u.bin, &eid);
	brutus_add_to_hash_tables(priv, self, eid);

	/* children? */
	n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_SUBFOLDERS);
	if (-1 == n)
		goto out;
	if ((guint32)row_set->_buffer[r]._buffer[n].Value._u.b) {
		self->flags |= CAMEL_FOLDER_CHILDREN;
		brutus_build_folder_hierachy(store, mapi_folder, self, FALSE, ex);
	} else
		self->flags |= CAMEL_FOLDER_NOCHILDREN;

out:
	BRUTUS_PRIV_UNLOCK(priv, folder_tree_mutex);

out_no_lock:
	CORBA_exception_free(ev);
	return retv;
}
