/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Brutus source file for implementing to/from Camel conversion functions.
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "brutus_camel.h"

/* system includes */
#include <string.h>
#include <strings.h>

/* camel includes */
#include <camel/camel-i18n.h>
#include <camel/camel-multipart.h>
#include <camel/camel-mime-utils.h>
#include <camel/camel-stream.h>
#include <camel/camel-stream-mem.h>

/* pure brutus includes */
#include <idl_output/types.h>
#include <idl_output/defines.h>
#include <idl_output/IAddrBook.h>

/* evolution-brutus includes */
#include <server/brutus.h>
#include <server/brutus_corba.h>
#include <server/brutus_mapi.h>
#include <server/brutus_cpid.h>
#include <server/brutus_conv.h>

#define BODY_TEXT_PLAIN 0x01
#define BODY_TEXT_HTML  0x02

static CORBA_char*
brutus_mapi_address_type(const char *addr)
{
	if ((!addr) || !strlen(addr) || (('[' != *addr) && ('/' != *addr)))
		return CORBA_string_dup("SMTP");

	if ('/' == *addr)
		return CORBA_string_dup("EX");

	if (g_strstr_len(addr, strlen("[SMTP:"), "SMTP:"))
		return CORBA_string_dup("SMTP");

	if (g_strstr_len(addr, strlen("[EX:"), "EX:"))
		return CORBA_string_dup("EX");

	if (g_strstr_len(addr, strlen("[X400:"), "X400:"))
		return CORBA_string_dup("X400");

	if (g_strstr_len(addr, strlen("[X500:"), "X500:"))
		return CORBA_string_dup("X500");

	if (g_strstr_len(addr, strlen("[3COM:"), "3COM:"))
		return CORBA_string_dup("3COM");

	if (g_strstr_len(addr, strlen("[ATT:"), "ATT:"))
		return CORBA_string_dup("ATT");

	if (g_strstr_len(addr, strlen("[FAX:"), "FAX:"))
		return CORBA_string_dup("FAX");

	if (g_strstr_len(addr, strlen("[MSFAX:"), "MSFAX:"))
		return CORBA_string_dup("MSFAX");

	if (g_strstr_len(addr, strlen("[MCI:"), "MCI:"))
		return CORBA_string_dup("MCI");

	if (g_strstr_len(addr, strlen("[MHS:"), "MHS:"))
		return CORBA_string_dup("MHS");

	if (g_strstr_len(addr, strlen("[MS:"), "MS:"))
		return CORBA_string_dup("MS");

	if (g_strstr_len(addr, strlen("[MSA:"), "MSA:"))
		return CORBA_string_dup("MSA");

	if (g_strstr_len(addr, strlen("[MSN:"), "MSN:"))
		return CORBA_string_dup("MSN");

	if (g_strstr_len(addr, strlen("[PROFS:"), "PROFS:"))
		return CORBA_string_dup("PROFS");

	if (g_strstr_len(addr, strlen("[SNADS:"), "SNADS:"))
		return CORBA_string_dup("SNADS");

	if (g_strstr_len(addr, strlen("[TELEX:"), "TELEX:"))
		return CORBA_string_dup("TELEX");

	return CORBA_string_dup("SMTP");
}

static gboolean
brutus_set_recipients(CamelMimeMessage *message,
		      BRUTUS_IMessage mapi_message,
		      BRUTUS_IMAPISession mapi_session,
		      PortableServer_POA poa,
		      CamelException *ex)
{
	int n;
	int to_cnt = 0;
	int cc_cnt = 0;
	int bcc_cnt = 0;
	CORBA_long mapi_recip_type;
	gboolean retv = FALSE;
	gboolean need_resolve = FALSE;
	const CamelInternetAddress *to_addr = NULL;
	const CamelInternetAddress *cc_addr = NULL;
	const CamelInternetAddress *bcc_addr = NULL;
	const char *name = NULL;
	const char *addr = NULL;
	char *msg = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_IAddrBook addr_book = CORBA_OBJECT_NIL;
	BRUTUS_ADRLIST addr_list = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	// extract recipients
	to_addr = camel_mime_message_get_recipients(message, CAMEL_RECIPIENT_TYPE_TO);
	cc_addr = camel_mime_message_get_recipients(message, CAMEL_RECIPIENT_TYPE_CC);
	bcc_addr = camel_mime_message_get_recipients(message, CAMEL_RECIPIENT_TYPE_BCC);

	// initiate MAPI address list
	addr_list._maximum =
		(to_addr ? camel_address_length(CAMEL_ADDRESS(to_addr)) : 0)
		+ (cc_addr ? camel_address_length(CAMEL_ADDRESS(cc_addr)) : 0)
		+ (bcc_addr ? camel_address_length(CAMEL_ADDRESS(bcc_addr)) : 0);
	addr_list._length = addr_list._maximum;
	addr_list._buffer = BRUTUS_ADRLIST_allocbuf(addr_list._maximum);

	// transfer addresses
	mapi_recip_type = BRUTUS_BRUTUS_MAPI_TO;
	for (n = 0; n < addr_list._length; n++) {
		name = NULL;
		addr = NULL;
		if (BRUTUS_BRUTUS_MAPI_TO == mapi_recip_type) {
			if (!camel_internet_address_get(to_addr, to_cnt++, &name, &addr))
				mapi_recip_type = BRUTUS_BRUTUS_MAPI_CC;
		}
		if (BRUTUS_BRUTUS_MAPI_CC == mapi_recip_type) {
			if (!camel_internet_address_get(cc_addr, cc_cnt++, &name, &addr))
				mapi_recip_type = BRUTUS_BRUTUS_MAPI_BCC;
		}
		if (BRUTUS_BRUTUS_MAPI_BCC == mapi_recip_type) {
			if (!camel_internet_address_get(bcc_addr, bcc_cnt++, &name, &addr))
				break;
		}

		// with or without PR_EMAIL_ADDRESS
		if (!need_resolve)
			need_resolve = !(addr || strlen(addr));

		addr_list._buffer[n].ulReserved1 = 0;
		addr_list._buffer[n].rgPropVals._maximum = 3 + !!(addr && strlen(addr)); // make optionally room for PR_EMAIL_ADDRESS
		addr_list._buffer[n].rgPropVals._length = addr_list._buffer[n].rgPropVals._maximum;
		addr_list._buffer[n].rgPropVals._buffer = BRUTUS_seq_SPropValue_allocbuf(addr_list._buffer[n].rgPropVals._maximum),

			addr_list._buffer[n].rgPropVals._buffer[0].ulPropTag = BRUTUS_BRUTUS_PR_RECIPIENT_TYPE;
		addr_list._buffer[n].rgPropVals._buffer[0].Value._d = BRUTUS_BRUTUS_PT_LONG;
		addr_list._buffer[n].rgPropVals._buffer[0].Value._u.l = mapi_recip_type;

		addr_list._buffer[n].rgPropVals._buffer[1].ulPropTag = BRUTUS_BRUTUS_PR_DISPLAY_NAME;
		addr_list._buffer[n].rgPropVals._buffer[1].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		addr_list._buffer[n].rgPropVals._buffer[1].Value._u.lpszA = name ? CORBA_string_dup(name) : (addr ? CORBA_string_dup(addr) : CORBA_string_dup(""));

		addr_list._buffer[n].rgPropVals._buffer[2].ulPropTag = BRUTUS_BRUTUS_PR_ADDRTYPE;
		addr_list._buffer[n].rgPropVals._buffer[2].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		addr_list._buffer[n].rgPropVals._buffer[2].Value._u.lpszA = brutus_mapi_address_type(addr);

		if (addr_list._buffer[n].rgPropVals._length <= 3)
			continue;

		addr_list._buffer[n].rgPropVals._buffer[3].ulPropTag = BRUTUS_BRUTUS_PR_EMAIL_ADDRESS;
		addr_list._buffer[n].rgPropVals._buffer[3].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		addr_list._buffer[n].rgPropVals._buffer[3].Value._u.lpszA = CORBA_string_dup(addr);
	}

	if (need_resolve) { // some transport providers require a call to ResolveName() regardless
		br = BRUTUS_IMAPISession_OpenAddressBook(mapi_session, "", BRUTUS_BRUTUS_MAPI_ACCESS_MODIFY, &addr_book, ev);
		if ((BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br) && (BRUTUS_BRUTUS_S_OK != br)) {
			d("Could not open address book");
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Could not open the Exchange address book for address resolution"));
			goto out;
		}

		br = BRUTUS_IAddrBook_ResolveName(addr_book, 0, &addr_list, ev);
		brutus_release_object(BRUTUS_IAddrBook_tc, &addr_book, poa, ev);

		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			msg = g_strdup("");
			for (n = 0; n < addr_list._length; n++) {
				if (addr_list._buffer[n].rgPropVals._length >= 3)
					msg = g_strjoin("\n", msg, addr_list._buffer[n].rgPropVals._buffer[1].Value._u.lpszA, NULL); // display name
			}
			camel_exception_setv(ex, CAMEL_EXCEPTION_SYSTEM,
					     "%s:%s",
					     _("Some recipients could not be resolved"), msg);
			g_free(msg);
			goto out;
		}
	}

	// set the recipients
	br = BRUTUS_IMessage_ModifyRecipients(mapi_message, 0, &addr_list, ev);
	if (BRUTUS_BRUTUS_S_OK != br) {
		d("Could not set recipients");
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not set recipients on outgoing message"));
		goto out;
	}

	retv = TRUE;

out:
	brutus_free_addrlist_content(&addr_list, poa);
	CORBA_exception_free(ev);
	return retv;
}

gboolean
brutus_camel_mime_message_to_mapi(CamelMimeMessage *message,
				  BRUTUS_IMessage mapi_message,
				  BRUTUS_IMAPISession mapi_session,
				  PortableServer_POA poa,
				  const BRUTUS_seq_SPropValue *ex_prop_values,
				  CamelException *ex)
{
	int n;
	int bare_prop_count; // later adjusted to allow for text/[plain,html] or multipart/alternative
	size_t msg_prop_count;
	gboolean retv = FALSE;
	guint8 body_flags = 0x00; // plain, html or both
	CORBA_string text_plain;  // plain body text
	CORBA_string text_html;   // html body text
	//
	// used for mime parts
	guint part_count;
	CamelMultipart *multi = NULL;
	CamelMimePart *part = NULL;
	CamelStreamMem *content = NULL;
	CamelContentType *ct = NULL;
	CamelDataWrapper *dw = NULL;
	const char *disposition = NULL;
	const char *filename = NULL;
	const char *charset = NULL;
	char *mime_type = NULL;
	char *buffer = NULL;
	//
	// used if the mime type is multipart/alternative
	CamelMimePart *temp_part = NULL;
	CamelStreamMem *temp_content = NULL;
	CamelDataWrapper *temp_dw = NULL;
	//
	// MAPI properties and interfaces
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_SPropProblemArray *mapi_problems = NULL;
	BRUTUS_seq_SPropValue msg_prop_values = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	// set recipients on MAPI message
	if (!brutus_set_recipients(message, mapi_message, mapi_session, poa, ex))
		goto out;

	//
	// message body
	//

	/** Get the mime parts from CamelMimemessge **/
	multi = (CamelMultipart*)camel_medium_get_content_object(CAMEL_MEDIUM(message));
	if (!multi) {
		d("Could not get content object");
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Mal-formatted CamelMimeMessage: No content object"));
		return FALSE;
	}

	//
	// The following if-else is modified from camel_groupwise_utils.c::camel_groupwise_util_item_from_message()
	//
	if (CAMEL_IS_MULTIPART(multi)) {
		part_count = camel_multipart_get_number(multi);
		for (n = 0; n < part_count; n++) {
			content = (CamelStreamMem*)camel_stream_mem_new();

			/*
			 * Assuming the first part always is the actual message
			 * and an attachment otherwise...
			 */
			part = camel_multipart_get_part(multi, n);
			ct = camel_mime_part_get_content_type(part);
			dw = camel_medium_get_content_object(CAMEL_MEDIUM(part));

			if (ct->subtype && !g_ascii_strcasecmp(ct->subtype, "alternative")) {
				temp_content = (CamelStreamMem*)camel_stream_mem_new();

				temp_part = camel_multipart_get_part((CamelMultipart*)dw, 1);
				if (temp_part) {
					temp_dw = camel_medium_get_content_object(CAMEL_MEDIUM(temp_part));
					camel_data_wrapper_write_to_stream(temp_dw, (CamelStream *)temp_content);
					buffer = g_malloc0(temp_content->buffer->len+1);
					buffer = memcpy(buffer, temp_content->buffer->data, temp_content->buffer->len);
					filename = camel_mime_part_get_filename(temp_part);
					disposition = camel_mime_part_get_disposition(temp_part);
					mime_type = camel_data_wrapper_get_mime_type(temp_dw);
//					send_as_attachment(cnc, item, temp_content, buffer, ct, temp_dw, filename, &attach_list);
					g_free(buffer);
					g_free(mime_type);
				}
				camel_object_unref(temp_content);
				camel_object_unref(temp_dw);
				camel_object_unref(dw);
				continue;
			}

			camel_data_wrapper_write_to_stream(dw, (CamelStream *)content);
			buffer = g_malloc0(content->buffer->len+1);
			buffer = memcpy(buffer, content->buffer->data, content->buffer->len);
			filename = camel_mime_part_get_filename(part);
			disposition = camel_mime_part_get_disposition(part);
			mime_type = camel_data_wrapper_get_mime_type(dw);

			if (0 == n && !g_ascii_strcasecmp(mime_type, "text/plain") ) {
/*				e_gw_item_set_content_type(item, mime_type); */
/*				e_gw_item_set_message(item, buffer); */
			} /* else  */
/*				send_as_attachment(cnc, item, content, buffer, ct, dw, filename, &attach_list); */

			g_free(buffer);
			g_free(mime_type);
			camel_object_unref(content);
			camel_object_unref(dw);
		}
	} else { // plain message
		ct = camel_mime_part_get_content_type((CamelMimePart *)message);
		dw = camel_medium_get_content_object(CAMEL_MEDIUM(message));
		charset = camel_content_type_param(dw->mime_type, "charset");

		content = (CamelStreamMem*)camel_stream_mem_new();
		camel_data_wrapper_write_to_stream(dw, (CamelStream *)content);
		camel_object_unref(dw);

		// body text: plain or html?
		if (!g_ascii_strcasecmp(ct->type, "text")) {
			do { // testing for supported text subtypes
				if (!g_ascii_strcasecmp(ct->subtype, "plain")) {
					body_flags |= BODY_TEXT_PLAIN;
					break;
				}

				if (!g_ascii_strcasecmp(ct->subtype, "html")) {
					body_flags |= BODY_TEXT_HTML;
					break;
				}
			} while (0);
		} else {
			/* only an attachment is present */
		}

		// copy body text, plain or html not both
		if (body_flags & BODY_TEXT_PLAIN) {
			text_plain = CORBA_string_alloc(content->buffer->len);
			memcpy(text_plain, content->buffer->data, content->buffer->len);
			text_plain[content->buffer->len] = '\0';
		} else if (body_flags & BODY_TEXT_HTML) {
			text_html = CORBA_string_alloc(content->buffer->len);
			memcpy(text_html, content->buffer->data, content->buffer->len);
			text_html[content->buffer->len] = '\0';
		} else {
			// no body text - copy attachment data
		}
		camel_object_unref(content);
	}

	// initiate final msg prop sequence
	bare_prop_count = 2 + !!(body_flags & BODY_TEXT_PLAIN) + !!(body_flags & BODY_TEXT_HTML); // allow simultaneous html and plain body text
	msg_prop_count = bare_prop_count + ex_prop_values->_length;

	msg_prop_values._maximum = msg_prop_count;
	msg_prop_values._length = bare_prop_count;
	msg_prop_values._buffer = BRUTUS_seq_SPropValue_allocbuf(msg_prop_values._maximum);

	// subject
	msg_prop_values._buffer[0].ulPropTag = BRUTUS_BRUTUS_PR_SUBJECT;
	msg_prop_values._buffer[0].Value._d = BRUTUS_BRUTUS_PT_STRING8;
	msg_prop_values._buffer[0].Value._u.lpszA = CORBA_string_dup(camel_mime_message_get_subject(message));

	// charset encoding
	msg_prop_values._buffer[1].ulPropTag = BRUTUS_BRUTUS_PR_INTERNET_CPID;
	msg_prop_values._buffer[1].Value._d = BRUTUS_BRUTUS_PT_LONG;
	msg_prop_values._buffer[1].Value._u.l = charset ? brutus_label_to_cpid(charset) : brutus_label_to_cpid("UTF-8"); // defaulting to UTF-8

	// body text
	if ((body_flags & BODY_TEXT_PLAIN)
	    && (body_flags & BODY_TEXT_HTML)) {
		msg_prop_values._buffer[msg_prop_values._length - 2].ulPropTag = BRUTUS_BRUTUS_PR_BODY;
		msg_prop_values._buffer[msg_prop_values._length - 2].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		msg_prop_values._buffer[msg_prop_values._length - 2].Value._u.lpszA = text_plain;
		msg_prop_values._buffer[msg_prop_values._length - 1].ulPropTag = BRUTUS_BRUTUS_PR_BODY_HTML;
		msg_prop_values._buffer[msg_prop_values._length - 1].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		msg_prop_values._buffer[msg_prop_values._length - 1].Value._u.lpszA = text_html;
	} else if (body_flags & BODY_TEXT_PLAIN) {
		msg_prop_values._buffer[msg_prop_values._length - 1].ulPropTag = BRUTUS_BRUTUS_PR_BODY;
		msg_prop_values._buffer[msg_prop_values._length - 1].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		msg_prop_values._buffer[msg_prop_values._length - 1].Value._u.lpszA = text_plain;
	} else if (body_flags & BODY_TEXT_HTML) {
		msg_prop_values._buffer[msg_prop_values._length - 1].ulPropTag = BRUTUS_BRUTUS_PR_BODY_HTML;
		msg_prop_values._buffer[msg_prop_values._length - 1].Value._d = BRUTUS_BRUTUS_PT_STRING8;
		msg_prop_values._buffer[msg_prop_values._length - 1].Value._u.lpszA = text_html;
	}

	// append the additional properties
	if (ex_prop_values && ex_prop_values->_length) {
		for (n = 0; n < ex_prop_values->_length; n++)
			brutus_copy_spropvalue(&msg_prop_values._buffer[n + msg_prop_values._length], &ex_prop_values->_buffer[n]);

		msg_prop_values._length = msg_prop_values._maximum;
	}

	mapi_problems = BRUTUS_seq_SPropProblem__alloc();
	mapi_problems->_maximum = 0;
	mapi_problems->_length = 0;
	mapi_problems->_buffer = NULL;
	br = BRUTUS_IMessage_SetProps(mapi_message, &msg_prop_values, FALSE, &mapi_problems, ev);
	CORBA_free(mapi_problems);
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not save changes to outgoing message"));
		goto out;
	}

	// all is well...
	retv = TRUE;

out:
	brutus_free_seq_spropvalue_content(&msg_prop_values, poa);

	CORBA_exception_free(ev);
	return retv;
}
