/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Implementation file for Brutus string utility functions.
 *    Copyright (C) 2005 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "brutus.h"
#include "brutus_string.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>

gboolean
brutus_get_xbit_number(const guint8 bits,
		       const char *num_str,
		       guint64 *ul)
{
	guint32 max;

	switch (bits) {
	case 8 :
		max = 0xff;
		break;
	case 16 :
		max = 0xffff;
		break;
	case 32 :
		max = 0xffffffff;
		break;
	default :
		return FALSE;
	}

	if (!num_str)
		return FALSE;

	if (*num_str == '-')
		return FALSE;

	char *tmp;
	unsigned long long num = strtoull(num_str, &tmp, 10);
	if (errno == ERANGE)
		return FALSE;
	
	if (*tmp != '\0')
		return FALSE;

	if (num > max)
		return FALSE;

	*ul = (guint64)num;

	return TRUE;
}

char*
brutus_append_string(char **dest, 
		     const char *src)
{
	char *retv;
	size_t size;


	if (!*dest) {
		size = sizeof(char) * (strlen(src) + 1);
		retv = (char*)calloc(size, sizeof(char));
	} else {
		size = sizeof(char) * (strlen(*dest) + strlen(src) + 1);
		retv = (char*)realloc((void*)*dest, size);
	}
	if (!retv) {
		free(*dest);
		return NULL;
	}

	strcat(retv, src);
	*dest = retv;

	return *dest;
}

char*
brutus_to_string(const char *fmt, 
		 ...) 
{
	int n, size = 100;
	char *p, *np;
	va_list ap;

	if ((p = malloc(size)) == NULL)
		return NULL;

	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf(p, size, fmt, ap);
		va_end(ap);
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			return p;
		/* Else try again with more space. */
		if (n > -1)    /* glibc 2.1 */
			size = n+1; /* precisely what is needed */
		else           /* glibc 2.0 */
			size *= 2;  /* twice the old size */
		if ((np = realloc(p, size)) == NULL) {
			free(p);
			return NULL;
		} else {
			p = np;
		}
	}
}

void
brutus_append_to_string(char **dest,
			const char *fmt, 
			...) 
{
	int n, size = 100;
	char *p, *np;
	va_list ap;

	if ((p = malloc(size)) == NULL)
		return;

	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf(p, size, fmt, ap);
		va_end(ap);
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			break;
		/* Else try again with more space. */
		if (n > -1)    /* glibc 2.1 */
			size = n+1; /* precisely what is needed */
		else           /* glibc 2.0 */
			size *= 2;  /* twice the old size */
		if ((np = realloc(p, size)) == NULL) {
			free(p);
			return;
		} else {
			p = np;
		}
	}

	brutus_append_string(dest, (const char*)p);
	free(p);
}

char*
brutus_build_init_ref_opt(const gboolean use_ssl,
			  const char *init_ref,
			  const char *ior_table_key,
			  const char *server_host,
			  const guint16 server_port)
{
	char *init_ref_opt = strdup("--ORBInitRef=");
	
	init_ref_opt = brutus_append_string(&init_ref_opt, init_ref);
	if (!init_ref_opt) 
		goto out;

	init_ref_opt = brutus_append_string(&init_ref_opt, use_ssl ? "=corbaloc:ssliop:" : "=corbaloc:iiop:");
	if (!init_ref_opt) 
		goto out;

	init_ref_opt = brutus_append_string(&init_ref_opt, server_host);
	if (!init_ref_opt) 
		goto out;

	init_ref_opt = brutus_append_string(&init_ref_opt, ":");
	if (!init_ref_opt) 
		goto out;

	char server_port_str[4];
	snprintf(server_port_str, 4, "%hu", server_port);
	init_ref_opt = brutus_append_string(&init_ref_opt, server_port_str);
	if (!init_ref_opt) 
		goto out;

	init_ref_opt = brutus_append_string(&init_ref_opt, "/");
	if (!init_ref_opt) 
		goto out;
	
	init_ref_opt = brutus_append_string(&init_ref_opt, ior_table_key);
	if (!init_ref_opt) 
		goto out;

out:
	return init_ref_opt;
}

char*
brutus_build_corbaloc_string(const char *brutus_server,
			     const char *ior_table_key,
			     const gboolean use_ssl,
			     const guint16 brutus_port)
{
	char port[32] = { '\0' };
	char *retv = NULL;

	snprintf(port, sizeof(port), "%hu", brutus_port);

	brutus_append_string(&retv, use_ssl ? "corbaloc:ssliop:" : "corbaloc:iiop:");
	brutus_append_string(&retv, (const char*)brutus_server);
	brutus_append_string(&retv, ":");
	brutus_append_string(&retv, (const char*)port);
	brutus_append_string(&retv, "/");
	brutus_append_string(&retv, (const char*)ior_table_key);

	return retv;
}
